#!/bin/bash

if [ $# -ne 10 ]
then
    echo "<app> [prefix] [reverse (true|false)]  [data_size] [model_init(default|ibm1-n|hmm-2)] [model_from] [model_to] [model(ibm1|hmm)] [hadoop(true|false)] p0 sample"
    exit
fi

log_file="aer.log"

echo "AER: --------------------------------" >> $log_file
echo "AER: --------------------------------" >> $log_file
echo "AER: " >> $log_file
echo "AER: `date`" >> $log_file
echo "AER: " >> $log_file
echo "AER: --------------------------------" >> $log_file

# directory for word alignment executables
bin=../src/bin

lang=$1

train_file=$lang.train
test_file=$lang.test
afile_file=$lang.afile

reverse=$2
size=$3

model_init=$4
model_from=$5
model_to=$6
model=$7
hadoop=$8
p0=$9
shift
sample=$9

training="yes"

if [ $training = "yes" ]
then
    if [ $hadoop = true ]
    then
	sh ./restart.sh 2> /dev/null
    fi
fi

########################
# id
########################
id=$lang-$reverse-$size
########################

echo "AER: ---------------------------" >> $log_file
echo "AER: (model=$model) lang=$lang direction=$reverse size=$size p0=$p0 sample=$sample hadoop=$hadoop" >> $log_file
echo "AER: ---------------------------" >> $log_file


# build a config file on the fly.
config_file="./config-$id-$sample-$model.ini"
rm -f $config_file
touch $config_file
path="../ini"
cat $path/config-default.ini >> $config_file
echo "" >> $config_file
echo "# from bash.sh" >> $config_file
echo "freq-word-count = 10" >> $config_file
echo "log_file = $log_file" >> $config_file
echo "data-format-reverse = $reverse" >> $config_file
echo "sampling-type = hmm" >> $config_file
echo "training-only = false" >> $config_file
echo "uniform-jump = false" >> $config_file
echo "" >> $config_file
cat $path/config-model-$model.ini >> $config_file

# $path/config-1-cut-is-10.ini  $path/config-2-reverse-is-$reverse.ini  $path/config-3-sample-is-hmm.ini  $path/config-4-decoding-is-true.ini $path/config-8-uniform-jump-is-false.ini $path/config-model-$model.ini > ./config.ini
# hadoop_options="$path/config-0-default.ini $path/config-1-cut-is-10.ini  $path/config-2-reverse-is-$reverse.ini  $path/config-3-sample-is-hmm.ini  $path/config-4-decoding-is-true.ini $path/config-8-uniform-jump-is-false.ini $path/config-model-$model.ini"

local_options=$config_file
hadoop_options=$config_file

#######################################################
# prepare data
#######################################################

if [ $training = "yes" ]
then
    cp $test_file bitext.test
    cp $afile_file bitext.afile
    
    head -n $size $train_file > bitext.train.$size
    cat bitext.train.$size bitext.test > bitext.train
    wc bitext.train

    # cp bitext.train bitext.test
fi

#######################################################
# EM
#######################################################

if [ $training = "yes" ]
then
    
    if [ $model_init = "default" ]
    then
	$bin/$model-mr-main i 0 $p0 $sample $hadoop $local_options < ./bitext.train > ./parameter-$id-0-$model-default.txt
	if [ $? -ne 0 ] 
	then 
	    echo "init failed"
	    exit 
	fi
	
	file="parameter-$id-0-$model-default.txt"
    else
    #################
    # continuation
	file="parameter-$id-0-$model_init.txt"
    fi

    if [ -e $file ]
    then
	echo "cp $file parameter.txt"
	cp $file parameter.txt 
	ls -lrt parameter.txt 
    else
	echo "Can not find $file"
	exit
    fi
fi

## only want to get parameter.txt
# exit

######## set up hadoop
if [ $training = "yes" ]
then
    if [ $hadoop = true ]
    then
	~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop dfs -rmr a 2> /dev/null
	~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop dfs -mkdir a
	~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop dfs -put ./bitext.train a
	~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop dfs -ls a
    fi
fi

for ((i=$model_from; i <= $model_to ; i++))
do
    # MOSES
    if [ $i -ne 5 ]
    then
	cp $test_file bitext.test
    else
	cp bitext.train bitext.test
    fi

    if [ $training = "yes" ]
    then

	if [ $hadoop = "true" ]
	then
	    ~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop dfs -rmr my-output 2> /dev/null
	    echo $cmd
	    ~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop jar $HADOOP_HOME/contrib/streaming/hadoop-0.20.2-streaming.jar -input a/bitext.train -output my-output -mapper "$bin/$model-mr-main m $i $p0 $sample $hadoop $hadoop_options" -reducer NONE -file ~/ini/config*.ini -file $bin/$model-mr-main -file parameter.txt
	else
	    $bin/$model-mr-main m $i $p0 $sample $hadoop $local_options < bitext.train > expected-count-$id-$sample-$model-$i.txt
	fi

	if [ $? -ne 0 ] 
	then 
	    echo "mapper failed"
	    exit 
	fi
	
	if [ $hadoop = "true" ]
	then
	    ~/project_/zhao_/third-party/hadoop-0.20.2/bin/hadoop dfs -cat my-output/part-* > expected-count-$id-$sample-$model-$i.txt
	    ls -lrt expected-count-$id-$sample-$model-$i.txt
	fi

	cp expected-count-$id-$sample-$model-$i.txt expected-count.txt
	ls -lrt expected-count.txt

	$bin/$model-mr-main r $i $p0 $sample $hadoop $local_options < expected-count.txt > parameter-$id-$sample-$model-$i.txt
	
	if [ $? -ne 0 ] 
	then 
	    echo "reducer failed"
	    exit 
	fi
	
	cp parameter-$id-$sample-$model-$i.txt parameter.txt
	ls -lrt parameter.txt

        # save space
	rm -f expected-count-$id-$sample-$model-$i.txt
    fi

    if [ $training = "no" ]
    then
	# $bin/$model-mr-main r $i $p0 $sample $hadoop $local_options < parameter-$id-$sample-$model-$i.txt > /dev/null
	echo "here"
	ls -lrt parameter-$id-$sample-$model-$i.txt
	echo "$bin/$model-mr-main r $i $p0 $sample $hadoop $local_options < parameter-$id-$sample-$model-$i.txt"

	$bin/$model-mr-main r $i $p0 $sample $hadoop $local_options < parameter-$id-$sample-$model-$i.txt > /dev/null

    
	if [ $? -ne 0 ] 
	then 
	    echo "reducer failed"
	    exit 
	fi
    fi

    
    if [ $model = "ibm1-classic" ]
    then
	cp ibm1-$i-ibm1-ibm1.afile ibm1-$i-ibm1-ibm1-$id-$sample.afile 
    elif [ $model = "hmm-classic-empty-2" ]
    then
	cp hmm-$i-hmm-ibm1.afile hmm-$i-hmm-ibm1-$id-$sample.afile
	cp hmm-$i-hmm-hmm.afile hmm-$i-hmm-hmm-$id-$sample.afile
    elif [ $model = "hmm-fertility-empty" ]
    then
	cp hmmf-$i-hmmf-ibm1.afile hmmf-$i-hmmf-ibm1-$id-$sample.afile
	cp hmmf-$i-hmmf-hmm.afile hmmf-$i-hmmf-hmm-$id-$sample.afile
    else
	echo "Unknown model: $model"
    fi

done


echo "AER: --------------------------------" >> $log_file
echo "AER: " >> $log_file
echo "AER: `date`" >> $log_file
echo "AER: " >> $log_file
echo "AER: --------------------------------" >> $log_file
echo "AER: --------------------------------" >> $log_file



