
#include <algorithm>
// #include <cmath>
#include "config/global.h"
#include "config/get-neighbors.h"

using namespace std;

/*
  We build all possible neighbor based on the traning bitext.
  It is a quaratic function of the target or source vocabulary.
*/

namespace MYCONFIG {

  unordered_map<string, unordered_set<string> > Get_Neighbors::all_candidates;
  unordered_map<string, int> Get_Neighbors::all_candicounts;
  /*
    neighbor type:
   
    del-1
    rep-1
    train
    len
 
  */
  
  void Get_Neighbors::get_nbs(const string& line, const int line_num, const string& neighbor_type, vector<string>* neighbors) {
    const string& file = Global::FLAGS_train_file;

    assert(neighbors);
    assert(neighbors->size() == 0);

    // get bitext
    vector<string> bitext;
    UTIL::split_string_using(line, "\t", &bitext);
    if (bitext.size() < 2) {  // empty or 1
      return;
    }
    assert(bitext.size() == 2);

    // get target words
    vector<string> trg;
    UTIL::split_string_using(bitext[0], " ", &trg);
    if (trg.size() == 0) {
      return;
    }
    set<string> set_trg;
    for (vector<string>::const_iterator it = trg.begin(); it != trg.end(); ++it) {
      set_trg.insert(*it);
    }
    assert(trg.size() == set_trg.size());  // because pre-processing has been done

    // get source words
    vector<string> src;
    UTIL::split_string_using(bitext[1], " ", &src);
    if (src.size() == 0) {
      return;
    }
    set<string> set_src;
    for (vector<string>::const_iterator it = src.begin(); it != src.end(); ++it) {
      set_src.insert(*it);
    }
    assert(src.size() == set_src.size());  // because pre-processing has been done

    // output the original bitext
    neighbors->push_back(line);

    // now we consider various kinds of neighbors...

    if (neighbor_type == "del-1") {
      // delete a word from source or target sentence

      // remove one target word
      if (trg.size() > 1) {
	for (int i = 0; i < (int)trg.size(); ++i) {  // the word at i will be removed
	  string str = "";
	  for (int j = 0; j < (int)trg.size(); ++j) {
	    if (j != i) {
	      str += trg[j] + " ";
	    }
	  }
	  str += "\t" + bitext[1];
	  neighbors->push_back(str);
	}
      }
      // remove one source word
      if (src.size() > 1) {
	for (int i = 0; i < (int)src.size(); ++i) {  // the word at i will be removed
	  string str = "";
	  str += bitext[0] + "\t";
	  for (int j = 0; j < (int)src.size(); ++j) {
	    if (j != i) {
	      str += src[j] + " ";
	    }
	  }
	  neighbors->push_back(str);
	}
      }
    } else if (neighbor_type == "train") {
      // all the training data

      assert(neighbors->size() == 1);

      string line2;

      ifstream in(file.c_str());
      if (!in.good()) {
	cerr << "Can not open " << file << endl;
	return;
      }
      while (UTIL::getline(in, &line2)) {
	vector<string> v;
	UTIL::split_string_using(line2, "\t", &v);
	if (v.size() != 2) {
	  continue;
	}
	
	neighbors->push_back(v[0] + "\t" + v[1]);
      }  // while
      in.close();

    } else if (neighbor_type == "any2") {
      // regard the current bitext,
      // any target and source sentence in the training data

      assert(neighbors->size() == 1);

      string sen1, sen2;

      ifstream in1(file.c_str());
      if (!in1.good()) {
	cerr << "Can not open " << file << endl;
	return;
      }
      while (UTIL::getline(in1, &sen1)) {
	vector<string> v1;
	UTIL::split_string_using(sen1, "\t", &v1);
	if (v1.size() != 2) {
	  continue;
	}
	
	ifstream in2(file.c_str());
	if (!in2.good()) {
	  cerr << "Can not open " << file << endl;
	  return;
	}
	while (UTIL::getline(in2, &sen2)) {
	  vector<string> v2;
	  UTIL::split_string_using(sen2, "\t", &v2);
	  if (v2.size() != 2) {
	    continue;
	  }

	  neighbors->push_back(v1[0] + "\t" + v2[1]);
	  // neighbors->push_back(v2[0] + "\t" + v1[1]);
	} // while
      }  // while
      in1.close();

    } else if (neighbor_type == "any1") {
      // any target sentence in the training data with the current source sentence
      // or 
      // any source sentence in the training data with the current target sentence

      assert(neighbors->size() == 1);

      ifstream in(file.c_str());
      if (!in.good()) {
	cerr << "Can not open " << file << endl;
	return;
      }

      string line2;

      while (UTIL::getline(in, &line2)) {
	vector<string> v;
	UTIL::split_string_using(line2, "\t", &v);
	if (v.size() != 2) {
	  continue;
	}

	if (line2 == line) {
	  // continue;
	}
	neighbors->push_back(bitext[0] + "\t" + v[1]);
	neighbors->push_back(v[0] + "\t" + bitext[1]);
      }  // while
      in.close();

    } else if (neighbor_type == "any-src") {
      // any source sentence in the training data with the current target sentence

      assert(neighbors->size() == 1);

      ifstream in(file.c_str());
      if (!in.good()) {
	cerr << "Can not open " << file << endl;
	return;
      }

      string line2;

      while (UTIL::getline(in, &line2)) {
	vector<string> v;
	UTIL::split_string_using(line2, "\t", &v);
	if (v.size() != 2) {
	  continue;
	}

	if (line2 == line) {
	  // continue;
	}
	neighbors->push_back(bitext[0] + "\t" + v[1]);

      }  // while
      in.close();

    } else if (neighbor_type == "nextk-src") {
      // any source sentence in the training data with the current target sentence

      assert(neighbors->size() == 1);

      ifstream in(file.c_str());
      if (!in.good()) {
	cerr << "Can not open " << file << endl;
	return;
      }

      string line2;

      int line_count = 0;
      while (UTIL::getline(in, &line2)) {
	++line_count;

	// only get next k src
	if (line_count - 1 < line_num) {
	  continue;
	} 

	// skip the same line
	if (line_count - 1 == line_num) {
	    continue;
	}

	if ((int)neighbors->size() >= Global::FLAGS_neighbor_type_nextk_count) {
	  break;
	}

	vector<string> v;
	UTIL::split_string_using(line2, "\t", &v);
	if (v.size() != 2) {
	  continue;
	}

	if (line2 == line) {
	  // continue;
	}
	neighbors->push_back(bitext[0] + "\t" + v[1]);

      }  // while

      in.close();

      if ((int)neighbors->size() < Global::FLAGS_neighbor_type_nextk_count) {
	ifstream in(file.c_str());
	if (!in.good()) {
	  cerr << "Can not open " << file << endl;
	  return;
	}
	
	string line2;
	
	int line_count = 0;
	while (UTIL::getline(in, &line2)) {
	  ++line_count;
	  
	  // skip the same line
	  if (line_count - 1 == line_num) {
	    continue;
	  }

	  if ((int)neighbors->size() >= Global::FLAGS_neighbor_type_nextk_count) {
	    break;
	  }
	  
	  vector<string> v;
	  UTIL::split_string_using(line2, "\t", &v);
	  if (v.size() != 2) {
	    continue;
	  }
	  
	  if (line2 == line) {
	    // continue;
	  }
	  neighbors->push_back(bitext[0] + "\t" + v[1]);
	  
	}  // while
	
	in.close();
      }  // if not enough

    } else if (neighbor_type == "change-1-in-src") {
      // use co-occur words to generate sentences.
      // Hope this one will solve the fertility issue.

      assert(neighbors->size() == 1);
      
      unordered_set<string> my_candidates;
      for (int i = 0; i < (int)trg.size(); ++i) {
	const string& t = trg[i];
	unordered_set<string>& m = all_candidates[t];
	for (unordered_set<string>::const_iterator it = m.begin(); it != m.end(); ++it) {
	  const string& tmp = *it;
	  if (Global::FLAGS_neighbor_with_only_new_words) {
	    // do not use words that are in the current translation
	    if (set_src.find(tmp) == set_src.end()) {  
	      my_candidates.insert(tmp);
	    }
	  } else {
	    my_candidates.insert(tmp);
	  }
	}
      }

      if (my_candidates.size() < 1) {
	cerr << "no neighborhood!" << endl;
	return;
      }

      // change each src
      int total_count = (int)my_candidates.size();
      if (Global::FLAGS_neighbor_type_change_1_in_src_count > 0) {
	total_count = min(Global::FLAGS_neighbor_type_change_1_in_src_count, (int)my_candidates.size());
      }

      cerr << "all-candidates: " << all_candidates.size() << endl;
      cerr << "my-candidates: " << my_candidates.size() << endl;
      cerr << "total-count: " << total_count << endl;

      vector<string> vec;
      for (unordered_set<string>::const_iterator it = my_candidates.begin(); it != my_candidates.end(); ++it) {
	vec.push_back(*it);
      }

      for (int len0 = 0; len0 < (int)src.size(); ++len0) {
	for (int count = 0; count < total_count; ++count) {
	  string nb = "";
	  for (int len = 0; len < len0; ++len) {
	    nb += src[len] + " ";
	  }
	  
	  // directory use my_word[count]
	  nb += vec[count] + " ";
	  
	  for (int len = len0 + 1; len < (int)src.size(); ++len) {
	    nb += src[len] + " ";
	  }
	
	  neighbors->push_back(bitext[0] + "\t" + nb);
	}
      }

    } else if (neighbor_type == "co-occur") {
      // use co-occur words to generate sentences.
      // Hope this one will solve the fertility issue.

      assert(neighbors->size() == 1);
      
      unordered_set<string> my_candidates;
      for (int i = 0; i < (int)trg.size(); ++i) {
	const string& t = trg[i];
	unordered_set<string>& m = all_candidates[t];
	for (unordered_set<string>::const_iterator it = m.begin(); it != m.end(); ++it) {
	  const string& tmp = *it;
	  if (Global::FLAGS_neighbor_with_only_new_words) {
	    // do not use words that are in the current translation
	    if (set_src.find(tmp) == set_src.end()) {  
	      my_candidates.insert(tmp);
	    }
	  } else {
	    my_candidates.insert(tmp);
	  }
	}
      }

      cerr << all_candidates.size() << endl;
      cerr << my_candidates.size() << endl;

      if (Global::FLAGS_neighbor_with_only_new_words) {
	assert(my_candidates.size() >= set_src.size());
      }
      if (my_candidates.size() < set_src.size()) {
	cerr << "no neighborhood!" << endl;
	return;
      }

      vector<int> candidate_count;
      vector<string> candidate_word;
      for (unordered_set<string>::const_iterator it = my_candidates.begin(); it != my_candidates.end(); ++it ) {
	const string& str = *it;
	int freq = all_candicounts[str];
	candidate_count.push_back(freq);
	candidate_word.push_back(str);
      }
       
      assert(candidate_count.size() == candidate_word.size());

      for (int count = 0; count < Global::FLAGS_neighbor_type_co_occur_count; ++count) {
	string nb = "";
	vector<int> my_count;
	vector<string> my_word;
	for (int i = 0; i < (int)candidate_count.size(); ++i) {
	  my_count.push_back(candidate_count[i]);
	  my_word.push_back(candidate_word[i]);
	}
	for (int len = 0; len < (int)src.size(); ++len) {
	  int total = 0;
	  for (int i = 0; i < (int)my_count.size(); ++i) {
	    total += my_count[i];
	  }
	  // generate a random integer [1, total]
	  // find the word and use it to generate the neighbor
	  int r = (rand() % total) + 1;
	  int i = 0;
	  int rr = 0;  // compare r and rr
	  for (; i < (int)my_count.size(); ++i) {
	    /*
	    if (my_count[i] == 0) {  // already used.
	      continue;
	    }
	    */
	    rr += my_count[i];
	    if (rr >= r) {
	      nb += my_word[i] + " ";
	      break;
	    }
	  }
	  assert(i < (int)my_count.size());
	  // remove from my_count and my_word
	  my_count.erase(my_count.begin() + i);
	  my_word.erase(my_word.begin() + i);
	  assert(my_count.size() == my_word.size());
	}
	neighbors->push_back(bitext[0] + "\t" + nb);
      }

    } else if (neighbor_type == "len") {
      // everything in the training data with the same length

      ifstream in(file.c_str());
      if (!in.good()) {
	cerr << "Can not open " << file << endl;
	return;
      }

      string line2;

      while (UTIL::getline(in, &line2)) {
	vector<string> v;
	UTIL::split_string_using(line2, "\t", &v);
	if (v.size() != 2) {
	  continue;
	}

	if (line2 == line) {
	  continue;
	}

	vector<string> strs1, strs2;
	UTIL::split_string_using(v[1], " ", &strs1);
	UTIL::split_string_using(v[0], " ", &strs2);
	if (strs1.size() == src.size()) {
	  neighbors->push_back(bitext[0] + "\t" + v[1]);
	}
	if (strs2.size() == trg.size()) {
	  neighbors->push_back(v[0] + "\t" + bitext[1]);
	}
      }  // while
      in.close();

    } else if ((neighbor_type == "rep-1") || (neighbor_type == "replace-1")) {
      // replace a word with another word in the same sentence

      // replace one target word
      if (trg.size() > 1) {
	for (int i = 0; i < (int)trg.size(); ++i) {  // the word at i will be replaced
	  for (int ii = 0; ii < (int)trg.size(); ++ii) {  // use word ii to replace word i
	    if (ii == i) {
	      continue;
	    }
	    string str = "";
	    for (int j = 0; j < (int)trg.size(); ++j) {
	      if (j != i) {
		str += trg[j] + " ";
	      } else {
		str += trg[ii] + " ";
	      }
	    }
	    str += "\t" + bitext[1];
	    neighbors->push_back(str);
	  }
	}
      }
      // replace one source word
      if (src.size() > 1) {
	for (int i = 0; i < (int)src.size(); ++i) {  // the word at i will be replaced
	  for (int ii = 0; ii < (int)src.size(); ++ii) {  // use word ii to replace word i
	    if (ii == i) {
	      continue;
	    }
	    string str = "";
	    str += bitext[0] + "\t";
	    for (int j = 0; j < (int)src.size(); ++j) {
	      if (j != i) {
		str += src[j] + " ";
	      } else {
		str += src[ii] + " ";
	      }
	    }
	    neighbors->push_back(str);
	  }
	}
      }
      /*
    } else if (neighbor_type == "add-1") {
    } else if (neighbor_type == "len-2") {
      // neighbor lenth is 2
      
      for (std::map<string, vector<string> >::const_iterator it1 = m.begin(); it1 != m.end(); ++it1) {
      // if (step % 10 == 0) {
      // cerr << "working on " << step << " out of " << STEP << "\t" << "(" << count << ")" << endl;
      // }
      const string& str1 = it1->first;
      const vector<string>& v1 = it1->second;
      for (std::map<string, vector<string> >::const_iterator it2 = m.begin(); it2 != m.end(); ++it2) {
      const string& str2 = it2->first;
      const vector<string>& v2 = it2->second;
      std::set<string> s;
      for (int i = 0; i < (int)v1.size(); ++i) {
      s.insert(v1[i]);
      }
      for (int i = 0; i < (int)v2.size(); ++i) {
      s.insert(v2[i]);
      }
      for (std::set<string>::const_iterator it3 = s.begin(); it3!= s.end(); ++it3) {
      const string& str3 = *it3;
      for (std::set<string>::const_iterator it4 = s.begin(); it4!= s.end(); ++it4) {
      const string& str4 = *it4;
	      
      out << str1 << " "
      << str2 << "\t"
      << str3 << " "
      << str4 << endl;
      ++count;
      }
      }
      }
      }
      // next data point
    } else if (neighbor_type == "len-1") {
      // neighbor length is 1
      for (std::map<string, vector<string> >::const_iterator it1 = m.begin(); it1 != m.end(); ++it1) {
      // if (step % 10 == 0) {
      // cerr << "working on " << step << " out of " << STEP << "\t" << "(" << count << ")" << endl;
      // }
      const string& str1 = it1->first;
      const vector<string>& v1 = it1->second;
	
      std::set<string> s;
      for (int i = 0; i < (int)v1.size(); ++i) {
      s.insert(v1[i]);
      }
	
      for (std::set<string>::const_iterator it3 = s.begin(); it3!= s.end(); ++it3) {
      const string& str3 = *it3;
	  
      out << str1 << "\t"
      << str3 << endl;
      ++count;
      }
      }
      // next data point
      */
    } else {
      cerr << "Error: wrong type of neighbor: " << neighbor_type << endl;
      exit(-1);
    }  // neighbor_type
  }
}  // namespace FERT2
