
#include <cassert>

#include "util/file-util.h"
#include "util/string-util.h"

#include "config/global.h"

namespace MYCONFIG {

  ///////////////////////////// variables ///////////////////////////////

  string my_global::FLAGS_option_ = "x";
  int my_global::FLAGS_loop_ = -1;
  double my_global::FLAGS_p0_ = -1;
  double my_global::FLAGS_samples_ = -1;
  bool my_global::FLAGS_hadoop_ = false;

  ///////////////////
  // config.ini

  string my_global::FLAGS_LL_file = "";

  double my_global::FLAGS_alpha = -1;
  double my_global::FLAGS_sigma = -1;

  string my_global::FLAGS_bin_dir = "";

  bool my_global::FLAGS_bound = true;

  bool my_global::FLAGS_output_prob = false;

  bool my_global::FLAGS_parler = true;

  int my_global::FLAGS_burn_in_eval = 10000;
  int my_global::FLAGS_burn_in_training = 100;
  int my_global::FLAGS_burn_in_training_max = 10000;

  bool my_global::FLAGS_contrastive = false;
  int my_global::FLAGS_constraints = 100;

  bool my_global::FLAGS_data_format_reverse = false;
  bool my_global::FLAGS_combine_jump = true;
  bool my_global::FLAGS_scale = false;
  double my_global::FLAGS_scale_value = 1;
  bool my_global::FLAGS_double_factorial = true;
  string my_global::FLAGS_distortion = "";

  bool my_global::FLAGS_debug = true;
  bool my_global::FLAGS_debug_output_counts = false;
  bool my_global::FLAGS_debug_output_details = false;
  bool my_global::FLAGS_debug_output_max = false;
  bool my_global::FLAGS_debug_save_space = true;
  bool my_global::FLAGS_debug_verbose = false;

  int my_global::FLAGS_number_of_empty_words = 0;
  string my_global::FLAGS_empty_word = "_empty_word_";
  string my_global::FLAGS_non_empty_word = "_non_empty_word_";

  string my_global::FLAGS_BOS = "<s>";
  string my_global::FLAGS_EOS = "</s>";

  int my_global::FLAGS_freq_word_count = 1;
  string my_global::FLAGS_wa_model = "ibm1";  // ibm1, prior or poisson

  string my_global::FLAGS_unknown_word = "_UNK_";
  int my_global::FLAGS_unknown_word_count = 3;

  double my_global::FLAGS_eta = 1.0e-8;

  bool my_global::FLAGS_enumerate = false;
  bool my_global::FLAGS_viterbi = true;

  int my_global::FLAGS_max_iter = 1000;

  int my_global::FLAGS_min_freq = 1;
  string my_global::FLAGS_file_prefix = "";
  bool my_global::FLAGS_fert1 = true;
  bool my_global::FLAGS_fert3 = true;

  string my_global::FLAGS_hadoop_home = "";

  bool my_global::FLAGS_ibm1_simulation = true;

  bool my_global::FLAGS_ignore_phi_factor = true;
  bool my_global::FLAGS_ignore_psi_factor = true;
  bool my_global::FLAGS_ignore_fert_prior = true;

  double my_global::FLAGS_init_lex = 0;
  double my_global::FLAGS_init_fert = 0;

  bool my_global::FLAGS_lambda = false;
  bool my_global::FLAGS_lambda_fixed = false;
  int my_global::FLAGS_lbfgs_loop = 20;
  bool my_global::FLAGS_log = false;
  bool my_global::FLAGS_local_co_occur = true;

  string my_global::FLAGS_model_file = "";
  int my_global::FLAGS_max_src_fert = 1;
  int my_global::FLAGS_max_trg_fert = -1;

  model_t my_global::FLAGS_model = IBM1_EM_LBFGS;

  string my_global::FLAGS_mapper = "";
  string my_global::FLAGS_reducer = "";
  

  string my_global::FLAGS_neighbor_type = "";
  bool my_global::FLAGS_neighbor_with_only_new_words = false;
  int my_global::FLAGS_neighbor_type_change_1_in_src_count = 10;
  int my_global::FLAGS_neighbor_type_nextk_count = -1;
  int my_global::FLAGS_neighbor_type_co_occur_count = -1;

  string my_global::FLAGS_project_id = "";

  string my_global::FLAGS_sampling_type = "quick";

  int my_global::FLAGS_re_start = -1;  // 10;
  int my_global::FLAGS_re_sample = -1;  // 1000;

  // string my_global::FLAGS_train_file = "";
  
  string my_global::FLAGS_log_file = "aer.log";

  string my_global::FLAGS_test_file = "";
  string my_global::FLAGS_a_file = "";
  string my_global::FLAGS_current_parameter = "";

  bool my_global::FLAGS_training_only = false;

  bool my_global::FLAGS_uniform = true;

  double my_global::FLAGS_weight_empty = 0;  // 2;
  double my_global::FLAGS_weight_other = 0;  // 1;

  bool my_global::FLAGS_uniform_jump = false;

  //////////////////////////
  // other global variables
  int my_global::t_co_occur_sum = 0;  // useful?
  int my_global::bitext_sentence_size_= 0;  // useful?

  int my_global::lex_size_ = 0;
  int my_global::trg_size_ = 0;

  unordered_map<int, pair<string, string> > my_global::lex_id2pair_;
  unordered_map<pair<string, string>, int> my_global::lex_pair2id_;
  unordered_map<int, string> my_global::slack_id2str_;
  unordered_map<string, int> my_global::slack_str2id_;

  unordered_map<string, int> my_global::t_freq_;  // t only

  // internal use
  unordered_map<string, unordered_map<string, int> > my_global::st_freq__;
  unordered_map<string, int> my_global::s_freq__;  // s only

  ///////////////////////////// functions ///////////////////////////////
  // return number of lines
  int my_global::read_co_occur_tables(istream& in) {
				      
    static bool run_once = false;
    if (run_once) {
      cerr << "you can only call me once: read_co_occur_tables" << endl;
      exit(-1);
    }
    run_once = true;

    string line;
    int line_count = 0;
    while (UTIL::getline(in, &line)) {
      if (line_count % 50000 == 0) {
	cerr << "Processing " << line_count << endl;
      }
      ++line_count;

      vector<string> v;
      UTIL::split_string_using(line, "\t", &v);

      if (v.size() != 2) {
	// cerr << "Skip: " << line << endl;
	cerr << "Skip... " << endl;
	cerr << "Line: " << line_count << endl;
	cerr << "Size: " << v.size() << endl;
	continue;
      }
    
      // change trg/v[0] to lower case
      UTIL::to_lower(&v[0]);

      vector<string> v_trg00, v_src00;
      if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	// reverse
	UTIL::split_string_using(v[1], " ", &v_trg00);
	UTIL::split_string_using(v[0], " ", &v_src00);
      } else {
	// normal
	UTIL::split_string_using(v[0], " ", &v_trg00);
	UTIL::split_string_using(v[1], " ", &v_src00);
      }

      if (my_global::FLAGS_wa_model == "hmm") {
	vector<string> tmp_src;
	tmp_src.push_back(MYCONFIG::my_global::FLAGS_BOS);
	for (int i = 0; i < (int)v_src00.size(); ++i) {
	  tmp_src.push_back(v_src00[i]);
	}
	tmp_src.push_back(MYCONFIG::my_global::FLAGS_EOS);
	const unsigned int size_src = tmp_src.size();
	assert(v_src00.size() + 2 == size_src);
	v_src00.clear();
	v_src00 = tmp_src;
	assert(v_src00.size() == size_src);

	vector<string> tmp_trg;
	tmp_trg.push_back(MYCONFIG::my_global::FLAGS_BOS);
	for (int i = 0; i < (int)v_trg00.size(); ++i) {
	  tmp_trg.push_back(v_trg00[i]);
	}
	tmp_trg.push_back(MYCONFIG::my_global::FLAGS_EOS);
	const unsigned int size_trg = tmp_trg.size();
	assert(v_trg00.size() + 2 == size_trg);
	v_trg00.clear();
	v_trg00 = tmp_trg;
	assert(v_trg00.size() == size_trg);
      } else {
	if (my_global::FLAGS_number_of_empty_words == 0) {
	} else if (my_global::FLAGS_number_of_empty_words == 1) {
	  v_trg00.push_back(my_global::FLAGS_empty_word);
	} else {
	  assert(my_global::FLAGS_number_of_empty_words == 2);
	  int size = v_trg00.size();
	  for (int i = 0; i < size; ++i) {
	    v_trg00.push_back(my_global::FLAGS_empty_word);
	  }
	}
      }

      // fill t_co_occur
      for (vector<string>::const_iterator it_t = v_trg00.begin(); it_t != v_trg00.end(); ++it_t) {
	const string& t = *it_t;
	
	t_freq_[t] += 1;
	t_co_occur_sum += 1;
      }
      
      // fill s_co_occur
      for (vector<string>::const_iterator it_s = v_src00.begin(); it_s != v_src00.end(); ++it_s) {
	const string& s = *it_s;
	
	s_freq__[s] += 1;
      }
	
      ++bitext_sentence_size_;

      // fill co_occur
      for (int i = 0; i  < (int)v_src00.size(); ++i) {
	
	const string& s = v_src00[i];
	unordered_map<string, int>& tmp1 = st_freq__[s];

	for (vector<string>::const_iterator it_t = v_trg00.begin(); it_t != v_trg00.end(); ++it_t) {
	  const string& t = *it_t;
	  tmp1[t] += 1;
	}
      }
    }  // line

    int count_index;
    
    // fill lex_id2pair_ and lex_pair2id_ from co_occur
    count_index = 0;
    for (unordered_map<string, unordered_map<string, int> >::const_iterator it_s = st_freq__.begin(); it_s != st_freq__.end(); ++it_s) {
      
      const string& src = it_s->first;
      const unordered_map<string, int>& map_trg = it_s->second;
      
      for (unordered_map<string, int>::const_iterator it_t = map_trg.begin(); it_t != map_trg.end(); ++it_t) {
	const string& trg = it_t->first;
	
	lex_pair2id_[make_pair(trg, src)] = count_index;
	lex_id2pair_[count_index] = make_pair(trg, src);
	++count_index;
      }
    }
      
    lex_size_ = count_index;
    
    // fill slack_id2str_ and slack_str2id from t_co_occur
    count_index = 0;
    for (unordered_map<string, int>::const_iterator it = t_freq_.begin(); it != t_freq_.end(); ++it) {
      const string& trg = it->first;
      
      slack_str2id_[trg] = count_index;
      slack_id2str_[count_index] = trg;
      ++count_index;
    }
    
    trg_size_ = count_index;
    
    return line_count;
  }
  
  // 
  // NOTE:
  // we all this in lbfgs-main using local files, and call this in delta-mapper in hadoop using hdfs file 
  //
  int my_global::read_co_occur(const string& train_file,
			     //
			     ConfigFile* config_default,
			    ConfigFile* config_site) {

    // read co-occur table so we can setup x for lbfgs.
    cerr << "[global]" << endl;
    cerr << "[global] train-file: " << train_file << endl;
    cerr << "[global]" << endl;

    // In hadoop, it is different from my_global::FLAGS_train_file.c_str());
    ifstream in(train_file.c_str());  
    int number_of_lines = read_co_occur_tables(in);
    in.close();

    if (my_global::st_freq__.size() == 0) {
      cerr << "can not read " << train_file << endl;
      exit(-1);
    }

    // model size
    cerr << "bitext size " << lex_size_ << endl;
    cerr << "target size " << trg_size_ << endl;
    cerr << "target size " << t_freq_.size() << endl;
    cerr << "source size " << s_freq__.size() << endl;
    
    return number_of_lines;
  }

  int my_global::str2id_old(const string& type, const string& trg, const string& src) {
    
    int id = 0;

    if (type == "size") {
      // lex, dual and slack variables
      id = 2 * lex_size_ + trg_size_;
    }else if (type == "lex") {
      assert(trg != "");
      assert(src != "");

      id = lex_pair2id_[make_pair(trg, src)];
      
      if (my_global::FLAGS_debug) {
	string str = id2str(id);
	assert(str == type + " " + trg + " " + src + " true");
      }
    } else if (type == "slack") {  // slack variables sum_e p(f|e) = 1
      assert(trg != "");
      assert(src != "");

      id = lex_size_ + lex_pair2id_[make_pair(trg, src)];

      if (my_global::FLAGS_debug) {
	string str = id2str(id);
	assert(str == type + " " + trg + " " + src + " true");
      }
    } else if (type == "dual") {  // dual variables for lex
      assert(trg != "");
      assert(src == "");

      id = 2 * lex_size_ + slack_str2id_[trg];

      if (my_global::FLAGS_debug) {
	string str = id2str(id);
	assert(str == type + " " + trg);
      }
    } else {
      assert(false);
    }
  
    return id;
  }

  int my_global::str2id(const string& type, const string& trg, const string& src) {
    
    int id = 0;

    if (type == "size") {
      // lex, dual and slack variables
      // id = lex_size_ + trg_size_;
      id = lex_size_;
    }else if (type == "lex") {
      assert(trg != "");
      assert(src != "");

      id = lex_pair2id_[make_pair(trg, src)];
      
      if (my_global::FLAGS_debug) {
	string str = id2str(id);
	assert(str == type + " " + trg + " " + src + " true");
      }
    } else {
      assert(false);
    }
  
    return id;
  }

  string my_global::id2str(int id) {
    
    string str = "";

    if ((id >= 0) && (id < lex_size_)) {
      const pair<string, string>& p = lex_id2pair_[id];
      str = "lex " + p.first + " " + p.second + " true";

      if (my_global::FLAGS_debug) {
	// int id2 = str2id("lex", p.first, p.second);
	// assert(id2 == id);
      }
    } else {
      assert(false);
    }

    return str;
  }

  string my_global::id2str_old(int id) {
    
    string str = "";

    if ((id >= 0) && (id < lex_size_)) {
      const pair<string, string>& p = lex_id2pair_[id];
      str = "lex " + p.first + " " + p.second + " true";

      if (my_global::FLAGS_debug) {
	// int id2 = str2id("lex", p.first, p.second);
	// assert(id2 == id);
      }
    } else if ((id >= lex_size_) && (id < 2 * lex_size_)) {
      id -= lex_size_;
      const pair<string, string>& p = lex_id2pair_[id];
      str = "slack " + p.first + " " + p.second + " true";

      if (my_global::FLAGS_debug) {
	// int id2 = str2id("dual", p.first, p.second);
	// assert(id2 == id);
      }
    } else if ((id >= 2 * lex_size_) && (id < 2 * lex_size_ + trg_size_)) {
      id -= 2 * lex_size_;
      const string str2 = slack_id2str_[id];
      str = "dual " + str2;

      if (my_global::FLAGS_debug) {
	// int id2 = str2id("slack", str2, "");
	// assert(id2 == id);
      }
    } else {
      assert(false);
    }

    return str;
  }

  void my_global::read_config(ConfigFile* config) {

    // read config.ini

    if (config->keyExists("log")) {
      my_global::FLAGS_log = config->read<bool>("log");
    }

    if (config->keyExists("uniform-jump")) {
      my_global::FLAGS_uniform_jump = config->read<bool>("uniform-jump");
    }

    if (config->keyExists("output-prob")) {
      my_global::FLAGS_output_prob = config->read<bool>("output-prob");
    }


    if (config->keyExists("lambda-fixed")) {
      my_global::FLAGS_lambda_fixed = config->read<bool>("lambda-fixed");
    }

    if (config->keyExists("scale")) {
      my_global::FLAGS_scale = config->read<bool>("scale");
    }

    if (config->keyExists("data-format-reverse")) {
      my_global::FLAGS_data_format_reverse = config->read<bool>("data-format-reverse");
    }

    if (config->keyExists("combine-jump")) {
      my_global::FLAGS_combine_jump = config->read<bool>("combine-jump");
    }

    if (config->keyExists("number-of-empty-words")) {
      my_global::FLAGS_number_of_empty_words = config->read<int>("number-of-empty-words");
    }

    if (config->keyExists("bound")) {
      my_global::FLAGS_bound = config->read<bool>("bound");
    }

    if (config->keyExists("hadoop-home")) {
      config->readInto(my_global::FLAGS_hadoop_home, "hadoop-home", string(""));
    }

    /*
    if (config->keyExists("train-file")) {
      config->readInto(my_global::FLAGS_train_file, "train-file", string(""));
    }
    */

    if (config->keyExists("distortion")) {
      config->readInto(my_global::FLAGS_distortion, "distortion", string("uniform"));
      if (my_global::FLAGS_distortion == "uniform") {
      } else if (my_global::FLAGS_distortion == "phi") {
      } else if (my_global::FLAGS_distortion == "empty-word") {
      } else {
	assert(false);
      }
    }

    if (config->keyExists("wa-model")) {
      config->readInto(my_global::FLAGS_wa_model, "wa-model", string("ibm1"));
      /*
      if (my_global::FLAGS_wa_model == "ibm1") {
      } else if (my_global::FLAGS_wa_model == "upoisson") {
      } else if (my_global::FLAGS_wa_model == "prior") {
      } else if (my_global::FLAGS_wa_model == "poisson") {
      } else if (my_global::FLAGS_wa_model == "hmm") {
      } else if (my_global::FLAGS_wa_model == "hmmf") {
      */
      if (my_global::FLAGS_wa_model == "ibm1") {
      } else if (my_global::FLAGS_wa_model == "hmm") {
      } else if (my_global::FLAGS_wa_model == "hmmf") {
      } else {
	assert(false);
      }
    }

    if (config->keyExists("sampling-type")) {
      config->readInto(my_global::FLAGS_sampling_type, "sampling-type", string("quick"));
    }

    if (config->keyExists("log-file")) {
      config->readInto(my_global::FLAGS_log_file, "log-file", string(""));
    }

    if (config->keyExists("test-file")) {
      config->readInto(my_global::FLAGS_test_file, "test-file", string(""));
    }

    if (config->keyExists("a-file")) {
      config->readInto(my_global::FLAGS_a_file, "a-file", string(""));
    }

    if (config->keyExists("current-parameter")) {
      config->readInto(my_global::FLAGS_current_parameter, "current-parameter", string(""));
    }

    if (config->keyExists("neighbor-type")) {
      config->readInto(my_global::FLAGS_neighbor_type, "neighbor-type", string(""));
    }

    if (config->keyExists("neighbor-type-nextk-count")) {
      if (my_global::FLAGS_neighbor_type == "neighbor-type-nextk-src") {
	if (config->keyExists("nextk-count")) {
	  my_global::FLAGS_neighbor_type_nextk_count = config->read<int>("neighbor-type-nextk-count");
	}
      }
    }

    if (config->keyExists("neighbor-with-only-new-words")) {
      my_global::FLAGS_neighbor_with_only_new_words = config->read<bool>("neighbor-with-only-new-words");
      if((my_global::FLAGS_neighbor_type != "neighbor-type-co-occur") &&
	 (my_global::FLAGS_neighbor_type != "neighbor-type-change-1-in-src")) {
	// cerr << "[global]: why do you define neighbor-type-only-new-words-in-nb while your nb is neither neighbor-type-change-1-in-src nor neighbor-type-co-occur?" << endl;
      }
    }

    if (config->keyExists("neighbor-type-change-1-in-src-count")) {
      my_global::FLAGS_neighbor_type_change_1_in_src_count = config->read<int>("neighbor-type-change-1-in-src-count");
      if (my_global::FLAGS_neighbor_type != "neighbor-type-change-1-in-src") {
	// cerr << "[global]: why do you define neighbor-type-change-1-in-src-count while your nb is not neighbor-type-change-1-in-src?" << endl;
      }
    }

    if (config->keyExists("neighbor-type-co-occur-count")) {
      if(my_global::FLAGS_neighbor_type != "neighbor-type-co-occur") {
	// cerr << "[global]: why do you define neighbor-type-co-occur_count while your nb is not neighbor-type-co-occur?" << endl;
      }
      my_global::FLAGS_neighbor_type_co_occur_count = config->read<int>("neighbor-type-co-occur-count");
    }

    if (config->keyExists("bin-dir")) {
      config->readInto(my_global::FLAGS_bin_dir, "bin-dir", string(""));
    }

    if (config->keyExists("freq-word-count")) {
      my_global::FLAGS_freq_word_count = config->read<int>("freq-word-count");
    }

    if (config->keyExists("unknown-word-count")) {
      my_global::FLAGS_unknown_word_count = config->read<int>("unknown-word-count");
    }

    if (config->keyExists("max-iter")) {
      my_global::FLAGS_max_iter = config->read<int>("max-iter");
    }

    if (config->keyExists("re-start")) {
      my_global::FLAGS_re_start = config->read<int>("re-start");
    }
    if (config->keyExists("re-sample")) {
      my_global::FLAGS_re_sample = config->read<int>("re-sample");
    }

    if (config->keyExists("burn-in-eval")) {
      my_global::FLAGS_burn_in_eval = config->read<int>("burn-ineval-");
    }
    if (config->keyExists("burn-in-training")) {
      my_global::FLAGS_burn_in_training = config->read<int>("burn-in-training");
    }
    if (config->keyExists("burn-in-training-max")) {
      my_global::FLAGS_burn_in_training_max = config->read<int>("burn-in-training-max");
    }

    if (config->keyExists("viterbi")) {
      my_global::FLAGS_viterbi = config->read<bool>("viterbi");
    }
    if (config->keyExists("enumerate")) {
      my_global::FLAGS_enumerate = config->read<bool>("enumerate");
    }
    if (my_global::FLAGS_enumerate) {
      assert(my_global::FLAGS_re_start == 1);
      assert(my_global::FLAGS_re_sample == 1);
    } else {
      // if not enumerate, we have to do viterbi
      assert(my_global::FLAGS_re_start > 0);
      assert(my_global::FLAGS_re_sample > 0);
      assert(my_global::FLAGS_viterbi);
    }

    if (config->keyExists("debug-verbose")) {
      my_global::FLAGS_debug_verbose = config->read<bool>("debug-verbose");
    }
    if (config->keyExists("debug")) {
      my_global::FLAGS_debug = config->read<bool>("debug");
    }
    if (config->keyExists("training-only")) {
      my_global::FLAGS_training_only = config->read<bool>("training-only");
    }

    if (config->keyExists("mapper")) {
      config->readInto(my_global::FLAGS_mapper, "mapper", string(""));
    }
    if (config->keyExists("reducer")) {
      config->readInto(my_global::FLAGS_reducer, "reducer", string(""));
    }

    if (config->keyExists("scale-value")) {
      my_global::FLAGS_scale_value = config->read<double>("scale-value");
    }


    if (config->keyExists("alpha")) {
      my_global::FLAGS_alpha = config->read<double>("alpha");
    }
    if (config->keyExists("sigma")) {
      my_global::FLAGS_sigma = config->read<double>("sigma");
    }

    if (config->keyExists("eta")) {
      my_global::FLAGS_eta = config->read<double>("eta");
    }

    if (config->keyExists("init-lex")) {
      my_global::FLAGS_init_lex = config->read<double>("init-lex");
    }
    if (config->keyExists("init-fert")) {
      my_global::FLAGS_init_fert = config->read<double>("init-fert");
    }

    if (config->keyExists("LL-file")) {
      config->readInto(my_global::FLAGS_LL_file, "LL-file", string(""));
    }
    assert(my_global::FLAGS_LL_file != "");

    if (config->keyExists("debug-save-space")) {
      my_global::FLAGS_debug_save_space = config->read<bool>("debug-save-space");
    }

    // ignore these factors?
    if (config->keyExists("ignore-phi-factor")) {
      my_global::FLAGS_ignore_phi_factor = config->read<bool>("ignore-phi-factor");
    }
    if (config->keyExists("ignore-psi-factor")) {
      my_global::FLAGS_ignore_psi_factor = config->read<bool>("ignore-psi-factor");
    }
    if (config->keyExists("ignore-fert-prior")) {
      my_global::FLAGS_ignore_fert_prior = config->read<bool>("ignore-fert-prior");
    }


    if (config->keyExists("debug-output-counts")) {
      my_global::FLAGS_debug_output_counts = config->read<bool>("debug-output-counts");
    }
    if (config->keyExists("debug-output-details")) {
      my_global::FLAGS_debug_output_details = config->read<bool>("debug-output-details");
    }
    if (config->keyExists("debug-output-max")) {
      my_global::FLAGS_debug_output_max = config->read<bool>("debug-output-max");
    }

    // my_global::FLAGS_lbfgs_loop = config->read<int>("lbfgs-loop");
    if (config->keyExists("file-prefix")) {
      config->readInto(my_global::FLAGS_file_prefix, "file-prefix", string(""));
    }
    assert(my_global::FLAGS_file_prefix != "");

    if (config->keyExists("model-file")) {
      config->readInto(my_global::FLAGS_model_file, "model-file", string(""));
    }
    assert(my_global::FLAGS_file_prefix != "");

    if (config->keyExists("project-id")) {
      config->readInto(my_global::FLAGS_project_id, "project-id", string(""));
    }
    assert(my_global::FLAGS_project_id != "");

    if (config->keyExists("ibm1-simulation")) {
      my_global::FLAGS_ibm1_simulation = config->read<bool>("ibm1-simulation");
    }

    if (config->keyExists("weight-empty")) {
      my_global::FLAGS_weight_empty = config->read<double>("weight-empty");
    }
    if (config->keyExists("weight-other")) {
      my_global::FLAGS_weight_other = config->read<double>("weight-other");
    }

    if (config->keyExists("lbfgs-loop")) {
      my_global::FLAGS_lbfgs_loop = config->read<int>("lbfgs-loop");
    }

    if (config->keyExists("min-freq")) {
      my_global::FLAGS_min_freq = config->read<int>("min-freq");
    }

    if (config->keyExists("contrastive")) {
      my_global::FLAGS_contrastive = config->read<bool>("contrastive");
    }

    if (config->keyExists("constraints")) {
      my_global::FLAGS_constraints = config->read<int>("constraints");
    }

    // done with config
  }

  void my_global::config(int argc, char* argv[]) {
    if (argc != 3) {
      cerr << "usage: " << argv[0] << " <config-default.ini> <config-site.ini>" << endl;
      exit(-1);
    }
    assert(argc == 3);
    const string& config_file_default = argv[1];
    const string& config_file_site = argv[2];
    
    ConfigFile config_default(config_file_default);
    ConfigFile config_site(config_file_site);
    my_global::read_config(&config_default);
    my_global::read_config(&config_site);
  }

}  // namespace
