
#ifndef _global_h_
#define _global_h_

#include <string>
#include <map>

#include "config/config.h"
#include "util/unordered-util.h"

namespace MYCONFIG {

  enum model_t {IBM1_EM_CLASSIC,  // the classic one in Brown 90
		IBM1_EM_LBFGS,  // still use the em formula, but use lbfgs to optimize
		IBM1_CO_OCCUR_LBFGS,  // consider all co-occur, and use lbfgs    
		UNKNOWN_MODEL};
    
  class my_global {
  public:

    ///////////////////
    // config.ini

    /* 
       reading from command line
    */
    static string FLAGS_option_;
    static int FLAGS_loop_;
    static double FLAGS_p0_;
    static double FLAGS_samples_;
    static bool FLAGS_hadoop_;

    /*
      global flags from config file
    */
    static string FLAGS_LL_file;

    static double FLAGS_alpha;
    static double FLAGS_sigma;

    static string FLAGS_bin_dir;

    static bool FLAGS_bound;  // bound for probability. true: 1, false: infinite

    static int FLAGS_burn_in_eval;
    static int FLAGS_burn_in_training;
    static int FLAGS_burn_in_training_max;

    static bool FLAGS_contrastive;
    static int FLAGS_constraints;

    static bool FLAGS_data_format_reverse;
    static bool FLAGS_combine_jump;
    static bool FLAGS_scale;
    static double FLAGS_scale_value;
    static bool FLAGS_double_factorial;
    static string FLAGS_distortion;

    static bool FLAGS_parler;

    static bool FLAGS_output_prob;
    
    static bool FLAGS_debug;
    static bool FLAGS_debug_output_counts;
    static bool FLAGS_debug_output_details;
    static bool FLAGS_debug_output_max;
    static bool FLAGS_debug_save_space;
    static bool FLAGS_debug_verbose;  // output extra stuff in lbfgs*.m1

    static int FLAGS_number_of_empty_words;
    static string FLAGS_empty_word;
    static string FLAGS_non_empty_word;

    static string FLAGS_BOS;
    static string FLAGS_EOS;

    static int FLAGS_freq_word_count;

    static string FLAGS_wa_model;  // ibm1, prior or poisson

    static string FLAGS_unknown_word;
    static int FLAGS_unknown_word_count;

    static double FLAGS_eta;

    static bool FLAGS_enumerate;
    static bool FLAGS_viterbi;

    static int FLAGS_min_freq;
    static string FLAGS_file_prefix;
    static bool FLAGS_fert1;
    static bool FLAGS_fert3;

    static string FLAGS_hadoop_home;

    static bool FLAGS_ibm1_simulation;  // a special case

    static bool FLAGS_ignore_phi_factor;
    static bool FLAGS_ignore_psi_factor;
    static bool FLAGS_ignore_fert_prior;

    static double FLAGS_init_lex;
    static double FLAGS_init_fert;

    static bool FLAGS_lambda;
    static bool FLAGS_lambda_fixed;
    static int FLAGS_lbfgs_loop;
    static bool FLAGS_log;
    // we only consider words that co-occured with each other
    // the ibm 1 is local co-occur model
    // but in general, the discrimitive models are not
    // and are computationally more expensive.
    static bool FLAGS_local_co_occur;

    static model_t FLAGS_model;
    static string FLAGS_model_file;
    static int FLAGS_max_iter;
    static int FLAGS_max_src_fert;
    static int FLAGS_max_trg_fert;

    static string FLAGS_mapper;
    static string FLAGS_reducer;

    static string FLAGS_neighbor_type;
    static bool FLAGS_neighbor_with_only_new_words;
    static int FLAGS_neighbor_type_change_1_in_src_count;
    static int FLAGS_neighbor_type_nextk_count;  // specify how many next-2k neighbors
    static int FLAGS_neighbor_type_co_occur_count;  // Specify how many co-occur neighbors.  

    static string FLAGS_project_id;

    static string FLAGS_sampling_type;
    static int FLAGS_re_start;
    static int FLAGS_re_sample;

    // static string FLAGS_train_file;
    static string FLAGS_current_parameter;
    static string FLAGS_test_file;
    static string FLAGS_a_file;

    static bool FLAGS_training_only;  // no decoding and evaluation

    static string FLAGS_log_file;

    static bool FLAGS_uniform;
    static double FLAGS_weight_empty;
    static double FLAGS_weight_other;
    static bool FLAGS_uniform_jump;

    ///////////////////
    // other global variabls

    // mapping between id (used by lbfgs) and string (used by *_Factor's)
    static int str2id_old(const string& type, const string& trg, const string& src);
    static string id2str_old(int id);

    static int str2id(const string& type, const string& trg, const string& src);
    static string id2str(int id);

    static unordered_map<int, pair<string, string> > lex_id2pair_;
    static unordered_map<pair<string, string>, int> lex_pair2id_;
    static unordered_map<int, string> slack_id2str_;
    static unordered_map<string, int> slack_str2id_;

    static int bitext_sentence_size_;
    static int t_co_occur_sum;  // t only for p(e)

    // first key: source (f)
    // second key: target (e)
    // first value: count
    // second value: index
    //
    // co_occur_[f][e].first: count
    // co_occur_[f][e].secon: index for x[]
    static unordered_map<string, unordered_map<string, int> > st_freq__;  // freq of st
    static unordered_map<string, int> s_freq__;  // freq of s
    static unordered_map<string, int> t_freq_;  // freq of t

    static int lex_size_;  // based on st_freq__
    static int trg_size_;  // based on t_freq_

  public:
    static void config(int argc, char* argv[]);
    static void read_config(ConfigFile* config);
    static int read_co_occur_tables(istream& in);
				     
    static int read_co_occur(const string& hdfs_bitext_file, 
			      //
			     ConfigFile* config_default,
			     ConfigFile* config_site);

  };
}

#endif
