
#include "format2giza.h"


void doit(const vector<string>& bitext, int pre_index, const string& reverse, 
	  //
	  map<int, set<int> >* m) {
  string bitext_str = bitext[pre_index];
  assert(bitext_str != "");
  vector<string> e, f;
  {
    vector<string> v;
    UTIL::split_string_using(bitext_str, "\t", &v);
    assert(v.size() == 2);
    
    if (reverse == "true") {
      string tmp = v[0];
      v[0] = v[1];
      v[1] = tmp;
    }
    
    UTIL::split_string_using(v[0], " ", &e);
    UTIL::split_string_using(v[1], " ", &f);
    
  }
  
  // 1. (optional) display all alignments, 
  // 2. checking i and j
  // build s_null
  set<int> s_null;
  for (int i = 1; i <= (int)e.size(); ++i) {
    s_null.insert(i);
  }

  string str = "";
  for(map<int, set<int> >::const_iterator it = m->begin(); it != m->end(); ++it) {
    
    int j = it->first;
    assert(j >= 0);
    assert(j <= (int)f.size());
    const set<int>& s = it->second;
    for (set<int>::const_iterator it2 = s.begin(); it2 != s.end(); ++it2) {
      int i = *it2;
      assert(i >= 1);
      assert(i <= (int)e.size());

      s_null.erase(i);
      
      char buffer[256];
      sprintf(buffer, "%d %d\n", j, i);
      str += string(buffer);
    }
  }
  
  // cout << str << endl;
  cout << "# index is " << pre_index << endl;
  
  // do the real thing!!!
  for (int i = 0; i < (int)e.size(); ++i) {
    cout << e[i] << " ";
  }
  cout << endl;
  

  cout << "NULL ({ ";
  for (set<int>::const_iterator it = s_null.begin(); it != s_null.end(); ++it) {
    cout << *it << " ";    
  }
  cout << "}) ";
  

  for (int j = 1; j <= (int)f.size(); ++j) {
    assert(j >= 0);
    assert(j <= (int)f.size());
    cout << f[j - 1] << " ({ ";
    
    const set<int>& s = (*m)[j];
    for (set<int>::const_iterator it2 = s.begin(); it2 != s.end(); ++it2) {
      int i = *it2;
      
      assert(i >= 1);
      assert(i <= (int)e.size());
      cout << i << " ";
      
      /*
# Sentence pair (1) source length 24 target length 20 alignment score : 1.47872e-24
２０００年 开局 、 我 国 国民 经济 面对 的 一 项 新 的 战略 任务 是 农业 结构 调整 。 
NULL ({ 9 11 13 }) at ({ }) the ({ }) beginning ({ 2 }) of ({ }) 2000 ({ 1 }) , ({ 3 }) china ({ 4 }) 's ({ 5 }) national ({ 6 }) economy ({ 7 }) faced ({ 8 }) a ({ 10 }) new ({ 12 }) strategic ({ 14 }) mission ({ 15 }) , ({ }) which ({ }) was ({ 16 }) the ({ }) structural ({ 18 }) adjustment ({ 19 }) of ({ }) agriculture ({ 17 }) . ({ 20 }) 
# Sentence pair (2) source length 14 target length 12 alignment score : 8.15834e-14
发展 城市 经济 还是 积累 工业化 条件 的 必由 之 路 。 
NULL ({ 8 }) developing ({ 1 }) urban ({ 2 }) economies ({ 3 }) is ({ }) also ({ }) a ({ }) path ({ }) to ({ }) accumulating ({ 4 5 9 10 11 }) the ({ }) conditions ({ 7 }) for ({ }) industrialization ({ 6 }) . ({ 12 }) 
      */
      
    }
    // close
    cout << "}) ";
    
  }
  cout << endl;
  
  // reset s;
  m->clear();
}

////////////////////////////////////////
//

int main(int argc, char* argv[]) {

  assert(argc == 4);
  const string& bitext_file = argv[1];
  const string& afile = argv[2];
  const string& reverse = argv[3];

  if (reverse == "true") {
  } else if (reverse == "false") {
  } else {
    assert(false);
  }

  // read bitext. For simplicity, I will read all them in.
  vector<string> bitext;
  bitext.push_back("");
  {
    string line;
    ifstream in (bitext_file.c_str());
    while (UTIL::getline(in, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, "\t", &v);
      if (v.size() == 0) {
	continue;
      }
      bitext.push_back(line);
    }
  }

  cerr << bitext.size() << endl;

  // read and transcode afile
  {
    map<int, set<int> > m;
    int pre_index = 0;

    string line;
    ifstream in(afile.c_str());
    while (UTIL::getline(in, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, " ", &v);
      if (v.size() == 0) {
	continue;
      }

      // trancode
      assert(v.size() == 4);
      int index = atoi(v[0].c_str());  // index of the bitext
      // index of e_i and f_j
      int i = atoi(v[2].c_str());
      int j = atoi(v[1].c_str());

      if (reverse == "true") {
	int tmp = i;
	i = j;
	j = tmp;
      }

      if (index > 10) {
	// break;
      }

      if (index != pre_index) {
	// we should output str
	if (pre_index == 0) {
	  // nothing to output
	  assert(m.size() == 0);
	  // continue;
	} else {
	  // now we have all information for one sentence pair
	  assert(m.size() != 0);

	  doit(bitext, pre_index, reverse,
	       //
	       &m);
	}

	pre_index = index;
      }

      // remember i and j for the same index
      if (m[i].size() == 0) {
	set<int> s;
	s.insert(j);
	m[i] = s;
      } else {
	set<int>& s = m[i];
	s.insert(j);
      }

    }

    // do not forget to output the last one.
    {
      assert(m.size() > 0);
      
      doit(bitext, pre_index, reverse,
	   //
	   &m);

    }

  }
 
  cerr << "done";
  return 0;
}
