
// hmm-classic_empty.cc

#include "config/global.h"
#include "hmm-classic-empty/hmm-classic-empty.h"
#include "ibm1-classic/ibm1-classic.h"
#include "wordalign-common/eval.h"
#include "wordalign-common/process-one-line.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>
#include <stack>

using namespace std;

namespace HMM_CLASSIC_EMPTY {

  Hmm_Classic_Empty::Hmm_Classic_Empty() {
  }

  Hmm_Classic_Empty::~Hmm_Classic_Empty() {
  }

  void Hmm_Classic_Empty::map() {
    
    vector<pair<vector<string>, vector<string> > > all_input;

    // Note: the first index is trg
    WORDALIGN_COMMON::Parameter_Freq parameter_trg_freq;
    WORDALIGN_COMMON::Parameter_Freq parameter_src_freq;
    WORDALIGN_COMMON::Parameter_Lex parameter_lex;
    WORDALIGN_COMMON::Parameter_Jump parameter_jump;
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;

    IBM1_CLASSIC::Ibm1_Classic::init_data_once(
					       //
					       &all_input,
					       &parameter_trg_freq,
					       &parameter_src_freq,
					       &parameter_lex,
					       &parameter_jump,
					       &parameter_lambda
					       );

    cerr << endl;
    cerr << "model: hmm" << endl;
    hmm_training(all_input, 
		 parameter_lex,
		 parameter_jump,
		 parameter_lambda
		 );
    
    
    // freq
    parameter_trg_freq.dump(cout, "trg-freq");
    cout << endl;

    parameter_src_freq.dump(cout, "src-freq");
    cout << endl;

    // lambda
    parameter_lambda.normalize(parameter_trg_freq);
    assert(parameter_lambda.key_size() > 0);
    parameter_lambda.dump(parameter_trg_freq, parameter_src_freq, WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH, cout);
    cout << endl;

    return;
  }

  void Hmm_Classic_Empty::reduce() {

    WORDALIGN_COMMON::Parameter_Freq parameter_trg_freq;
    WORDALIGN_COMMON::Parameter_Freq parameter_src_freq;
    WORDALIGN_COMMON::Parameter_Lex parameter_lex;
    WORDALIGN_COMMON::Parameter_Jump parameter_jump;
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;
    
    IBM1_CLASSIC::Ibm1_Classic::reduce_common(false,
					      &parameter_trg_freq,
					      &parameter_src_freq,
					      &parameter_lex,
					      &parameter_jump,
					      &parameter_lambda
					      );


    // if (MYCONFIG::my_global::FLAGS_debug) {
    if (true) {
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      parameter_lex.get_topk(WORDALIGN_COMMON::TOPK, &my_topk);
      
      cerr << "topk (lex) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	cerr << p.first << "\t" << "p(" << p.second << ")" << endl;
	my_topk.pop();
      }
      cerr << endl;
    }

    if (!MYCONFIG::my_global::FLAGS_training_only) {
      // decode and eval
      if (true) {
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result;
	const string str = "hmm-ibm1";
	//// hmm_decode_new(parameter_lex, prob_fert, prob_p0p1, str1, loop, &result);
	IBM1_CLASSIC::Ibm1_Classic::ibm1_decode_new(parameter_lex, str, &result);
	WORDALIGN_COMMON::Eval::eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result, str, MYCONFIG::my_global::FLAGS_loop_);
	cerr << "INFO: IBM1 decoding done at ";
	// system("date >> stderr");
	cerr << endl;
      }
      
      if (true) {
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result;
	const string str = "hmm-hmm";
	//// hmm_decode_new(parameter_lex, prob_fert, prob_p0p1, str1, loop, &result);
	hmm_decode_new(parameter_lex, parameter_jump, str, &result);
	WORDALIGN_COMMON::Eval::eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result, str, MYCONFIG::my_global::FLAGS_loop_);
	cerr << "INFO: HMM decoding done at ";
	// system("date >> stderr");
	cerr << endl;
      }
    }

    return;
  }

  void Hmm_Classic_Empty::compute_expected_counts(
					    const vector<pair<vector<string>, vector<string> > >& all_input,
					    const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					    const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
					    //
					    unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
					    unordered_map<int, unordered_map<int, unordered_map<int, double> > >* total_expected_count_jump
					    ) {
      
    int line_count = 0;
    const int LINE_COUNT = all_input.size();
    int STEP = LINE_COUNT / 100;
    ++STEP;
    
    for (vector<pair<vector<string>, vector<string> > >::const_iterator it = all_input.begin(); it != all_input.end(); ++it) {
      
      if (MYCONFIG::my_global::FLAGS_hadoop_) {
	cerr << "reporter:status:" << line_count << "/" << LINE_COUNT << endl;
      } else {
	if (line_count % STEP == 0) {
	  cerr << ".";
	}
      }
      ++line_count;
      
      // DP is better than Gibbs
      const vector<string>& v_trg = it->first;
      const vector<string>& v_src = it->second;

      const int I = (int)v_trg.size();
      int II_tmp = I / 2;
      if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
	II_tmp = I;
      } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
	assert(I % 2 == 0);
      } else {
	assert(false);
      }
      const int II = II_tmp;

      vector<unordered_map<int, double> > forward;
      vector<unordered_map<int, double> > backward;
      unordered_map<int, double> fwbw;
      unordered_map<int, double> fwbw2;

      /*
	compute forward backword probabilities (alpha and beta)
      */

      // compute alpha
      compute_forward(
		      v_trg, 
		      v_src, 
		      line_count,
		      parameter_lex,
		      parameter_jump,
		      //
		      &forward);
      // compute beta
      compute_backward(
		       v_trg, 
		       v_src, 
		       line_count,
		       parameter_lex,
		       parameter_jump,
		       //
		       &backward);
      // compute data likelihood, they should all look similar (indenpendent of j)
      compute_forward_backward(
			       v_trg, 
			       v_src,
			       forward, 
			       backward,
			       //
			       &fwbw);

      compute_forward_backward_2(
				 v_trg, 
				 v_src,
				 parameter_lex,
				 parameter_jump,
				 forward, 
				 backward,
				 //
				 &fwbw2);


      const double prob_data = fwbw[0];
      assert(prob_data >= 0);
      check_fwbw(prob_data, &fwbw, &fwbw2);

      /*
	compute expected counts using alpha and beta
      */

      for (int j = 0; j < (int)v_src.size(); ++j) {

	unordered_map<string, unordered_map<string, double> > expected_count_lex_1;
	unordered_map<string, unordered_map<string, double> > expected_count_lex_2;
	unordered_map<string, double> lex_sum_1;
	unordered_map<string, double> lex_sum_2;
	double lex_sum = 0;

	unordered_map<int, unordered_map<int, unordered_map<int, double> > > expected_count_jump_1;
	unordered_map<int, unordered_map<int, unordered_map<int, double> > > expected_count_jump_2;
	unordered_map<int, unordered_map<int, double> > jump_sum_1;
	unordered_map<int, unordered_map<int, double> > jump_sum_2;
	double jump_sum = 0;

	const string& src = v_src[j];

	// compute expected counts for lex
	for (int i = 0; i < (int)v_trg.size(); ++i) {
	  const string& trg = v_trg[i];

	  double tmp = 0;
	  if (prob_data > 0) {
	    tmp = forward[j][i] * backward[j][i] / prob_data;
	  } else {
	    tmp = WORDALIGN_COMMON::SMALL_LEX_SMOOTH;
	  } 
	  if (tmp) {
	    assert(tmp > 0);
	    if (false) {
	      cerr << "lex" << "\t" << trg << "\t" << src << "\t" << tmp << endl;
	    }
	    expected_count_lex_1[trg][src] += tmp;
	    expected_count_lex_2[src][trg] += tmp;
	  }
	}

	if (false) {
	  cerr << endl;
	}

	// compute expected counts for jump
	if (j > 0) {
	  for (int i = 0; i < (int)v_trg.size(); ++i) {
	    const string& trg = v_trg[i];
	    
	    double p_lex = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	    if (p_lex == 0) {
	      continue;
	    }
	    
	    for (int i_pre = 0; i_pre < (int)v_trg.size(); ++i_pre) {
	      
	      WORDALIGN_COMMON::JUMP_TYPE jt;
	      if (j == 1) {
		jt = WORDALIGN_COMMON::FROM_BOS;
	      } else if (j == (int)v_src.size() - 1) {
		jt = WORDALIGN_COMMON::TO_EOS;
	      } else {
		jt = WORDALIGN_COMMON::OTHERS;
	      }

	      double p_jump = parameter_jump.get(jt, i_pre, i, I, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	      if (p_jump == 0) {
		continue;
	      }
	      
	      double tmp = 0;
	      if (prob_data > 0) {
		tmp = p_lex * p_jump * forward[j-1][i_pre] * backward[j][i] / prob_data;
	      } else {
		tmp = WORDALIGN_COMMON::SMALL_JUMP_SMOOTH;
	      }

	      if (tmp) {
		assert(tmp > 0);
		if (false) {
		  cerr << endl;
		  cerr << "\t";
		  cerr << "trg " << trg << "\t";
		  cerr << "src " << src << "\t";
		  cerr << "i_pre " << i_pre << "\t";
		  cerr << "i " << i << "\t";
		  cerr << "p_lex " << p_lex << "\t";
		  cerr << "p_jump " << p_jump << endl;
		  cerr << endl;
		}
		
		if (false) {
		  cerr << "jump" << "\t" << jt << "\t" << i_pre << "\t" << i << "\t" << tmp << "\t" << p_jump << endl;
		}

		int type = -1;
		if (j == 1) {
		  type = 0;
		} else if (j == (int)v_src.size() -1) {
		  type = 2;
		} else {
		  type = 1;
		}
		assert(type != -1);

		expected_count_jump_1[type][i_pre][i] += tmp;
		expected_count_jump_2[type][i][i_pre] += tmp;
	      }
	      
	    } // for i_pre
	  } // for i
	} // j > 0

	// update counts for lex
	for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = expected_count_lex_1.begin(); it1 != expected_count_lex_1.end(); ++it1) {
	  const string& trg = it1->first;
	  const unordered_map<string, double>& tmp = it1->second;
	  
	  for (unordered_map<string, double>::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	    const string& src = it2->first;
	    double count = it2->second;
	    
	    (*total_expected_count_lex)[trg][src] += count;
	    lex_sum_1[trg] += count;
	    lex_sum_2[src] += count;
	    lex_sum += count;
	  }
	}
	
	// update counts for jump
	if (j > 0) {
	  for (unordered_map<int, unordered_map<int, unordered_map<int, double> > >::const_iterator it0 = expected_count_jump_1.begin(); it0 != expected_count_jump_1.end(); ++it0) {
	    int type = it0->first;
	    const unordered_map<int, unordered_map<int, double> >& tmp0 = it0->second;
	    
	    for (unordered_map<int, unordered_map<int, double> >::const_iterator it1 = tmp0.begin(); it1 != tmp0.end(); ++it1) {
	      int i_pre = it1->first;
	      const unordered_map<int, double>& tmp = it1->second;
	      
	      for (unordered_map<int, double>::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
		int i = it2->first;
		double count = it2->second;
		
		if (i > II) {
		  // We know it is p0, so we do not count this 
		  // othersize, there is a very high prob for width = 0
		} else {
		  if (i_pre > II) {
		    i_pre -= II;
		  }
		  (*total_expected_count_jump)[type][i_pre][i] += count;
		}
		jump_sum_1[type][i_pre] += count;
		jump_sum_2[type][i] += count;
		jump_sum += count;
	      }
	    }
	  }
	} else {  // j == 0
	  assert(expected_count_jump_1.size() == 0);
	  assert(expected_count_jump_2.size() == 0);
	  assert(jump_sum == 0);
	}

	if (prob_data > 0) {
	  assert(fabs(lex_sum - 1) < 0.00001);
	  if (j > 0) {
	    assert(fabs(jump_sum - 1) < 0.00001);
	  }
	}

      }  // for j
    }  // all_pairs
    // cerr << "training done" << endl;
    
    return;
  }

  void Hmm_Classic_Empty::hmm_training(const vector<pair<vector<string>, vector<string> > >& all_input,
				       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				       const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				       const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda
				       ) {
    
    unordered_map<string, unordered_map<string, double> > total_expected_count_lex;
    unordered_map<int, unordered_map<int, unordered_map<int, double> > > total_expected_count_jump;
    
    // compute expected counts
    compute_expected_counts(
			    all_input,
			    //
			    parameter_lex,
			    parameter_jump,
			    &total_expected_count_lex,
			    &total_expected_count_jump
			    );
    
    
    // dump expected counts for lex
    {
      for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = total_expected_count_lex.begin(); it1 != total_expected_count_lex.end(); ++it1) {
	const string& trg = it1->first;
	const unordered_map<string, double>& tmp = it1->second;
	for (unordered_map<string, double>::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	  const string& src = it2->first;
	  double count = it2->second;
	  cout << "lex" << "\t" << trg << "\t" << src << "\t" << count << endl;
	}
      }
    }

    cout << endl;

    // dump expected counts for jump
    {
      for (unordered_map<int, unordered_map<int, unordered_map<int, double> > >::const_iterator it1 = total_expected_count_jump.begin(); it1 != total_expected_count_jump.end(); ++it1) {
	int type = it1->first;
	const unordered_map<int, unordered_map<int, double> >& tmp1 = it1->second;
	
	for (unordered_map<int, unordered_map<int, double> >::const_iterator it2 = tmp1.begin(); it2 != tmp1.end(); ++it2) {
	    int i_pre = it2->first;
	  const unordered_map<int, double>& tmp2 = it2->second;

	  for (unordered_map<int, double>::const_iterator it3 = tmp2.begin(); it3 != tmp2.end(); ++it3) {
	      int i = it3->first;
	      double count = it3->second;
	      
	      cout << "jump" << "\t" << type << "\t" << abs(i - i_pre) << "\t" << count << endl;
	  }
	}
      }
    }

    cout << endl;

    if (MYCONFIG::my_global::FLAGS_debug) {
      // if (true) {
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      parameter_lex.get_topk(WORDALIGN_COMMON::TOPK, &my_topk);
      
      cerr << "topk (lex) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	cerr << p.first << "\t" << "p(" << p.second << ")" << endl;
	my_topk.pop();
      }
      cerr << endl;
    }

    return;
  }

  void Hmm_Classic_Empty::viterbi(const vector<string>& v_trg,
			    const vector<string>& v_src,
			    int line_count,
			    const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
			    const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
			    //
			    unordered_map<int, int>* v_max_align) {
    

    // Using DP (max)
    vector<unordered_map<int, double> > forward;
    vector<unordered_map<int, int> > back_pointer;

    const int I = v_trg.size();  
    const int J = v_src.size();

    // const int I = (int)v_trg.size();
    int II_tmp = I / 2;
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      II_tmp = I;
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(I % 2 == 0);
    } else {
      assert(false);
    }
    const int II = II_tmp;

    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

    for (int j = 0; j < J; ++j) {
      unordered_map<int, double> m1;
      if (j == 0) {
	m1[0] = 1;
      } else {
	m1[0] = 0;
      }
      forward.push_back(m1);

      unordered_map<int, int> m2;
      back_pointer.push_back(m2);
    }
    // cerr << forward->size() << endl;

    for (int j = 1; j <= J - 1; ++j) {
      const string& src = v_src[j];

      for (int i = 0; i < I; ++i) {
	const string& trg = v_trg[i];

	forward[j][i] = 0;

	// max over i_pre
	for (int i_pre = 0; i_pre < I; ++i_pre) {

	  if (line_count == 5) {
	    if (j == 43) {
	      if (i == 54) {
		if (i_pre > -1) {
		  // cerr << "aha" << endl;
		}
	      }
	    }
	  }

	  // we can put this out of the i_pre loop, but I put it here for debugging reason
	  double p_lex = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	  if (p_lex == 0) {
	    continue;
	  }

	  assert(i != 0);
	  if (j != J - 1) {
	    assert(i != II - 1);
	  } else {
	    assert(i == II - 1);
	  }

	  double p_jump = 0;
	  if (j == 1) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::FROM_BOS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else if (j == J - 1) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::TO_EOS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::OTHERS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  }
	  if (p_jump == 0) {
	    continue;
	  }
	  
	  if (false) {
	    cerr << "src " << src << endl;
	    cerr << "trg " << trg << endl;
	    cerr << "i_pre " << i_pre << endl;
	    cerr << "i " << i << endl;
	    cerr << "forward " << forward[j-1][i_pre] << endl;
	    cerr << "p_lex " << p_lex << endl;
	    cerr << "p_jump " << p_jump << endl;
	    cerr << "add " << forward[j-1][i_pre] * p_jump * p_lex << endl;
	  }

	  if (j == 1) {
	    if (MYCONFIG::my_global::FLAGS_combine_jump) {
	    } else {
	      assert(i_pre == 0);
	      assert(forward[j-1][i_pre] == 1);
	    }
	  } else if (j == J - 1) {
	    assert(i == II - 1);
	    assert(p_lex == 1);
	    if (MYCONFIG::my_global::FLAGS_combine_jump) {
	    } else {
	      assert(p_jump == 1);
	    }
	  }

	  if (forward[j-1][i_pre] * p_jump * p_lex > forward[j][i]) {
	    forward[j][i] = forward[j-1][i_pre] * p_jump * p_lex;
	    back_pointer[j][i] = i_pre;
	  }
	}  // i_pre
      }  // i
    }  // j
    
    // get the max at the end
    int max_index = -1;
    double max_value = -1;
    for (int index = 0; index < I; ++index) {
      int j = J - 1;
      double value = forward[j][index];
      if (value > max_value) {
	max_value = value;
	max_index = index;
      }
    }

    assert(max_index != -1);
    assert(max_value != -1);
    if (max_value > 0) {
      assert(max_index == II - 1);
    } else {
      max_index = II - 1;
    }

    // retrieve all the max indexes.
    unordered_map<int, int> max_back_pointer;
    max_back_pointer[J - 1] = max_index;
    for (int j = J - 1; j > 0; --j) {
      max_index = back_pointer[j][max_index];
      max_back_pointer[j - 1] = max_index;
    }

    assert(max_index == 0);
    
    for (int j = 0; j < (int)v_src.size(); ++j) {
      int max_i = max_back_pointer[j];
      (*v_max_align)[j] = max_i;
    }  // for each j

    assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_src[J - 1] == MYCONFIG::my_global::FLAGS_EOS);
    assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

    assert((*v_max_align)[0] == 0);
    assert((*v_max_align)[J - 1] == II - 1);

    return;
  }

  void Hmm_Classic_Empty::hmm_decode_new(const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				   const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				   const string& str,
				   //
				   unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result) {


    ifstream in(MYCONFIG::my_global::FLAGS_test_file.c_str());
    vector<pair<vector<string>, vector<string> > > all_input;
    
    {
      int line_count = 0;
      string line;
      while (UTIL::getline(in, &line)) {
	
	vector<string> v_trg, v_src;
	bool b = WORDALIGN_COMMON::Wordalign::process_one_line(line, 0, 0, &v_trg, &v_src);
	if (!b) {
	  continue;
	}
	
	++line_count;
	if (line_count % 2 == 0) {
	  // cerr << ".";
	}
	
	pair<vector<string>, vector<string> > p = make_pair(v_trg, v_src);
	all_input.push_back(p);
      }
      // cerr << endl;
    }

    char afile[1024];
    sprintf(afile, "%s-%d-%s.afile", MYCONFIG::my_global::FLAGS_project_id.c_str(), MYCONFIG::my_global::FLAGS_loop_, str.c_str());
    ofstream out(afile);

    int line_count = 0;
    const int LINE_COUNT = all_input.size();
    int STEP = LINE_COUNT / 100;
    ++STEP;

    for (vector<pair<vector<string>, vector<string> > >::const_iterator it = all_input.begin(); it != all_input.end(); ++it) {

      if (line_count % STEP == 0) {
	cerr << ".";
      }
      ++line_count;
      
      const vector<string>& v_trg = it->first;
      const vector<string>& v_src = it->second;

      const int I = (int)v_trg.size();
      int II_tmp = I / 2;
      if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
	II_tmp = I;
      } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
	assert(I % 2 == 0);
      } else {
	assert(false);
      }
      const int II = II_tmp;

      assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

      // DP is better than Gibbs
      // get viterbi
      unordered_map<int, int> v_max_align;
      viterbi(v_trg, v_src,
	      line_count,
	      parameter_lex,
	      parameter_jump,
	      //
	      &v_max_align);
      assert(v_max_align.size() == v_src.size());

      for (int j = 1; j < (int)v_src.size() - 1; ++j) {  // (0, J - 1)
	int max_i = v_max_align[j];
	if (max_i >= II) {
	  // removing empty alignment
	  continue;
	}
	if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	  // reverse
	  out << line_count << " " << j << " " << max_i << " S" << endl;
	} else {
	  // normal
	  out << line_count << " " << max_i << " " << j << " S" << endl;
	}
	(*result)[line_count][max_i][j] = true;
      }  // for each j
    }  // for each line
    cerr << endl;
    in.close();
    out.close();

    return;
  }
  
  void Hmm_Classic_Empty::compute_forward(
				    const vector<string>& v_trg, const vector<string>& v_src,
				    int line_count,
				    const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				    const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				    //
				    vector<unordered_map<int, double> >* forward
				    ) {

    const int I = v_trg.size();  
    const int J = v_src.size();

    // const int I = (int)v_trg.size();
    int II_tmp = I / 2;
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      II_tmp = I;
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(I % 2 == 0);
    } else {
      assert(false);
    }
    const int II = II_tmp;

    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

    for (int j = 0; j < J; ++j) {
      unordered_map<int, double> m;
      if (j == 0) {
	assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
	m[0] = 1;
      } else {
	m[0] = 0;
      }
      forward->push_back(m);
    }
    // cerr << forward->size() << endl;

    for (int j = 1; j <= J - 1; ++j) {
      const string& src = v_src[j];

      for (int i = 0; i < I; ++i) {
	const string& trg = v_trg[i];

	(*forward)[j][i] = 0;

	// old: we can put this out of the i_pre loop, but I put it here for debugging reason
	double p_lex = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	if (p_lex == 0) {
	  continue;
	}

	// sum over i_pre
	for (int i_pre = 0; i_pre < I; ++i_pre) {

	  if (j == 4) {
	    if (i == 1) {
	      if (i_pre == 0) {
		// cerr << "aha" << endl;
	      }
	    }
	  }

	  assert(i != 0);
	  if (j != J - 1) {
	    assert(i != II - 1);
	  } else {
	    assert(i == II - 1);
	  }

	  double p_jump = 0;
	  if (j == 1) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::FROM_BOS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else if (j == J - 1) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::TO_EOS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::OTHERS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  }
	  if (p_jump == 0) {
	    continue;
	  }
	  
	  assert((*forward)[j-1][i_pre] >= 0);
	  if ((*forward)[j-1][i_pre] > 0) {
	    if (false) {
	      cerr << endl;
	      cerr << "line_count " << line_count << endl;
	      cerr << "j " << j << endl;
	      cerr << "src " << src << endl;
	      cerr << "trg " << trg << endl;
	      cerr << "i_pre " << i_pre << endl;
	      cerr << "i " << i << endl;
	      cerr << "forward " << (*forward)[j-1][i_pre] << endl;
	      cerr << "p_lex " << p_lex << endl;
	      cerr << "p_jump " << p_jump << endl;
	      cerr << "add " << (*forward)[j-1][i_pre] * p_jump * p_lex << endl;
	    }

	    if (j == 1) {
	      assert(i_pre == 0);
	      assert((*forward)[j-1][i_pre] == 1);
	    } else if (j == J - 1) {
	      assert(i == II - 1);
	      assert(p_lex == 1);
	      if (MYCONFIG::my_global::FLAGS_combine_jump) {
	      } else {
		assert(p_jump == 1);
	      }
	    }
	    (*forward)[j][i] += (*forward)[j-1][i_pre] * p_jump * p_lex;
	  }

	}  // i_pre
      }  // i
    }  // j
    
    return;
  }


  void Hmm_Classic_Empty::compute_backward(
				     const vector<string>& v_trg, const vector<string>& v_src,
				     int line_count,
				     const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				     const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				     //
				     vector<unordered_map<int, double> >* backward
				     ) {
    
    const int I = v_trg.size();  
    const int J = v_src.size();

    // const int I = (int)v_trg.size();
    int II_tmp = I / 2;
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      II_tmp = I;
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(I % 2 == 0);
    } else {
      assert(false);
    }
    const int II = II_tmp;

    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

    for (int j = 0; j < J; ++j) {
      unordered_map<int, double> m;
      if (j == J - 1) {
	assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);
	m[II - 1] = 1;
      } else {
	m[II - 1] = 0;
      }
      backward->push_back(m);
    }
    // cerr << backward->size() << endl;

    for (int j = J - 2; j >= 0; --j) {
      const string& src = v_src[j+1];

      for (int i = 0; i < I; ++i) {
	(*backward)[j][i] = 0;
	
	// sum over i_post
	for (int i_post = 0; i_post < I; ++i_post) {
	  const string& trg = v_trg[i_post];

	  if (j == 2) {
	    if (i == 0) {
	      if (i_post >= 0) {
		// cerr << "aha" << endl;
	      }
	    }
	  }

	  double p_lex = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	  if (p_lex == 0) {
	    continue;
	  }

	  assert(i_post != 0);
	  if (j < (int)v_src.size() - 2) {
	    assert(i_post != II - 1);
	  } else {
	    assert(i_post == II - 1);
	  }
	  
	  double p_jump = 0;
	  if (j == 0) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::FROM_BOS, i, i_post, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else if (j == J - 2) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::TO_EOS, i, i_post, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::OTHERS, i, i_post, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  }
	  if (p_jump == 0) {
	    continue;
	  }
	  
	  assert((*backward)[j+1][i_post] >= 0);
	  if ((*backward)[j+1][i_post] > 0) {
	    if (false) {
	      cerr << "src " << src << endl;
	      cerr << "trg " << trg << endl;
	      cerr << "i " << i << endl;
	      cerr << "i_post " << i_post << endl;
	      cerr << "backward " << (*backward)[j+1][i_post] << endl;
	      cerr << "p_lex " << p_lex << endl;
	      cerr << "p_jump " << p_jump << endl;
	      cerr << "add " << (*backward)[j+1][i_post] * p_jump * p_lex << endl;
	    }
	    
	    if (j == J - 2) {
	      assert(i_post == II - 1);
	      assert(p_lex == 1);
	      if (MYCONFIG::my_global::FLAGS_combine_jump) {
	      } else {
		assert(p_jump == 1);
	      }
	      assert((*backward)[j+1][i_post] == 1);
	    } else if (j == 0) {
	      if (MYCONFIG::my_global::FLAGS_combine_jump) {
	      } else {
		assert(i == 0);
	      }
	    } else {
	    }
	    (*backward)[j][i] += (*backward)[j+1][i_post] * p_jump * p_lex;
	  }

	}  // i_post
      }  // i
    } // j

    return;
  }

  void Hmm_Classic_Empty::compute_forward_backward(
					     const vector<string>& v_trg, const vector<string>& v_src,
					     const vector<unordered_map<int, double> >& forward, 
					     const vector<unordered_map<int, double> >& backward,
					     //
					     unordered_map<int, double>* fwbw) {
    
    for (int j = 0; j < (int)v_src.size(); ++j) {
      (*fwbw)[j] = 0;
      // sum over i
      for (int i = 0; i < (int)v_trg.size(); ++i) {
	(*fwbw)[j] += get_2(forward, j, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) * get_2(backward, j, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
      }
    }

    return;
  }

  void Hmm_Classic_Empty::compute_forward_backward_2(
					       const vector<string>& v_trg, const vector<string>& v_src,
					       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					       const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
					       const vector<unordered_map<int, double> >& forward, 
					       const vector<unordered_map<int, double> >& backward,
					       //
					       unordered_map<int, double>* fwbw2) {
    

    for (int j = 1; j < (int)v_src.size(); ++j) {
      (*fwbw2)[j] = 0;
      // sum over i and i_pre
      for (int i = 0; i < (int)v_trg.size(); ++i) {
	for (int i_pre = 0; i_pre < (int)v_trg.size(); ++i_pre) {
	  const string& src = v_src[j];
	  const string& trg = v_trg[i];
	  double p_lex = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	  if (p_lex == 0) {
	    continue;
	  }
	  
	  double p_jump = 0;
	  if (j == 1) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::FROM_BOS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else if (j == (int)v_src.size() - 1) {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::TO_EOS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  } else {
	    p_jump = parameter_jump.get(WORDALIGN_COMMON::OTHERS, i_pre, i, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  }

	  if (false) {
	    cerr << "src " << src << endl;
	    cerr << "trg " << trg << endl;
	    cerr << "p_lex " << p_lex << endl;
	    cerr << "add " << p_lex * p_jump * get_2(forward, j - 1, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) * get_2(backward, j, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) << endl;
	    cerr << "result " << (*fwbw2)[j] + p_lex * p_jump * get_2(forward, j - 1, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) * get_2(backward, j, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) << endl;
	  }
	  
	  if (j == 0) {
	    if ((i == 0) && (i_pre== 0)) {
	      (*fwbw2)[j] += p_lex * p_jump * 1 * get_2(backward, j, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	    }
	  } else {
	    (*fwbw2)[j] += p_lex * p_jump * get_2(forward, j - 1, i_pre, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) * get_2(backward, j, i, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
	  }
	}
      }
    }
    return;
  }

  void Hmm_Classic_Empty::print_top_translations(
				     const string& trg,
				     const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				     const WORDALIGN_COMMON::Parameter_Jump& parameter_jump
				     ) {

    // output some examples

    priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
    parameter_lex.get_topk(WORDALIGN_COMMON::TOPK2, &my_topk);
  
    // we what to output the high probability words first.
    stack<pair<double, string> > s;
    while (!my_topk.empty()) {
      pair<double, string> p = my_topk.top();
      s.push(p);
      my_topk.pop();
    }
    while (!s.empty()) {
      pair<double, string> p = s.top();
      cerr << "\"" << p.second << "\"" << " (" << p.first << ") ";
      s.pop();
    }
    
    cerr << endl;

    return;
  }

  void Hmm_Classic_Empty::check_fwbw(double prob_data, 
			       const unordered_map<int, double>* fwbw,
			       const unordered_map<int, double>* fwbw2
			       ) const {

    for (unordered_map<int, double>::const_iterator it = fwbw->begin(); it != fwbw->end(); ++it) {
      double tmp = it->second;
      assert(fabs(prob_data - tmp) <= fabs(0.01 * prob_data));
    }

    for (unordered_map<int, double>::const_iterator it = fwbw2->begin(); it != fwbw2->end(); ++it) {
      double tmp = it->second;
      assert(fabs(prob_data - tmp) <= fabs(0.01 * prob_data));
    }
  }

}  // namespace
