
// hmm-classic_empty.h
#include <vector>
#include <cstdlib>
#include <iostream>
#include "util/unordered-util.h"
#include "wordalign-common/prob-table-util.h"

using namespace std;

namespace HMM_CLASSIC_EMPTY {

  // HMM
  // the transition probability is defined upon jump width
  // pay attention to the I in p(a|a', I) 
  //
  class Hmm_Classic_Empty {

  public:
    
  public:
    Hmm_Classic_Empty();
    ~Hmm_Classic_Empty();
    
    void map();
    void reduce();

    static void compute_forward(
			 const vector<string>& v_trg, const vector<string>& v_src,
			 int line_count,
			 const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
			 const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
			 //
			 vector<unordered_map<int, double> >* forward
			 );

    void compute_backward(
			  const vector<string>& v_trg, const vector<string>& v_src,
			  int line_count,
			  const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
			  const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
			  //
			  vector<unordered_map<int, double> >* backward
			  );

    // computing the data likelihood
    void compute_forward_backward(
				  const vector<string>& v_trg, const vector<string>& v_src,
				  const vector<unordered_map<int, double> >& forward, 
				  const vector<unordered_map<int, double> >& backward,
				  //
				  unordered_map<int, double>* fwbw);

    // another way of computing the data likelihood
    void compute_forward_backward_2(
				    const vector<string>& v_trg, const vector<string>& v_src,
				    const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				    const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				    const vector<unordered_map<int, double> >& forward, 
				    const vector<unordered_map<int, double> >& backward,
				    //
				    unordered_map<int, double>* fwbw);
    

    /*
    double get_forward(int e, int e) const;
    double get_backward(int f, int f) const;
    bool check_forward_backward() const;
    */
    void compute_expected_counts(
				 const vector<pair<vector<string>, vector<string> > >& all_input,
				 const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				 const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				 //
				 unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
				 unordered_map<int, unordered_map<int, unordered_map<int, double> > >* total_expected_count_jump
				 );

    void hmm_training(const vector<pair<vector<string>, vector<string> > >& all_input,
		      const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
		      const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
		      const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda
		      );
    
    static void viterbi(const vector<string>& v_trg,
			const vector<string>& v_src,
			int line_count,
			const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
			const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
			//
			unordered_map<int, int>* v_max_align
			);

    // *_new does not call system(...)
    static void hmm_decode_new(const WORDALIGN_COMMON::Parameter_Lex& parameter_lex, 
				const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				const string& str,
				//
				unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result);

    static void print_top_translations(
				       const string& trg,
				       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				       const WORDALIGN_COMMON::Parameter_Jump& parameter_jump
				       );

    void check_fwbw(double prob_data, 
		    const unordered_map<int, double>* fwbw,
		    const unordered_map<int, double>* fwbw2
		    ) const;

  };

}
