
// hmm-fertility-empty.cc

#include "config/global.h"
#include "hmm-classic-empty/hmm-classic-empty.h"
#include "hmm-fertility-empty/hmm-fertility-empty.h"
#include "ibm1-classic/ibm1-classic.h"
#include "wordalign-common/eval.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>

using namespace std;

namespace HMM_FERTILITY_EMPTY {

  // important, for debugging
  const bool RUN_SRAND = false;
  const string VITERBI_INIT = "ibm1";  // init alignments for Gibbs
  const bool ibm1_viterbi_without_empty_words = true;

  void Hmm_Fertility_Empty::Gibbs_training(const vector<pair<vector<string>, vector<string> > >& all_input,
					   int N_training,
					   int N_decoding,
					   const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
					   const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
					   const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					   const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
					   const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda,
					   //
					   unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
					   unordered_map<int, unordered_map<int, double> >* total_expected_count_jump,
					   unordered_map<string, unordered_map<int, unordered_map<int, double> > >* total_expected_count_fert
					   ) {
    
    // compute expected counts by Gibbs
    
    int line_count = 0;
    const int LINE_COUNT = all_input.size();
    int STEP = LINE_COUNT / 100;
    if (STEP == 0) {
      STEP = 1;
    }
    
    for (vector<pair<vector<string>, vector<string> > >::const_iterator it = all_input.begin(); it != all_input.end(); ++it) {
      
      if (MYCONFIG::my_global::FLAGS_hadoop_) {
	cerr << "reporter:status:" << line_count << "/" << LINE_COUNT << endl;
      } else {
	if (line_count % STEP == 0) {
	  cerr << ".";
	}
      }
      ++line_count;
      
      // for decoding
      unordered_map<int, int> v_max_align;
      unordered_map<int, unordered_set<int> > v_max_fert;
      double max_log_viterbi = 1;
      
      Gibbs_one_sentence(it->first,
			 it->second, 
			 parameter_trg_freq,
			 parameter_src_freq,
			 parameter_lex, 
			 parameter_jump,
			 parameter_lambda,
			 line_count,
			 N_training,
			 false,
			 //
			 &v_max_align, &v_max_fert, &max_log_viterbi,
			 total_expected_count_lex, 
			 total_expected_count_jump, 
			 total_expected_count_fert
			 );
      
    }  // for loop
    cerr << endl;

    return;
  }

  // Gibbs sampling
  void Hmm_Fertility_Empty::Gibbs_one_sentence(const vector<string>& v_trg,
					       const vector<string>& v_src,
					       const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
					       const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
					       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					       const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
					       const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda,
					       int line_count,
					       int N,
					       bool decoding,
					       //
					       unordered_map<int, int>* v_max_align,
					       unordered_map<int, unordered_set<int> >* v_max_fert,
					       double* max_log_viterbi,
					       unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
					       unordered_map<int, unordered_map<int, double> >* total_expected_count_jump,
					       unordered_map<string, unordered_map<int, unordered_map<int, double> > >* total_expected_count_fert
					       ) {

    // the un-normalized probability of the bitext sentence
    double data_uprob = 0;
    // for expected counts
    unordered_map<string, unordered_map<string, double> > expected_count_lex;
    unordered_map<int, unordered_map<int, double> > expected_count_jump;
    unordered_map<string, unordered_map<int, unordered_map<int, double> > > expected_count_fert;
    // unordered_map<string, double> expected_count_e;
    
    assert(v_trg.size() > 1);
    assert(v_src.size() >= 1);

    // init v_align and v_fert
    unordered_map<int, unordered_set<int> > v_fert;
    unordered_map<int, int> v_align;

    assert(VITERBI_INIT == "ibm1");
    if (VITERBI_INIT == "random") {
      
      for (int j = 0; j < (int)v_src.size(); ++j) {
	
	// const string& src = v_src[j];
	
	// empty word locates at pos
	int max_align = v_trg.size() - 1;
	assert(v_trg[max_align] == MYCONFIG::my_global::FLAGS_EOS);
	
	
	// random
	int rand_total = v_trg.size();
	assert(rand_total >= 2);
	max_align = rand() % rand_total;
	
	v_align[j] = max_align;
	v_fert[max_align].insert(j);
      }

    } else if (VITERBI_INIT == "ibm1") {

      const int I = (int)v_trg.size();
      int II_tmp = I / 2;
      if (II_tmp == 32) {
	assert(II_tmp == 32);
      }
      if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
	II_tmp = I;
      } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
	assert(I % 2 == 0);
      } else {
	assert(false);
      }
      const int II = II_tmp;
      assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_src[v_src.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);
      assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

      for (int j = 0; j < (int)v_src.size(); ++j) {
	
	const string& src = v_src[j];

	int max_align = -1;
	double max_prob = -1;
	// for debugging
	unordered_map<int, double> tmp_prob;
	  
	if (WORDALIGN_COMMON::Parameter_Jump::get_word_type(src) == WORDALIGN_COMMON::BOS) {
	  for (int i = 0; i < (int)v_trg.size(); ++i) {
	    const string& trg = v_trg[i];
	    if (WORDALIGN_COMMON::Parameter_Jump::get_word_type(trg) == WORDALIGN_COMMON::BOS) {
	      max_align = i;
	      max_prob = 1;
	      assert(max_align == 0);
	      break;
	    }
	  }
	} else if (WORDALIGN_COMMON::Parameter_Jump::get_word_type(src) == WORDALIGN_COMMON::EOS) {
	  for (int i = 0; i < (int)v_trg.size(); ++i) {
	    const string& trg = v_trg[i];
	    if (WORDALIGN_COMMON::Parameter_Jump::get_word_type(trg) == WORDALIGN_COMMON::EOS) {
	      max_align = i;
	      max_prob = 1;
	      assert(max_align == II - 1);
	      break;
	    }
	  }
	} else {
	  
	  // update to Viterbi alignments instead of empty word alignments
	  for (int cur_align = 0; cur_align < (int)v_trg.size(); ++cur_align) {
	    const string& trg = v_trg[cur_align];
	    double prob = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);

	    // what if it aligns to empty word
	    if (cur_align >= II) {
	      assert(WORDALIGN_COMMON::Parameter_Jump::get_word_type(trg) == WORDALIGN_COMMON::EMPTY);
	      assert(j > 0);
	      int pre_align = v_align[j - 1];
	      if (abs(pre_align - cur_align) % II != 0) {
		// consistent
		prob = 0;
	      } else if (ibm1_viterbi_without_empty_words){
		// we ignore all empty words when using ibm1
		prob = 0;
	      }
	    }

	    tmp_prob[cur_align] = prob;

	    if (prob > max_prob) {
	      max_prob = prob;
	      max_align = cur_align;
	    }
	  }
	}

	assert(max_align != -1);
	assert(max_prob != -1);
	v_align[j] = max_align;
	v_fert[max_align].insert(j);
      }
      
    } else if (VITERBI_INIT == "hmm") {
      
      unordered_map<int, int> v_max_align;
      HMM_CLASSIC_EMPTY::Hmm_Classic_Empty::viterbi(
						    v_trg, v_src,
						    line_count,
						    parameter_lex,
						    parameter_jump,
						    &v_max_align);
      
      
      assert(v_max_align.size() == v_src.size());
      for (int j = 0; j < (int)v_src.size(); ++j) {
	int max_align = v_max_align[j];
	v_align[j] = max_align;
	v_fert[max_align].insert(j);
      }
    } else {
      assert(false);
    }

    // make sure the fertility_empty set is not "empty"
    for (int i = 0; i < (int)v_trg.size(); ++i) {
      if (v_fert[i].size() == 0) {
	unordered_set<int> tmp;
	v_fert[i] = tmp;
      }
    }

    // init viterbi alignment
    *v_max_align = v_align;
    *v_max_fert = v_fert;

    double log_viterbi = get_log_viterbi(v_trg, 
					 v_src, 
					 v_align,
					 v_fert,
					 parameter_trg_freq,
					 parameter_src_freq,
					 parameter_lex,
					 parameter_jump,
					 parameter_lambda
					 );
    // cerr << log_viterbi << endl;
    assert(log_viterbi <= 0);
    
    // init max_log_viterbi
    if (*max_log_viterbi == 1) {
      *max_log_viterbi = log_viterbi;
    } else {
      assert(*max_log_viterbi <= 0);
    }

    int RE_START = N_training_re_start;
    if (decoding) {
      RE_START = N_decoding_re_start;
    }

    if (RUN_SRAND) {
      srand(0);
    }

    // TODO (shaojun) put restart out of one_sentence
    for (int i = 0; i < RE_START; ++i) {
      for (int n = 0; n < N; ++n) {
	Gibbs_one_move_new(v_trg, v_src, 
			   parameter_trg_freq, parameter_src_freq, 
			   parameter_lex, parameter_jump, parameter_lambda, 
			   line_count,
			   i,
			   n,
			   // (n > N/5),
			   (n >= 0),
			   decoding,
			   //
			   v_max_align, v_max_fert, 
			   max_log_viterbi, &log_viterbi,
			   &v_align, &v_fert, 
			   &expected_count_lex, 
			   &expected_count_jump, 
			   &expected_count_fert, 
			   &data_uprob);

	assert(log_viterbi <= 0);

      }
    }

    if (MYCONFIG::my_global::FLAGS_debug) {
      // lex
      cerr << endl;
      for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = expected_count_lex.begin(); it1 != expected_count_lex.end(); ++it1) {
	const string& trg = it1->first;
	const unordered_map<string, double>& map_tmp = it1->second;
	
	for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	  const string& src = it2->first;
	  double count = it2->second;
	  cerr << count << "\t" << trg << "\t" << src << endl;
	}      
      }
      cerr << endl;
    }

    // update total_expected_count_lex
    for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = expected_count_lex.begin(); it1 != expected_count_lex.end(); ++it1) {
      const string& trg = it1->first;
      const unordered_map<string, double>& map_tmp = it1->second;
      unordered_map<string, double>& map_tmp_total = (*total_expected_count_lex)[trg];
      
      for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	const string& src = it2->first;
	double count = it2->second;
	
	map_tmp_total[src] += count / data_uprob;
      }
    }

    // update total_expected_count_jump
    for (unordered_map<int, unordered_map<int, double> >::const_iterator it1 = expected_count_jump.begin(); it1 != expected_count_jump.end(); ++it1) {
      const int& trg = it1->first;
      const unordered_map<int, double>& map_tmp = it1->second;
      unordered_map<int, double>& map_tmp_total = (*total_expected_count_jump)[trg];
      
      for (unordered_map<int, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	const int& src = it2->first;
	double count = it2->second;
	
	map_tmp_total[src] += count / data_uprob;
      }
    }

    // update total_expected_count_fert
    for (unordered_map<string, unordered_map<int, unordered_map<int, double> > >::const_iterator it1 = expected_count_fert.begin(); it1 != expected_count_fert.end(); ++it1) {
      const string& trg = it1->first;
      // assert(trg != MYCONFIG::my_global::FLAGS_empty_word);

      const unordered_map<int, unordered_map<int, double> >& map_tmp = it1->second;
      
      for (unordered_map<int, unordered_map<int, double> >::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	int fert = it2->first;
	const unordered_map<int, double>& map_tmp_2 = it2->second;

	for (unordered_map<int, double>::const_iterator it3 = map_tmp_2.begin(); it3 != map_tmp_2.end(); ++it3) {
	  int shared_by = it3->first;
	  double count = it3->second;
	
	  (*total_expected_count_fert)[trg][fert][shared_by] += count / data_uprob;
	}
      }
    }

    return;
  }

  void Hmm_Fertility_Empty::Gibbs_decode_new(
					     const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
					     const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
					     const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					     const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
					     const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda,
					     int N_decoding,
					     const string& str,
					     //
					     unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result) {

    ifstream in(MYCONFIG::my_global::FLAGS_test_file.c_str());

    char afile[1024];
    sprintf(afile, "%s-%d-%s.afile", MYCONFIG::my_global::FLAGS_project_id.c_str(), MYCONFIG::my_global::FLAGS_loop_, str.c_str());
    ofstream out(afile);

    string line;
    int line_count = 0;
    while (UTIL::getline(in, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, "\t", &v);
      assert(v.size() == 2);
      
      ++line_count;
      
      if (line_count % 2 == 0) {
	cerr << ".";
      }

      // change trg/v[0] to lower case
      UTIL::to_lower(&v[0]);

      vector<string> v_trg00, v_src00;
      if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	// reverse
	UTIL::split_string_using(v[1], " ", &v_trg00);
	UTIL::split_string_using(v[0], " ", &v_src00);
      } else {  
	// normal
	UTIL::split_string_using(v[0], " ", &v_trg00);
	UTIL::split_string_using(v[1], " ", &v_src00);
      }
      
      // add <s> and </s> to the source sentence
      vector<string> v_src0;
      v_src0.push_back(MYCONFIG::my_global::FLAGS_BOS);
      for (int i = 0; i < (int)v_src00.size(); ++i) {
	v_src0.push_back(v_src00[i]);
      }
      v_src0.push_back(MYCONFIG::my_global::FLAGS_EOS);
      
      // add <s> and </s> to the target sentence
      vector<string> v_trg0;
      v_trg0.push_back(MYCONFIG::my_global::FLAGS_BOS);
      for (int i = 0; i < (int)v_trg00.size(); ++i) {
	v_trg0.push_back(v_trg00[i]);
      }
      v_trg0.push_back(MYCONFIG::my_global::FLAGS_EOS);
      
      assert(v_src00.size() >= 1);
      assert(v_src0.size() >= 3);
      assert(v_src00.size() + 2 == v_src0.size());
      
      assert(v_trg00.size() >= 1);
      assert(v_trg0.size() >= 3);
      assert(v_trg00.size() + 2 == v_trg0.size());
      
      // decode one sentence
      unordered_map<string, unordered_map<string, double> > total_expected_count_lex;
      unordered_map<int, unordered_map<int, double> > total_expected_count_jump;
      unordered_map<string, unordered_map<int, unordered_map<int, double> > > total_expected_count_fert;
      
      // for decoding
      unordered_map<int, int> v_max_align;
      unordered_map<int, unordered_set<int> > v_max_fert;
      double max_viterbi = 0;
      
      Gibbs_one_sentence(v_trg0, 
			 v_src0,
			 parameter_trg_freq, 
			 parameter_src_freq, 
			 parameter_lex, 
			 parameter_jump,
			 parameter_lambda,
			 line_count,
			 N_decoding,
			 true,
			 //
			 &v_max_align, &v_max_fert, &max_viterbi,
			 &total_expected_count_lex, 
			 &total_expected_count_jump, 
			 &total_expected_count_fert);
      
      // loop
      for (int j = 0; j < (int)v_max_align.size(); ++j) {
	int i = v_max_align[j];
	assert(v_trg0.size() == v_max_fert.size());
	if (i + 1 == (int)v_max_fert.size()) { // ?
	  // out << 0 << "\t" << j + 1 << "\t" << 1 << endl;
	  assert(v_trg0[i] == MYCONFIG::my_global::FLAGS_EOS);
	} else {
	  if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	    // reverse
	    out << line_count << " " << j << " " << i << " S" << endl;
	  } else {
	    // normal
	    out << line_count << " " << i << " " << j << " S" << endl;
	  }
	  (*result)[line_count][i][j] = true;
	}
      }
      // out << "end" << endl;
    }
    cerr << endl;
    in.close();
    out.close();

    return;
  }

}
