
// hmm-fertility-empty.cc

#include "config/global.h"
#include "hmm-classic-empty/hmm-classic-empty.h"
#include "hmm-fertility-empty/hmm-fertility-empty.h"
#include "ibm1-classic/ibm1-classic.h"
#include "wordalign-common/eval.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>
#include <stack>

using namespace std;

namespace HMM_FERTILITY_EMPTY {

  // unordered_map<string, int> Hmm_Fertility_Empty::total_count_of_e_;

  int Hmm_Fertility_Empty::N_training_re_start = 3;  // 3 is too small, and 12 is good
  int Hmm_Fertility_Empty::N_training = 10;
  int Hmm_Fertility_Empty::N_decoding_re_start = 12;  // 3 is too small, and 12 is good
  int Hmm_Fertility_Empty::N_decoding = 100;

  Hmm_Fertility_Empty::Hmm_Fertility_Empty() {
    if (MYCONFIG::my_global::FLAGS_sampling_type == "for-toy-debugging") {
      N_training_re_start = 10;
      N_training = 1000;
      N_decoding_re_start = 10;
      N_decoding = 1000;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "quick") {
      N_training_re_start = 2;
      N_training = 5;
      N_decoding_re_start = 2;
      N_decoding = 5;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "standard") {  // standard
      N_training_re_start = 5;
      N_training = 15;
      N_decoding_re_start = 5;
      N_decoding = 15;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "hmm") {  // increased training, reduced decoding; for 491 data and fast running
      N_training_re_start = 1;  // 5;  // 1 is good!
      N_training = 15;
      // N_training = 50;
      // N_training = 100;
      N_training = MYCONFIG::my_global::FLAGS_samples_;
      assert(N_decoding_re_start = 12);
      assert(N_decoding == 100);
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "final") {  // reduced training; for 20k, 50k data
      N_training_re_start = 2;
      N_training = 5;
      assert(N_decoding_re_start == 12);
      assert(N_decoding == 100);
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "improved") {

      N_training_re_start *= 4;

    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "long") {
      N_training_re_start *= 5;
      N_training *= 10;
      N_decoding_re_start *= 20;
      N_decoding *= 20;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "longest") {
      N_training_re_start *= 10;
      N_training *= 20;
      N_decoding_re_start *= 50;
      N_decoding *= 100;
    } else {
      assert(false);
    }
  }

  Hmm_Fertility_Empty::~Hmm_Fertility_Empty() {
  }

  void Hmm_Fertility_Empty::map() {
    
    /* 
       init
    */
    vector<pair<vector<string>, vector<string> > > all_input;

    // Note: the first index is trg
    WORDALIGN_COMMON::Parameter_Freq parameter_trg_freq;
    WORDALIGN_COMMON::Parameter_Freq parameter_src_freq;
    WORDALIGN_COMMON::Parameter_Lex parameter_lex;
    WORDALIGN_COMMON::Parameter_Jump parameter_jump;
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;

    IBM1_CLASSIC::Ibm1_Classic::init_data_once(
					       //
					       &all_input,
					       &parameter_trg_freq,
					       &parameter_src_freq,
					       &parameter_lex,
					       &parameter_jump,
					       &parameter_lambda
					       );

    cerr << endl << endl;
    cerr << "model: hmmf" << endl;

    /*
      training
    */
    unordered_map<string, unordered_map<string, double> > total_expected_count_lex;
    unordered_map<int, unordered_map<int, double> > total_expected_count_jump;
    unordered_map<string, unordered_map<int, unordered_map<int, double> > > total_expected_count_fert;

    Gibbs_training(all_input, 
		   N_training,
		   N_decoding,
		   parameter_trg_freq, 
		   parameter_src_freq, 
		   parameter_lex, 
		   parameter_jump,
		   parameter_lambda,
		   //
		   &total_expected_count_lex,
		   &total_expected_count_jump,
		   &total_expected_count_fert
		   );
    

    // dump freq
    parameter_trg_freq.dump(cout, "trg-freq");
    cout << endl;

    parameter_src_freq.dump(cout, "src-freq");
    cout << endl;

    // dump expected counts for lex
    for (unordered_map<string, unordered_map<string, double> >::iterator it1 = total_expected_count_lex.begin(); it1 != total_expected_count_lex.end(); ++it1) {
      const string& trg = it1->first;
      unordered_map<string, double>& map_tmp = it1->second;
      for (unordered_map<string, double>::iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	const string& src = it2->first;
	double& count = it2->second;
	cout << "lex" << "\t" << trg << "\t" << src << "\t" << count << endl;
      }
    }
    
    cout << endl;

    // dump expected counts for jump
    {
      for (unordered_map<int, unordered_map<int, double> >::const_iterator it1 = total_expected_count_jump.begin(); it1 != total_expected_count_jump.end(); ++it1) {
	int i_pre = it1->first;
	const unordered_map<int, double>& tmp = it1->second;
	
	for (unordered_map<int, double>::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	  int i = it2->first;
	  double count = it2->second;

	  // TODO: jt is not a constant
	  int jt = 1;

	  cout << "jump" << "\t" << jt << "\t" << abs(i - i_pre) << "\t" << count << endl;
	}
      }
    }

    cout << endl;

    // dump expected counts for fert
    {
      for (unordered_map<string, unordered_map<int, unordered_map<int, double> > >::const_iterator it1 = total_expected_count_fert.begin(); it1 != total_expected_count_fert.end(); ++it1) {
	const string& trg = it1->first;
	const unordered_map<int, unordered_map<int, double> >& tmp = it1->second;
	
	for (unordered_map<int, unordered_map<int, double> >::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {

	  int fert = it2->first;
	  const unordered_map<int, double>& tmp2 = it2->second;
	  
	  for (unordered_map<int, double>::const_iterator it3 = tmp2.begin(); it3 != tmp2.end(); ++it3) {
	    
	    int shared_by = it3->first;
	    double count = it3->second;
	    
	    cout << "lambda" << "\t" << trg << "\t" << fert << "\t" << shared_by << "\t" << count << endl;
	  }
	}
      }
    }

    cout << endl;

    return;
  }

  void Hmm_Fertility_Empty::reduce() {
    
    WORDALIGN_COMMON::Parameter_Freq parameter_trg_freq;
    WORDALIGN_COMMON::Parameter_Freq parameter_src_freq;
    WORDALIGN_COMMON::Parameter_Lex parameter_lex;
    WORDALIGN_COMMON::Parameter_Jump parameter_jump;
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;

    IBM1_CLASSIC::Ibm1_Classic::reduce_common(true,
					      &parameter_trg_freq,
					      &parameter_src_freq,
					      &parameter_lex,
					      &parameter_jump,
					      &parameter_lambda
					      );

    /*
      update_parameters for lambda's
    */

    // TODO: what is the difference between update_lambda and count_lambda normalize and dump?

    /*
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;
    double parameter_lambda_non_empty_word;
    double parameter_lambda_empty_word;
    
    update_lambda(
		  count_lambda,
		  //
		  &parameter_lambda,
		  &parameter_lambda_non_empty_word,
		  &parameter_lambda_empty_word
		  );

    // dump
    {
      for (unordered_map<string, double>::const_iterator it =  parameter_lambda.begin(); it != parameter_lambda.end(); ++it) {
	const string& trg = it->first;
	double lambda = it->second;
	if (lambda > WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH) {
	  cout << "lambda" << "\t" << trg << "\t" << lambda << endl;
	}
      }
      cout << "lambda_non_empty_word" << "\t" << parameter_lambda_non_empty_word << endl;
      cout << "lambda_empty_word" << "\t" << parameter_lambda_empty_word << endl;
    }

    // output

    if (MYCONFIG::my_global::FLAGS_debug) {
      priority_queue<pair<double, string>, vector<pair<double, pair<string, string> > >, WORDALIGN_COMMON::topk3> my_topk;
      count_lex.get_topk(WORDALIGN_COMMON::TOPK, &my_topk);
      
      cerr << "topk (lex) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, pair<string, string> > p = my_topk.top();
	cerr << "\"" << p.first << "\"" << "\t" << "p(" << p.second.first << " | " << p.second.second << ")" << endl;
	//
	const string& trg = p.second.second;
	// cerr << trg << "\t" << MYCONFIG::my_global::t_freq_[trg] << endl;
	int freq = HMM_FERTILITY_EMPTY::Hmm_Fertility_Empty::total_count_of_e_[trg];
	if (freq >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	  // if (MYCONFIG::my_global::t_freq_[trg] >= 1) {
	  double lambda = parameter_lambda.get(trg);
	  cerr << lambda << "\tp(lambda | " << trg << ")" << " "
	       << freq << " ";
	  HMM_CLASSIC::Hmm_Classic::print_top_translations(trg, count_lex, count_jump);
	}
	my_topk.pop();
      }  // not empty
    } // true
    cerr << endl;

    bool too_small = false;
    // if (MYCONFIG::my_global::FLAGS_debug) {

    if (MYCONFIG::my_global::FLAGS_debug) {
      // if (true) {
      // output lambda for empty word
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      // priority_queue<pair<double, string> > my_topk(topk);
      for (unordered_map<string, double>::iterator it1 = parameter_lambda.begin(); it1 != parameter_lambda.end(); ++it1) {
	const string& trg = it1->first;
	int freq = HMM_FERTILITY_EMPTY::Hmm_Fertility_Empty::total_count_of_e_[trg];
	if (freq >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	  double prob = -1.0 * it1->second;		     
	  						     
	  my_topk.push(make_pair(prob, trg));		     
	  if ((int)my_topk.size() > WORDALIGN_COMMON::TOPK) {
	    my_topk.pop();				     
	    assert((int)my_topk.size() <= WORDALIGN_COMMON::TOPK);
	  }
	}
      }
      
      cerr << "topk (lambda) " << my_topk.size() << endl;
      if ((int)my_topk.size() < WORDALIGN_COMMON::TOPK) {
	too_small = true;
      } else {
	assert((int)my_topk.size() == WORDALIGN_COMMON::TOPK);
      }

      stack<pair<double, string> > s1;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	s1.push(p);
	my_topk.pop();
      }

      while (!s1.empty()) {
	pair<double, string> p = s1.top();
	int freq = HMM_FERTILITY_EMPTY::Hmm_Fertility_Empty::total_count_of_e_[p.second];
	cerr << -1 * p.first << "\tp(lambda | " << p.second << ")" << " "
	     << freq << " ";
	HMM_CLASSIC::Hmm_Classic::print_top_translations(p.second, count_lex, count_jump);
	s1.pop();
      }

      cerr << endl;
      cerr << "--" << endl;
      // cerr << "p(lambda|unk) " << parameter_lambda[MYCONFIG::my_global::FLAGS_unknown_word] << endl;
      cerr << "p(lambda|*) " << parameter_lambda_non_empty_word << "\t";
      cerr << "p(lambda|eps) " << parameter_lambda_empty_word << "\t";
      system("date");
      cerr << "--" << endl;
      cerr << endl;
    }
    
    if (MYCONFIG::my_global::FLAGS_debug) {
      // if (true) {
      if (!too_small) {
	priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
	// priority_queue<pair<double, string> > my_topk(topk);
	for (unordered_map<string, double>::iterator it1 = parameter_lambda.begin(); it1 != parameter_lambda.end(); ++it1) {
	  const string& trg = it1->first;
	  int freq = HMM_FERTILITY_EMPTY::Hmm_Fertility_Empty::total_count_of_e_[trg];
	  if (freq >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	    double prob = it1->second;
	    
	    my_topk.push(make_pair(prob, trg));
	    if ((int)my_topk.size() > WORDALIGN_COMMON::TOPK) {
	      my_topk.pop();
	      assert((int)my_topk.size() <= WORDALIGN_COMMON::TOPK);
	    }
	  }
	}
	
	cerr << "topk (lambda) " << my_topk.size() << endl;
	while (!my_topk.empty()) {
	  pair<double, string> p = my_topk.top();
	  int freq = HMM_FERTILITY_EMPTY::Hmm_Fertility_Empty::total_count_of_e_[p.second];
	  cerr << p.first << "\tp(lambda | " << p.second << ")" << " "
	       << freq << " ";
	  HMM_CLASSIC::Hmm_Classic::print_top_translations(p.second, count_lex, count_jump);
	  my_topk.pop();
	}
      }
    }
    cerr << endl;
    */

    // if (MYCONFIG::my_global::FLAGS_debug) {
    if (true) {
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      parameter_lex.get_topk(WORDALIGN_COMMON::TOPK, &my_topk);
      
      cerr << "topk (lex) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	cerr << p.first << "\t" << "p(" << p.second << ")" << endl;
	my_topk.pop();
      }
      cerr << endl;
    }

    if (!MYCONFIG::my_global::FLAGS_training_only) {
      // decode and eval
      if (true) {
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result;
	const string str = "hmmf-ibm1";
	//// hmm_decode_new(parameter_lex, prob_fert, prob_p0p1, str1, loop, &result);
	IBM1_CLASSIC::Ibm1_Classic::ibm1_decode_new(parameter_lex, str, &result);
	WORDALIGN_COMMON::Eval::eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result, str, MYCONFIG::my_global::FLAGS_loop_);
	cerr << "INFO: IBM1 decoding done at ";
	// system("date >> stderr");
      }
      
      if (true) {
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result;
	const string str = "hmmf-hmm";
	//// hmm_decode_new(parameter_lex, prob_fert, prob_p0p1, str1, loop, &result);
	HMM_CLASSIC_EMPTY::Hmm_Classic_Empty::hmm_decode_new(parameter_lex, parameter_jump, str, &result);
	WORDALIGN_COMMON::Eval::eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result, str, MYCONFIG::my_global::FLAGS_loop_);
	cerr << "INFO: HMM decoding done at ";
	// system("date >> stderr");
      }
    }

    return;
  }


}  // namespace
