
// hmm-classic.h
#include <vector>
#include <cstdlib>
#include <iostream>
#include "util/unordered-util.h"
#include "wordalign-common/prob-table-util.h"

using namespace std;

namespace HMM_FERTILITY_EMPTY {

  // HMM
  // the transition probability is defined upon jump width
  // pay attention to the I in p(a|a', I) 
  //

  class Hmm_Fertility_Empty {

  public:

    static int N_training_re_start;
    static int N_training;
    static int N_decoding_re_start;
    static int N_decoding;
    
    // static unordered_map<string, int> total_count_of_e_;

  public:
    Hmm_Fertility_Empty();
    ~Hmm_Fertility_Empty();
    
    void map();
    void reduce();

    void Gibbs_training(const vector<pair<vector<string>, vector<string> > >& all_input,
			int N_training,
			int N_decoding,
			const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
			const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
			const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
			const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
			const WORDALIGN_COMMON::Parameter_Lambda& prob_lambda,
			//
			unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
			unordered_map<int, unordered_map<int, double> >* total_expected_count_jump,
			unordered_map<string, unordered_map<int, unordered_map<int, double> > >* total_expected_count_fert
			);
    void Gibbs_one_sentence(const vector<string>& v_trg,
			    const vector<string>& v_src,
			    const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
			    const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
			    const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
			    const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
			    const WORDALIGN_COMMON::Parameter_Lambda& prob_lambda,
			    int line_count,
			    int N,
			    bool decoding,
			    //
			    unordered_map<int, int>* v_max_align,
			    unordered_map<int, unordered_set<int> >* v_max_fert,
			    double* max_log_viterbi,
			    unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
			    unordered_map<int, unordered_map<int, double> >* total_expected_count_jump,
			    unordered_map<string, unordered_map<int, unordered_map<int, double> > >* total_expected_count_fert);
    void Gibbs_one_move_new(
			    const vector<string>& v_trg, const vector<string>& v_src,
			    const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
			    const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
			    const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
			    const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
			    const WORDALIGN_COMMON::Parameter_Lambda& prob_lambda,
			    const int line_count,
			    const int loop_1,
			    const int loop_2,
			    bool save,
			    bool decoding,
			    //
			    unordered_map<int, int>* v_max_align,
			    unordered_map<int, unordered_set<int> >* v_max_fert,
			    double* max_log_viterbi,
			    double* log_viterbi,
			    unordered_map<int, int>* v_align,
			    unordered_map<int, unordered_set<int> >* v_fert,
			    unordered_map<string, unordered_map<string, double> >* expected_count_lex,
			    unordered_map<int, unordered_map<int, double> >* total_expected_count_jump,
			    unordered_map<string, unordered_map<int, unordered_map<int, double> > >* expected_count_fert,
			    // unordered_map<string, double>* expected_count_e,
			    double* data_uprob);
    
    void update_for_lex_and_jump(const vector<string>& v_trg, const vector<string>& v_src,
				 const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
				 const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
				 const unordered_map<int, int>& v_align,
				 int jj, // current j
				 int old_align,
				 int new_align,
				 int pre_align,
				 int post_align,
				 // 
				 double* log_viterbi
				 );

    void update_for_distortion(
			       const vector<string>& v_trg, const vector<string>& v_src,
			       const unordered_map<int, int>& v_align,
			       int old_align,
			       int new_align,
			       int old_align_old_fert,
			       int old_align_new_fert,
			       int new_align_old_fert,
			       int new_align_new_fert,
			       //
			       double* log_viterbi
			       );

    void update_for_fertility(
			      const vector<string>& v_trg, const vector<string>& v_src,
			      const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
			      const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
			      const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
			      const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
			      const WORDALIGN_COMMON::Parameter_Lambda& prob_lambda,
			      const unordered_map<int, int>& v_align,
			      int old_align,
			      int new_align,
			      int old_align_old_fert,
			      int old_align_new_fert,
			      int new_align_old_fert,
			      int new_align_new_fert,
			      //
			      double* log_viterbi
			      );

    void Gibbs_decode_new(
			  const WORDALIGN_COMMON::Parameter_Freq& prob_trg_freq,
			  const WORDALIGN_COMMON::Parameter_Freq& prob_src_freq,
			  const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
			  const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
			  const WORDALIGN_COMMON::Parameter_Lambda& prob_lambda,
			  int N_decoding,
			  const string& str,
			  //
			  unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result);

    double get_log_viterbi(
			   const vector<string>& v_trg, 
			   const vector<string>& v_src,
			   const unordered_map<int, int>& v_align,
			   const unordered_map<int, unordered_set<int> >& v_fert,
			   const WORDALIGN_COMMON::Parameter_Freq& prob_trg_freq,
			   const WORDALIGN_COMMON::Parameter_Freq& prob_src_freq,
			   const WORDALIGN_COMMON::Parameter_Lex& prob_lex,
			   const WORDALIGN_COMMON::Parameter_Jump& prob_jump,
			   const WORDALIGN_COMMON::Parameter_Lambda& prob_lambda
			   );


  };

}
