
// hmm-fertility-empty-one-move.cc

#include "config/global.h"
#include "ibm1-classic/ibm1-classic.h"
// #include "hmm-classic/hmm-classic.h"
#include "hmm-fertility-empty/hmm-fertility-empty.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>
#include <stack>

using namespace std;

namespace HMM_FERTILITY_EMPTY {

  void Hmm_Fertility_Empty::Gibbs_one_move_new(const vector<string>& v_trg, const vector<string>& v_src, 
					       const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
					       const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
					       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					       const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
					       const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda,
					       const int line_count,
					       const int loop_1,
					       const int loop_2,
					       bool after_burn_in,
					       bool decoding,
					       //
					       unordered_map<int, int>* v_max_align,
					       unordered_map<int, unordered_set<int> >* v_max_fert,
					       double* max_log_viterbi,
					       double* log_viterbi,
					       unordered_map<int, int>* v_align,
					       unordered_map<int, unordered_set<int> >* v_fert,
					       unordered_map<string, unordered_map<string, double> >* expected_count_lex,
					       unordered_map<int, unordered_map<int, double> >* expected_count_jump,
					       unordered_map<string, unordered_map<int, unordered_map<int, double> > >* expected_count_fert,
					       // unordered_map<string, double>* expected_count_e,
					       double* data_uprob) {
    

    const int I = (int)v_trg.size();
    int II_tmp = I / 2;
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      II_tmp = I;
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(I % 2 == 0);
    } else {
      assert(false);
    }
    const int II = II_tmp;
    assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_src[v_src.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);
    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

    // Gibbs sampling
    for (int jj = 0; jj < (int)v_src.size(); ++jj) {
      
      const string& src = v_src[jj];

      if (line_count == 9) {
	if (loop_1 == 0) {
	  if (loop_2 == 3) {
	    if (jj == 20) {
	      assert(jj == 20);
	    }
	  }
	}
      }
      
      // we do not nned to chanage the alignment for <s> and </s>
      if (src == MYCONFIG::my_global::FLAGS_BOS) {
	assert(jj == 0);
	continue;
      } else if (src == MYCONFIG::my_global::FLAGS_EOS) {
	assert(jj == (int)v_src.size() - 1);
	continue;
      }
      
      const int old_align = (*v_align)[jj];
      const int old_align_old_fert = (*v_fert)[old_align].size();

      vector<double> pp;  // pp = p(a_j | not_a_j, E, F) for each a_j
      
      for (int ii = 0; ii < (int)v_trg.size(); ++ii) {  
	const string& trg = v_trg[ii];
	if (trg == MYCONFIG::my_global::FLAGS_BOS) {
	  // do not change the alignment to <s>
	  pp.push_back(0);
	  continue;
	} else if (trg == MYCONFIG::my_global::FLAGS_EOS) {
	  // do not change the alignment to </s>
	  pp.push_back(0);
	  continue;
	} else if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	  assert(trg == MYCONFIG::my_global::FLAGS_empty_word);
	}

	// TODO
	/*
	  Due to the jump constraints for empty words,
	  in the following cases, 
	  the probability is zero.
	 */

	if (v_trg.size() == 64) {
	  assert(v_trg.size() == 64);
	}

	// if previous is empty
	assert(jj > 0);
	assert(jj < (int)v_src.size() - 1);
	int pre_align = (*v_align)[jj - 1];
	int post_align = (*v_align)[jj + 1];
	if (v_trg[ii] == MYCONFIG::my_global::FLAGS_empty_word) {
	  if(abs(ii - pre_align) % II != 0) {
	    pp.push_back(0);
	    continue;
	  }
	}
	if (v_trg[post_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	  if(abs(post_align - ii) % II != 0) {
	    pp.push_back(0);
	    continue;
	  }
	}

	const int new_align = ii;
	if (ii >= II) {
	  assert(v_trg[ii] == MYCONFIG::my_global::FLAGS_empty_word);
	  assert(abs(ii - pre_align) % II == 0);
	}

	int new_align_old_fert = (*v_fert)[new_align].size();
	int old_align_new_fert = old_align_old_fert;
	int new_align_new_fert = new_align_old_fert;
	if (new_align != old_align) {
	  --old_align_new_fert;
	  ++new_align_new_fert;
	}

	if (MYCONFIG::my_global::FLAGS_debug) {
	  cerr << "source: " << jj << " " << v_src[jj] << endl;
	  cerr << "target: " << ii << " " << v_trg[ii] << endl;
	  cerr << "old_align: " << old_align << " " << v_trg[old_align] << endl;
	  cerr << "old_align_old_fert: " << old_align_old_fert << endl;
	  cerr << "old_align_new_fert: " << old_align_new_fert << endl;
	  cerr << "new_align: " << new_align << " " << v_trg[new_align] << endl;
	  cerr << "new_align_old_fert: " << new_align_old_fert << endl;
	  cerr << "new_align_new_fert: " << new_align_new_fert << endl;
	}
	
	// we will update this variable.
	double my_log_viterbi = *log_viterbi;
	assert(my_log_viterbi <= 0);

	// update lex diff
	update_for_lex_and_jump(v_trg, v_src,
				parameter_lex,
				parameter_jump,
				*v_align,
				jj,
				old_align,
				new_align,
				(*v_align)[jj-1],
				(*v_align)[jj+1],
				//
				&my_log_viterbi);
	assert(my_log_viterbi <= 0);

	// update distortion
	update_for_distortion(
			      v_trg, v_src,
			      *v_align,
			      old_align,
			      new_align,
			      old_align_old_fert,
			      old_align_new_fert,
			      new_align_old_fert,
			      new_align_new_fert,
			      //
			      &my_log_viterbi
			      );
	assert(my_log_viterbi <= 0);

	// update fert/lambda diff
	update_for_fertility(
			     v_trg, v_src,
			     parameter_trg_freq,
			     parameter_src_freq,
			     parameter_lex,
			     parameter_jump,
			     parameter_lambda,
			     *v_align,
			     old_align,
			     new_align,
			     old_align_old_fert,
			     old_align_new_fert,
			     new_align_old_fert,
			     new_align_new_fert,
			     //
			     &my_log_viterbi
			     );
	assert(my_log_viterbi <= 0);

	// done with my_log_viterbi;

	if (decoding) {
	  if (my_log_viterbi > *max_log_viterbi) {
	    *max_log_viterbi = my_log_viterbi;
	    *v_max_align = *v_align;
	    *v_max_fert = *v_fert;
	    // we do not update v_align and v_fert because we may move to another state
	    // so we have to update v_max_* below
	    // update
	    (*v_max_align)[jj] = new_align;
	    // update
	    (*v_max_fert)[old_align].erase(jj);
	    (*v_max_fert)[new_align].insert(jj);
	  }
	}

	// we will move to this state proportioan to this weight
	double weight = exp(my_log_viterbi);
	pp.push_back(weight);
      }  // for ii

      // normalize, so weight become probability
      double sum = 0;
      assert(pp.size() == v_trg.size());
      for (int i = 0; i < (int)pp.size(); ++i) {
	sum += pp[i];
      }

      if (MYCONFIG::my_global::FLAGS_debug) {
	for (int i = 0; i < (int)pp.size(); ++i) {
	  cerr << pp[i] / sum << " " << v_src[jj] << " " << v_trg[i] << " " << pp[i] << endl;
	}
      }

      if (sum > 0) {
	for (int i = 0; i < (int)pp.size(); ++i) {
	  pp[i] /= sum;
	}
      } else {
	assert(sum == 0);
	continue;
      }

      // sample alignment from the conditiona distribution "pp"
      // pp = p(a_j | not_a_j, E, F) for each a_j
      double r = UTIL::rand_r();
      while (r == 0) {
	r = UTIL::rand_r();
      }
      double sum_p = 0;
      int new_align = 0;
      assert(pp[0] == 0);
      assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
      for (int i = 1; i < (int)pp.size(); ++i) {  // starting from 1 because the first one is <s>
	sum_p += pp[i];
	if (sum_p >= r) {
	  new_align = i;
	  break;
	}
      }
      assert(v_trg[new_align] != MYCONFIG::my_global::FLAGS_BOS);
      assert(v_trg[new_align] != MYCONFIG::my_global::FLAGS_EOS);
      assert(new_align > 0);
      assert(new_align < (int)v_trg.size() - 1);
      
      // remember old and new align and fert variables
      int new_align_old_fert = (*v_fert)[new_align].size();
      int old_align_new_fert = old_align_old_fert;
      int new_align_new_fert = new_align_old_fert;
      if (new_align != old_align) {
	--old_align_new_fert;
	++new_align_new_fert;
      }

      // update current alignments and fert variables
      (*v_align)[jj] = new_align;
      (*v_fert)[old_align].erase(jj);
      (*v_fert)[new_align].insert(jj);

      // verify
      assert((int)(*v_fert)[old_align].size() == old_align_new_fert);
      assert((int)(*v_fert)[new_align].size() == new_align_new_fert);

      // update lex diff
      update_for_lex_and_jump(v_trg, v_src,
			      parameter_lex,
			      parameter_jump,
			      *v_align,
			      jj,
			      old_align,
			      new_align,
			      (*v_align)[jj-1],
			      (*v_align)[jj+1],
			      //
			      log_viterbi);
      assert(*log_viterbi <= 0);
      
      // update distortion
      update_for_distortion(
			    v_trg, v_src,
			    *v_align,
			    old_align,
			    new_align,
			    old_align_old_fert,
			    old_align_new_fert,
			    new_align_old_fert,
			    new_align_new_fert,
			    //
			    log_viterbi
			    );
      assert(*log_viterbi <= 0);

      // update fert/lambda diff
      update_for_fertility(
			   v_trg, v_src,
			   parameter_trg_freq,
			   parameter_src_freq,
			   parameter_lex,
			   parameter_jump,
			   parameter_lambda,
			   *v_align,
			   old_align,
			   new_align,
			   old_align_old_fert,
			   old_align_new_fert,
			   new_align_old_fert,
			   new_align_new_fert,
			   //
			   log_viterbi
			   );
      assert(*log_viterbi <= 0);

      // done with log_viterbi

      // save this sample
      if (after_burn_in) {
	// update expected_count_lex
	for (int j = 0; j < (int)v_src.size(); ++j) {
	  const string& src = v_src[j];
	  int align = (*v_align)[j];
	  const string& trg = v_trg[align];
	  unordered_map<string, double>& expected_count_lex_tmp = (*expected_count_lex)[trg];
	  ++expected_count_lex_tmp[src];

	  if (j > 0) {
	    int align_pre = (*v_align)[j-1];
	    ++(*expected_count_jump)[align_pre][align];
	  }
	}
	
	// update expected_count_fert
	{
	  assert ((MYCONFIG::my_global::FLAGS_wa_model == "hmmf") ||
		  (MYCONFIG::my_global::FLAGS_wa_model == "upoisson") ||
		  (MYCONFIG::my_global::FLAGS_wa_model == "prior") ||
		  (MYCONFIG::my_global::FLAGS_wa_model == "poisson"));

	  for (int i = 0; i < (int)v_trg.size(); ++i) {
	    const string& trg = v_trg[i];
	    unordered_map<int, unordered_map<int, double> >& expected_count_fert_tmp = (*expected_count_fert)[trg];
	    const unordered_set<int> set_fert = get_1(*v_fert, i);
	    int fert = set_fert.size();
	    int shared_by = 1;

	    // fertility for empty words are shared by all empty words in v_trg.
	    if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	      assert(v_trg.size() - II > 0);
	      shared_by = v_trg.size() - II;
	    }

	    ++expected_count_fert_tmp[fert][shared_by];
	  }
	}
	
	++(*data_uprob);
      }

    }  // for jj
      
    return;
  }

  double Hmm_Fertility_Empty::get_log_viterbi(
				const vector<string>& v_trg, 
				const vector<string>& v_src,
				const unordered_map<int, int>& v_align,
				const unordered_map<int, unordered_set<int> >& v_fert,
				const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
				const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
				const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
				const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda
				) {
    // init log_viterbi
    double log_viterbi = 0;  // log (N!/n1! n2! ... nm!) \prob p(fj|e_{a_j})

    // lex and jump
    for (int j = 0; j < (int)v_src.size(); ++j) {
      int cur_align = -1;
      {
	unordered_map<int, int>::const_iterator it = v_align.find(j);
	assert(it != v_align.end());
	int align = it->second;
	cur_align = align;
	double prob = parameter_lex.get(v_trg[align], v_src[j], WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	assert(prob > 0);
	assert(prob <= 1);
	log_viterbi += log(prob);
	assert(log_viterbi <= 0);
      }

      assert(cur_align != -1);

      if (j > 0) {
	unordered_map<int, int>::const_iterator it = v_align.find(j-1);
	assert(it != v_align.end());
	int pre_align = it->second;
	if (v_trg.size() == 82) {
	  assert(1 == 1);
	}
	double prob = parameter_jump.get(WORDALIGN_COMMON::OTHERS, pre_align, cur_align, v_trg.size(), WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH);
	assert(prob > 0);
	assert(prob <= 1);
	log_viterbi += log(prob);
	assert(log_viterbi <= 0);
      }
    }
    assert(log_viterbi <= 0);
    
    // fert
    assert(MYCONFIG::my_global::FLAGS_wa_model == "hmmf");
      
    // lambda
    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      assert(v_trg[v_trg.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(v_trg[v_trg.size() - 1] == MYCONFIG::my_global::FLAGS_empty_word);
    } else {
      assert(false);
    }
    
    bool b_empty_words = false;
    int phi_empty_word = 0;
    for (int i = 0; i < (int)v_trg.size(); ++i) {
      const string& trg = v_trg[i];
      
      if (trg == MYCONFIG::my_global::FLAGS_BOS) {
	continue;
      } else if (trg == MYCONFIG::my_global::FLAGS_EOS) {
	continue;
      } else if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	b_empty_words = true;
	// cumulate phi for empty_word
	unordered_map<int, unordered_set<int> >::const_iterator it = v_fert.find(i);
	assert(it != v_fert.end());
	const unordered_set<int>& set_tmp = it->second;
	phi_empty_word += set_tmp.size();
	// we will add lambda factor for empty words in the end;
	continue;
      }
      
      // retrieve phi_i
      unordered_map<int, unordered_set<int> >::const_iterator it = v_fert.find(i);
      assert(it != v_fert.end());
      const unordered_set<int>& set_tmp = it->second;
      const int phi_i = set_tmp.size();
      
      WORDALIGN_COMMON::WORD_TYPE wt = WORDALIGN_COMMON::Parameter_Jump::get_word_type(trg);
      assert(wt == WORDALIGN_COMMON::NON_EMPTY);
      int freq = parameter_trg_freq.get(trg);
      
      // retrieve lambda_i
      double lambda_i = 0;
      if (freq < MYCONFIG::my_global::FLAGS_freq_word_count) {
	lambda_i = parameter_lambda.get_lambda_for_non_empty_word(parameter_trg_freq, parameter_src_freq);
      } else {
	lambda_i = parameter_lambda.get(trg, WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH);
      }
      
      // scalar for the empty word's Poisson distribution
      if (MYCONFIG::my_global::FLAGS_scale == true) {
	if (v_trg[i] == MYCONFIG::my_global::FLAGS_empty_word) {
	  assert(false);
	  const int size = v_trg.size() - 1;
	  assert(size >= 1);
	  lambda_i *= size;
	}
      }
      
      // update log_viterbi
	// formula: lambda_i^phi_i e^(-lambda_i) / (phi_i)!
      log_viterbi += phi_i * log(lambda_i);
      log_viterbi -= lambda_i;
      for (int phi = 1; phi <= phi_i; ++phi) {
	log_viterbi -= log(phi);
      }
    }  // for

    // lambda factor for empty words
    if (b_empty_words) {
      const int I = (int)v_trg.size();
      int II_tmp = I / 2;
      if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
	II_tmp = I;
      } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
	assert(I % 2 == 0);
      } else {
	assert(false);
      }
      const int II = II_tmp;
      assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_src[v_src.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);
      assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

      const int factor = v_trg.size() - II;
      assert(factor >= 1);

      double lambda_empty_word = parameter_lambda.get_lambda_for_empty_word(parameter_trg_freq, parameter_src_freq);
      if (lambda_empty_word == 0) {
	lambda_empty_word = 1.0e-8;
      }
      lambda_empty_word *= factor;

      log_viterbi += phi_empty_word * log(lambda_empty_word);
      log_viterbi -= lambda_empty_word;
      for (int phi = 1; phi <= phi_empty_word; ++phi) {
	log_viterbi -= log(phi);
      }
    }

    return log_viterbi;
  }

  /*
   distortion (wa-model and distortion model)
   1. upoisson: uniform
   2. uniform: uniform
   3. poisson: 
     a) empty-word: give more chances to empty words
     b) phi: proportional to fert
     c) uniform: uniform
  */
  void Hmm_Fertility_Empty::update_for_distortion(const vector<string>& v_trg, const vector<string>& v_src,
						  const unordered_map<int, int>& v_align,
						  int old_align,
						  int new_align,
						  int old_align_old_fert,
						  int old_align_new_fert,
						  int new_align_old_fert,
						  int new_align_new_fert,
						  //
						  double* log_viterbi
						  ) {

    // poisson
    if (MYCONFIG::my_global::FLAGS_wa_model != "poisson") {
      return;
    }

    // hmmf is not poisson
    assert(false);

    assert(MYCONFIG::my_global::FLAGS_wa_model == "poisson");

    if (MYCONFIG::my_global::FLAGS_distortion == "uniform") {
    } else if (MYCONFIG::my_global::FLAGS_distortion == "empty-word") {
      // proportioanl to k(e_{a_j})
      // pros: empty-word > uniform > phi
      // cons: this results in very big lambda for the empty word
      if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	const int size = v_trg.size() - 1;
	assert(size >= 1);
	assert (MYCONFIG::my_global::FLAGS_scale_value > 0);
	assert (MYCONFIG::my_global::FLAGS_scale_value <= 1);
	*log_viterbi -= log(size * MYCONFIG::my_global::FLAGS_scale_value);
      }
      if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	const int size = v_trg.size() - 1;
	assert(size >= 1);
	assert (MYCONFIG::my_global::FLAGS_scale_value > 0);
	assert (MYCONFIG::my_global::FLAGS_scale_value <= 1);
	*log_viterbi += log(size * MYCONFIG::my_global::FLAGS_scale_value);
	// *log_viterbi += log(v_trg.size() - 1);
      }
    } else if (MYCONFIG::my_global::FLAGS_distortion == "phi") {
      // proportional to \phi
      // this results in very small \lambda for freq words
      assert(v_trg.size() >= 2);

      if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	old_align_old_fert *= v_trg.size() - 1;
	old_align_new_fert *= v_trg.size() - 1;
      }

      if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	new_align_old_fert *= v_trg.size() - 1;
	new_align_new_fert *= v_trg.size() - 1;
      }

      assert(old_align_old_fert > 0);
      *log_viterbi -= old_align_old_fert * log(old_align_old_fert);


      if (old_align_new_fert > 0) {
	*log_viterbi += old_align_new_fert * log(old_align_new_fert);
      }
      if (new_align_old_fert > 0) {
	*log_viterbi -= new_align_old_fert * log(new_align_old_fert);
      }
      assert(new_align_new_fert > 0);
      *log_viterbi += new_align_new_fert * log(new_align_new_fert);
    } else {
      cerr << "wrong option for distion: " << MYCONFIG::my_global::FLAGS_distortion << endl;
      assert(false);
    }

    return;
  }

  void Hmm_Fertility_Empty::update_for_lex_and_jump(const vector<string>& v_trg, const vector<string>& v_src,
						    const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
						    const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
						    const unordered_map<int, int>& v_align,
						    int jj, // current j
						    int old_align,
						    int new_align,
						    int pre_align,
						    int post_align,
						    // 
						    double* log_viterbi
						    ) {

    // we do not change the alignment for <s> and </s>
    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      assert(v_trg[v_trg.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(v_trg[v_trg.size() - 1] == MYCONFIG::my_global::FLAGS_empty_word);
    } else {
      assert(false);
    }
    assert(jj > 0);
    assert(jj < (int)v_src.size() - 1);

    double old_align_prob = parameter_lex.get(v_trg[old_align], v_src[jj], WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
    if (old_align_prob == 0) {
      old_align_prob = WORDALIGN_COMMON::SMALL_LEX_SMOOTH;
    }
    *log_viterbi -= log(old_align_prob);
    assert(*log_viterbi <= 0);

    double new_align_prob = parameter_lex.get(v_trg[new_align], v_src[jj], WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
    if (new_align_prob == 0) {
      new_align_prob = WORDALIGN_COMMON::SMALL_LEX_SMOOTH;
    }
    *log_viterbi += log(new_align_prob);
    assert(*log_viterbi <= 0);

    double old_jump_prob_pre = parameter_jump.get(WORDALIGN_COMMON::OTHERS, pre_align, old_align, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
    if (old_jump_prob_pre <= 0) {
      // cerr << "ha" << endl;
    }
    *log_viterbi -= log(old_jump_prob_pre);
    assert(*log_viterbi <= 0);

    double new_jump_prob_pre = parameter_jump.get(WORDALIGN_COMMON::OTHERS, pre_align, new_align, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
    if (new_jump_prob_pre == 0) {
      // cerr << "ha" << endl;
    }
    *log_viterbi += log(new_jump_prob_pre);
    assert(*log_viterbi <= 0);

    if (old_align == 38) {
      if (post_align == 31) {
	assert(old_align == 38);
      }
    }

    double old_jump_prob_post = parameter_jump.get(WORDALIGN_COMMON::OTHERS, old_align, post_align, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
    if (old_jump_prob_post == 0) {
      // cerr << "ha" << endl;
    }
    *log_viterbi -= log(old_jump_prob_post);
    assert(*log_viterbi <= 0);

    double new_jump_prob_post = parameter_jump.get(WORDALIGN_COMMON::OTHERS, new_align, post_align, v_trg.size(), WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
    if (new_jump_prob_post == 0) {
      // cerr << "ha" << endl;
    }
    *log_viterbi += log(new_jump_prob_post);
    assert(*log_viterbi <= 0);
    
    return;
  }

  void Hmm_Fertility_Empty::update_for_fertility(
						 const vector<string>& v_trg, const vector<string>& v_src,
						 const WORDALIGN_COMMON::Parameter_Freq& parameter_trg_freq,
						 const WORDALIGN_COMMON::Parameter_Freq& parameter_src_freq,
						 const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
						 const WORDALIGN_COMMON::Parameter_Jump& parameter_jump,
						 const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda,
						 const unordered_map<int, int>& v_align,
						 int old_align,
						 int new_align,
						 int old_align_old_fert,
						 int old_align_new_fert,
						 int new_align_old_fert,
						 int new_align_new_fert,
						 //
						 double* log_viterbi
						 ) {
    
    assert(MYCONFIG::my_global::FLAGS_wa_model == "hmmf");

    const int I = (int)v_trg.size();
    int II_tmp = I / 2;
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      II_tmp = I;
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(I % 2 == 0);
    } else {
      assert(false);
    }
    const int II = II_tmp;
    assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_src[v_src.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);
    assert(v_trg[0] == MYCONFIG::my_global::FLAGS_BOS);
    assert(v_trg[II - 1] == MYCONFIG::my_global::FLAGS_EOS);

    if (old_align != new_align) {
      /*
	remove old
      */
      double old_lambda = parameter_lambda.get_lambda_for_non_empty_word(parameter_trg_freq, parameter_src_freq);
      int old_freq = parameter_trg_freq.get(v_trg[old_align]);
      if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	old_lambda = parameter_lambda.get_lambda_for_empty_word(parameter_trg_freq, parameter_src_freq);
      } else if (old_freq >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	old_lambda = parameter_lambda.get(v_trg[old_align], WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH);
      }
      
      if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	const int factor = v_trg.size() - II;
	assert(factor >= 1);
	old_lambda *= factor;
      }
      
      // update
      /*
	\lambda(e_i)^{\phi_i} e^{-\lambda(e_i)}
	---------------------------------------
	\phi_i !
      */
      
      assert(old_align_old_fert == old_align_new_fert + 1);
      assert (MYCONFIG::my_global::FLAGS_double_factorial == true);
      if (true) {
	*log_viterbi -= log(old_lambda);
	*log_viterbi += log(old_align_old_fert);
      } else {
	*log_viterbi -= log(old_lambda);
	//// *log_viterbi += old_lambda;
	// *log_viterbi += log(old_align_old_fert);
      }
      
      /* 
	 add new
      */
      double new_lambda = parameter_lambda.get_lambda_for_non_empty_word(parameter_trg_freq, parameter_src_freq);
      int new_freq = parameter_trg_freq.get(v_trg[new_align]);
      if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	new_lambda = parameter_lambda.get_lambda_for_empty_word(parameter_trg_freq, parameter_src_freq);
      } else if (new_freq >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	new_lambda = parameter_lambda.get(v_trg[new_align], WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH);
      }
      
      if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	const int factor = v_trg.size() - II;
	assert(factor >= 1);
	new_lambda *= factor;
      }
      
      // update
      /*
	\lambda(e_i)^{\phi_i} e^{-\lambda(e_i)}
	  ---------------------------------------
	  \phi_i !
      */
      
      assert(new_align_old_fert == new_align_new_fert - 1);
      assert (MYCONFIG::my_global::FLAGS_double_factorial == true);
      if (true) {
	*log_viterbi += log(new_lambda);
	*log_viterbi -= log(new_align_new_fert); // -
      } else {
	*log_viterbi += log(new_lambda);
	//// *log_viterbi -= new_lambda;
	// *log_viterbi -= log(new_align_new_fert); // -
      }
    } else {
      assert(old_align_old_fert == old_align_new_fert);
      assert(new_align_old_fert == new_align_new_fert);
    }

    return;
  }

}  // namespace
