
// fert1.cc

#include "config/global.h"
#include "ibm1-classic/ibm1-classic.h"
#include "wordalign-common/eval.h"
#include "wordalign-common/process-one-line.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>
#include <stack>

using namespace std;

namespace IBM1_CLASSIC {
  
  // important, for debugging
  const bool VITERBI_INIT = true;  // init alignments for Gibbs
  const bool RUN_SRAND = false;

  Ibm1_Classic::Ibm1_Classic() {
  }

  Ibm1_Classic::~Ibm1_Classic() {
  }

  bool Ibm1_Classic::init_data_once(
			     //
			     vector<pair<vector<string>, vector<string> > >* all_input,
			     WORDALIGN_COMMON::Parameter_Freq* parameter_trg_freq,
			     WORDALIGN_COMMON::Parameter_Freq* parameter_src_freq,
			     WORDALIGN_COMMON::Parameter_Lex* parameter_lex,
			     WORDALIGN_COMMON::Parameter_Jump* parameter_jump,
			     WORDALIGN_COMMON::Parameter_Lambda* parameter_lambda
			     ) {

    static int run_once = 0;
    ++run_once;

    assert(run_once == 1);
    assert(all_input->size() == 0);
    assert(parameter_trg_freq->key_size() == 0);
    assert(parameter_src_freq->key_size() == 0);
    assert(parameter_lex->key_size() == 0);
    assert(parameter_jump->key_size() == 0);
    assert(parameter_lambda->key_size() == 0);

    bool b_trg_freq = false;
    bool b_src_freq = false;
    bool b_lex = false;
    bool b_jump = false;
    bool b_lambda = false;

    // read parameters
    {
      string line;
      
      ifstream in(MYCONFIG::my_global::FLAGS_current_parameter.c_str());
      while (UTIL::getline(in, &line)) {
	vector<string> v;
	UTIL::split_string_using(line, "\t", &v);
	if (v.size() == 0) {
	  continue;
	}
	if (v[0] == "trg-freq") {

	  assert(v.size() == 3);
	  b_trg_freq = true;

	  const string& trg = v[1];
	  int freq = atoi(v[2].c_str());
	  // freq
	  double tmp = parameter_trg_freq->get(trg);
	  if (tmp != 0) {
	    assert(tmp == freq);
	  } else {
	    parameter_trg_freq->add(trg, freq);
	  }

	} else if (v[0] == "src-freq") {

	  assert(v.size() == 3);
	  b_src_freq = true;

	  const string& src = v[1];
	  int freq = atoi(v[2].c_str());
	  // freq
	  double tmp = parameter_src_freq->get(src);
	  if (tmp != 0) {
	    assert(tmp == freq);
	  } else {
	    parameter_src_freq->add(src, freq);
	  }

	} else if (v[0] == "lex") {

	  assert(v.size() == 4);

	  b_lex = true;
	  const string& trg = v[1];
	  const string& src = v[2];
	  double value = atof(v[3].c_str());
	  // assert(parameter_lex->get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH) == 0);
	  parameter_lex->set(trg, src, value);

	} else if (v[0] == "jump") {

	  assert(v.size() == 4);

	  b_jump = true;
	  int type = atoi(v[1].c_str());
	  int width = atoi(v[2].c_str());
	  double value = atof(v[3].c_str());

	  if (type == 0) {
	    parameter_jump->set2(WORDALIGN_COMMON::FROM_BOS, width, value);
	  } else if (type == 1) {
	    parameter_jump->set2(WORDALIGN_COMMON::OTHERS, width, value);
	  } else if (type == 2) {
	    parameter_jump->set2(WORDALIGN_COMMON::TO_EOS, width, value);
	  } else {
	    assert(false);
	  }

	} else if (v[0] == "lambda") {

	  assert(v.size() == 5);
	  b_lambda = true;

	  const string& trg = v[1];
	  int fert = atoi(v[2].c_str());
	  assert(fert == 1);
	  int shared_by = atoi(v[3].c_str());
	  assert(shared_by == 1);
	  double value = atof(v[4].c_str());
	  // assert(parameter_lambda->get(trg, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH) == 0);

	  // init data, using set
	  double tmp = parameter_lambda->get(trg, 0);
	  assert(tmp == 0);
	  parameter_lambda->set(trg, value);

	} else if (v[0] == "length_trg") {
	} else if (v[0] == "length_src") {
	} else {
	  assert(false);
	}
      }
    }
    
    assert(b_trg_freq == true);
    assert(b_src_freq == true);
    assert(b_lex == true);
    assert(b_jump == true);
    assert(b_lambda == true);

    assert(parameter_trg_freq->key_size() > 0);
    assert(parameter_src_freq->key_size() > 0);

    parameter_lex->normalize();
    assert(parameter_lex->key_size() > 0);

    parameter_jump->normalize();
    assert(parameter_jump->key_size() > 0);

    // we used set, so there is no point to normalize
    // parameter_lambda->normalize(*parameter_trg_freq);
    assert(parameter_lambda->key_size() > 0);

    cerr << "lambda for the empty word: " << parameter_lambda->get("_empty_word_", 0) << endl;

    // read in all the map input
    {
      cerr << "reading..." << endl;
      
      string line;
      int line_count = 0;
      while (UTIL::getline(cin, &line)) {

	vector<string> v_trg, v_src;
	bool b = WORDALIGN_COMMON::Wordalign::process_one_line(line, WORDALIGN_COMMON::TRG_LEN_LIMIT, WORDALIGN_COMMON::SRC_LEN_LIMIT, &v_trg, &v_src);
	if (!b) {
	  continue;
	}

	++line_count;
	if (line_count % 1000 == 0) {
	  cerr << ".";
	}

	pair<vector<string>, vector<string> > p = make_pair(v_trg, v_src);
	all_input->push_back(p);
      }
    }
    cerr << endl;

    cerr << "bitext size " << all_input->size() << endl;
    cerr << "parameter_lex size1 " << parameter_lex->key_size() << endl;
    cerr << "parameter_lex size2 " << parameter_lex->get_total_size() << endl;
    cerr << "parameter_jump size " << parameter_jump->key_size() << endl;
    cerr << "parameter_lambda size " << parameter_lambda->key_size() << endl;

    return true;
  }

  void Ibm1_Classic::map() {
    
    vector<pair<vector<string>, vector<string> > > all_input;

    // Note: the first index is trg
    WORDALIGN_COMMON::Parameter_Freq parameter_trg_freq;
    WORDALIGN_COMMON::Parameter_Freq parameter_src_freq;
    WORDALIGN_COMMON::Parameter_Lex parameter_lex;
    WORDALIGN_COMMON::Parameter_Jump parameter_jump;
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;

    init_data_once(
		   //
		   &all_input,
		   &parameter_trg_freq,
		   &parameter_src_freq,
		   &parameter_lex,
		   &parameter_jump,
		   &parameter_lambda
		   );

    cerr << endl;
    cerr << "model: ibm1" << endl;
    ibm1_training(all_input, 
		  parameter_lex,
		  parameter_lambda
		  //
		  );

    // freq
    parameter_trg_freq.dump(cout, "trg-freq");
    cout << endl;

    parameter_src_freq.dump(cout, "src-freq");
    cout << endl;

    // jump
    parameter_jump.normalize();
    parameter_jump.dump(cout, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
    cout << endl;
    
    // lambda
    parameter_lambda.normalize(parameter_trg_freq);
    parameter_lambda.dump(parameter_trg_freq, parameter_src_freq, WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH, cout);
    cout << endl;
    
    return;
  }

  void Ibm1_Classic::reduce_common(
				   bool fertility_model,
				   //
				   WORDALIGN_COMMON::Parameter_Freq* parameter_trg_freq,
				   WORDALIGN_COMMON::Parameter_Freq* parameter_src_freq,
				   WORDALIGN_COMMON::Parameter_Lex* parameter_lex,
				   WORDALIGN_COMMON::Parameter_Jump* parameter_jump,
				   WORDALIGN_COMMON::Parameter_Lambda* parameter_lambda
				   ) {
    
    int line_count = 0;
    string line;
    while (UTIL::getline(cin, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, "\t", &v);
      if (v.size() == 0) {
	continue;
      }

      ++line_count;
      if (line_count % 100000 ==0) {
	cerr << ".";
      }

      if (v[0] == "trg-freq") {

	assert(v.size() == 3);
	const string& trg = v[1];
	int freq = atoi(v[2].c_str());

	// freq
	// Note: there are redundant (trg, freq) pairs
	// remove redudant!!
	double tmp = parameter_trg_freq->get(trg);
	if (tmp != 0) {
	  assert(tmp == freq);
	} else {
	  parameter_trg_freq->add(trg, freq);
	}

      } else if (v[0] == "src-freq") {

	assert(v.size() == 3);
	const string& src = v[1];
	int freq = atoi(v[2].c_str());

	// freq
	// Note: there are redundant (src, freq) pairs
	// remove redudant!!
	double tmp = parameter_src_freq->get(src);
	if (tmp != 0) {
	  assert(tmp == freq);
	} else {
	  parameter_src_freq->add(src, freq);
	}

      } else if (v[0] == "lex") {
	assert(v.size() == 4);
	const string& trg = v[1];
	const string& src = v[2];
	double value = atof(v[3].c_str());
	
	// lex
	// Note: there are redundant (trg, src, parameter) pairs
	parameter_lex->add(trg, src, value);

      } else if (v[0] == "jump") {

	assert(v.size() == 4);
	int type = atoi(v[1].c_str());
	int width = atoi(v[2].c_str());
	double value = atof(v[3].c_str());
	
	// Jump
	// Note: there are redundant (width, parameter) pairs
	if (type == 0) {
	  parameter_jump->add2(WORDALIGN_COMMON::FROM_BOS, width, value);
	} else if (type == 1) {
	  parameter_jump->add2(WORDALIGN_COMMON::OTHERS, width, value);
	} else if (type == 2) {
	  parameter_jump->add2(WORDALIGN_COMMON::TO_EOS, width, value);
	} else {
	  assert(false);
	}

      } else if (v[0] == "lambda") {

	assert(v.size() == 5);
	const string& trg = v[1];
	int fert = atof(v[2].c_str());
	int shared_by = atof(v[3].c_str());
	double value = atof(v[4].c_str());

	if (trg == "_empty_word_") {
	  assert(trg != "");
	}
	
	// lambda
	// Note: there are redundant (width, parameter) pairs
	if (fertility_model) {
	  parameter_lambda->add(trg, fert, shared_by, value);
	} else {
	  assert(fert == 1);
	  assert(shared_by == 1);
	  parameter_lambda->set(trg, value);
	}

      } else if (v[0] == "length_trg") {
      } else if (v[0] == "length_src") {
      } else {
	assert(false);
      }
    }  // while
    cerr << endl;

    // freq
    parameter_trg_freq->dump(cout, "trg-freq");
    cout << endl;

    parameter_src_freq->dump(cout, "src-freq");
    cout << endl;

    // lex
    parameter_lex->normalize();
    parameter_lex->dump(cout, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
    cout << endl;

    // jump
    parameter_jump->normalize();
    parameter_jump->dump(cout, WORDALIGN_COMMON::SMALL_JUMP_SMOOTH);
    cout << endl;

    // lambda
    parameter_lambda->normalize(*parameter_trg_freq);
    parameter_lambda->dump(*parameter_trg_freq, *parameter_src_freq, WORDALIGN_COMMON::SMALL_LAMBDA_SMOOTH, cout);
    cout << endl;

    return;
  }

  void Ibm1_Classic::reduce() {

    WORDALIGN_COMMON::Parameter_Freq parameter_trg_freq;
    WORDALIGN_COMMON::Parameter_Freq parameter_src_freq;
    WORDALIGN_COMMON::Parameter_Lex parameter_lex;
    WORDALIGN_COMMON::Parameter_Jump parameter_jump;
    WORDALIGN_COMMON::Parameter_Lambda parameter_lambda;
    
    IBM1_CLASSIC::Ibm1_Classic::reduce_common(false,
					      &parameter_trg_freq,
					      &parameter_src_freq,
					      &parameter_lex,
					      &parameter_jump,
					      &parameter_lambda);

    // if (MYCONFIG::my_global::FLAGS_debug) {
    if (true) {
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      parameter_lex.get_topk(WORDALIGN_COMMON::TOPK, &my_topk);
      
      cerr << "topk (lex) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	cerr << p.first << "\t" << "p(" << p.second << ")" << endl;
	my_topk.pop();
      }
      cerr << endl;
    }

    if (!MYCONFIG::my_global::FLAGS_training_only) {
      // decode and eval
      {
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result;
	const string str1 = "ibm1-ibm1";
	// ibm1_decode_new(parameter_lex, parameter_fert, parameter_p0p1, str1, loop, &result);
	ibm1_decode_new(parameter_lex, str1, &result);
	WORDALIGN_COMMON::Eval::eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result, "ibm1", MYCONFIG::my_global::FLAGS_loop_);
	cerr << "INFO: IBM1 decoding done at ";
	// system("date >> stderr");
	cerr << endl;
      }
    }

    return;
  }

  void Ibm1_Classic::compute_expected_counts(
					     const vector<pair<vector<string>, vector<string> > >& all_input,
					     const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
					     //
					     unordered_map<string, unordered_map<string, double> >* total_expected_count_lex
					     ) {
      
    int line_count = 0;
    const int LINE_COUNT = all_input.size();
    int STEP = LINE_COUNT / 100;
    if (STEP == 0) {
      STEP = 1;
    }
    
    for (vector<pair<vector<string>, vector<string> > >::const_iterator it = all_input.begin(); it != all_input.end(); ++it) {
      
      if (MYCONFIG::my_global::FLAGS_hadoop_) {
	cerr << "reporter:status:" << line_count << "/" << LINE_COUNT << endl;
      } else {
	if (line_count % STEP == 0) {
	  cerr << ".";
	}
      }
      ++line_count;
      
      const vector<string>& v_trg = it->first;
      const vector<string>& v_src = it->second;

      // DP is better than Gibbs
      for (int j = 0; j < (int)v_src.size(); ++j) {
	const string& src = v_src[j];
	// compute sum
	double sum = 0;
	for (int i = 0; i < (int)v_trg.size(); ++i) {
	  const string& trg = v_trg[i];
	  double parameter = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	  
	  sum += parameter;
	}

	assert(sum > 0);

	// compute expected counts
	for (int i = 0; i < (int)v_trg.size(); ++i) {
	  const string& trg = v_trg[i];
	  double parameter = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	  

	  (*total_expected_count_lex)[trg][src] += parameter / sum;
	}
      }
    }  // all_pairs
    // cerr << "training done" << endl;
    cerr << endl;
    
    return;
  }
  
  void Ibm1_Classic::ibm1_training(const vector<pair<vector<string>, vector<string> > >& all_input,
				   const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				   const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda
				   //
				   ) {
    
    // compute expected counts by Gibbs
    unordered_map<string, unordered_map<string, double> > total_expected_count_lex;

    compute_expected_counts(all_input, parameter_lex, &total_expected_count_lex);

    // dump expected counts for parameter_lex
    {
      char str[256];
      if (MYCONFIG::my_global::FLAGS_output_prob) {
	cerr << "output to " << str << " ..." << endl;
      }
      
      for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = total_expected_count_lex.begin(); it1 != total_expected_count_lex.end(); ++it1) {
	const string& trg = it1->first;
	const unordered_map<string, double>& tmp = it1->second;
	for (unordered_map<string, double>::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	  const string& src = it2->first;
	  double count = it2->second;
	  cout << "lex" << "\t" << trg << "\t" << src << "\t" << count << endl;
	}
      }
    }

    return;
  }
  
  void Ibm1_Classic::ibm1_decode_new(const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				const string& str,
				//
				unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result) {
    

    ifstream in(MYCONFIG::my_global::FLAGS_test_file.c_str());
    // ifstream in("/p/mt2/users/shaojun/acl2010/chi-eng/zhao/acl2010.test");

    char afile[1024];
    sprintf(afile, "%s-%d-%s.afile", MYCONFIG::my_global::FLAGS_project_id.c_str(), MYCONFIG::my_global::FLAGS_loop_, str.c_str());
    ofstream out(afile);

    string line;
    int line_count = 0;
    while (UTIL::getline(in, &line)) {

      vector<string> v_trg, v_src;
      bool b = WORDALIGN_COMMON::Wordalign::process_one_line(line, 0, 0, &v_trg, &v_src);
      if (!b) {
	continue;
      }
      
      ++line_count;

      assert(v_src[0] == MYCONFIG::my_global::FLAGS_BOS);
      assert(v_src[v_src.size() - 1] == MYCONFIG::my_global::FLAGS_EOS);

      // decoding loop
      for (int j = 1; j < (int)v_src.size() - 1; ++j) {
	const string& src = v_src[j];
	double max_prob = -1;
	int max_i = -1;
	for (int i = 0; i < (int)v_trg.size(); ++i) {
	  const string& trg = v_trg[i];
	  double prob = parameter_lex.get(trg, src, WORDALIGN_COMMON::SMALL_LEX_SMOOTH);
	  if (prob > max_prob) {
	    max_prob = prob;
	    max_i = i;
	  }
	}
	if (max_i == -1) {
	  assert(max_prob == -1);
	  continue;
	} else if (v_trg[max_i] == MYCONFIG::my_global::FLAGS_empty_word) {
	  continue;
	}

	// unordered_map<int, unordered_set<int> >& tmp = (*result)[line_count];;
	if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	  // reverse
	  out << line_count << " " << j << " " << max_i << " S" << endl;
	} else {
	  // normal
	  out << line_count << " " << max_i << " " << j << " S" << endl;
	}
	(*result)[line_count][max_i][j] = true;
      }  // for each j
    }  // for each line
    in.close();
    out.close();

    return;
  }
  void Ibm1_Classic::print_top_translations(
					    const string& trg,
					    const WORDALIGN_COMMON::Parameter_Lex& parameter_lex
					    ) {
    
    // output some examples

    priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
    parameter_lex.get_topk(WORDALIGN_COMMON::TOPK2, &my_topk);
  
    // we what to output the high probability words first.
    stack<pair<double, string> > s;
    while (!my_topk.empty()) {
      pair<double, string> p = my_topk.top();
      s.push(p);
      my_topk.pop();
    }
    while (!s.empty()) {
      pair<double, string> p = s.top();
      cerr << "\"" << p.second << "\"" << " (" << p.first << ") ";
      s.pop();
    }
    
    cerr << endl;

    return;
  }

}  // namespace
