
// ibm1-classic.h
#include <vector>
#include <cstdlib>
#include <iostream>
#include "util/unordered-util.h"
#include "wordalign-common/prob-table-util.h"

using namespace std;

namespace IBM1_CLASSIC {

  /*
    A simple fertility model that adds a prior for the fertility distribution.
    It assums *uniform* fertility probability.
    It assumes *uniform* distortion probability.
  */
  class Ibm1_Classic {
    
  public:
    Ibm1_Classic();
    ~Ibm1_Classic();
    
    void map();
    void reduce();

    static bool init_data_once(
			       //
			       vector<pair<vector<string>, vector<string> > >* all_input,
			       WORDALIGN_COMMON::Parameter_Freq* parameter_trg_freq,
			       WORDALIGN_COMMON::Parameter_Freq* parameter_src_freq,
			       WORDALIGN_COMMON::Parameter_Lex* parameter_lex,
			       WORDALIGN_COMMON::Parameter_Jump* parameter_jump,
			       WORDALIGN_COMMON::Parameter_Lambda* parameter_lambda
			       );

    void compute_expected_counts(
				 const vector<pair<vector<string>, vector<string> > >& all_input,
				 const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
				 //
				 unordered_map<string, unordered_map<string, double> >* total_expected_count_lex
				 );

    void ibm1_training(const vector<pair<vector<string>, vector<string> > >& all_input,
		       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex,
		       const WORDALIGN_COMMON::Parameter_Lambda& parameter_lambda
		       //
		       );

    static void reduce_common(
			      bool fertility_model,
			      //
			      WORDALIGN_COMMON::Parameter_Freq* parameter_trg_freq,
			      WORDALIGN_COMMON::Parameter_Freq* parameter_src_freq,
			      WORDALIGN_COMMON::Parameter_Lex* parameter_lex,
			      WORDALIGN_COMMON::Parameter_Jump* parameter_jump,
			      WORDALIGN_COMMON::Parameter_Lambda* parameter_lambda
			      );
    
    // *_new does not call system(...)
    static void ibm1_decode_new(const WORDALIGN_COMMON::Parameter_Lex& parameter_lex, 
				const string& str,
				//
				unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result);

    static void print_top_translations(
				       const string& trg,
				       const WORDALIGN_COMMON::Parameter_Lex& parameter_lex
				       );
  };

}
