
// fet1-main.cc

#include "config/config.h"
#include "ibm1-fertility/ibm1-fertility.h"
#include "config/global.h"

using namespace IBM1_FERTILITY;
using namespace MYCONFIG;

int main(int argc, char* argv[]) {

  if (argc <= 3) {
    cerr << "usage: " << argv[0] << " m/r loop multiple-config.ini-files < input > output" << endl;
    exit(-1);
  }

  bool mapper = true;
  if (string(argv[1]) == "m") {
    mapper = true;
  } else {
    assert(string(argv[1]) == "r");
    mapper = false;
  }

  MYCONFIG::my_global::FLAGS_loop = atoi(argv[2]);  

  for (int i = 3; i < argc; ++i) {
    const string& config_file = argv[i];

    ConfigFile config(config_file);
    my_global::read_config(&config);
  }

  // cerr << "train: " << my_global::FLAGS_train_file.c_str() << endl;
  cerr << "test: " << my_global::FLAGS_test_file.c_str() << endl;
  cerr << "a: " << my_global::FLAGS_a_file.c_str() << endl;

  /*
    output
  */

  cerr << endl;
  cerr << "------------" << endl;
  system("date");
  cerr << "[message-2010-06-09-07:38] using Prob_Lex" << endl;
  cerr << "[message-2010-05-17-17:38] adding upoisson" << endl;
  cerr << "------------" << endl;
  cerr << endl;

  // 1
  cerr << "sampling-type: " << MYCONFIG::my_global::FLAGS_sampling_type << endl;
  cerr << "scale-value: " << MYCONFIG::my_global::FLAGS_scale_value << endl;

  // 2
  assert(MYCONFIG::my_global::FLAGS_distortion != "");

  bool typical = false;
  if (MYCONFIG::my_global::FLAGS_wa_model == "ibm1") {
    cerr << "model: model-ibm1" << endl;
    typical = true;
  } else if (MYCONFIG::my_global::FLAGS_wa_model == "prior") {
    cerr << "model: model-prior" << endl;
    typical = true;
  } else if (MYCONFIG::my_global::FLAGS_wa_model == "poisson") {
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(MYCONFIG::my_global::FLAGS_scale == false);
      if (MYCONFIG::my_global::FLAGS_distortion == "uniform") {
	cerr << "model: model-sam" << endl;
	typical = true;
      }
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 1) {
      assert(MYCONFIG::my_global::FLAGS_scale == true);
      if (MYCONFIG::my_global::FLAGS_distortion == "empty-word") {
	cerr << "model: model-dan" << endl;
	typical = true;
      }
    }
    // end of poisson
  } else if (MYCONFIG::my_global::FLAGS_wa_model == "upoisson") {
    assert(MYCONFIG::my_global::FLAGS_number_of_empty_words == 1);
    assert(MYCONFIG::my_global::FLAGS_scale == true);
    cerr << "model: model-dan-upoisson" << endl;
    typical = true;
    // end of poisson
  } else { 
    assert(false);
  }
  if (!typical) {
    cerr << "model: not typical" << endl;
  }
  cerr << endl;

  // details
  cerr << "Model: " << MYCONFIG::my_global::FLAGS_wa_model << endl;
  cerr << "double_factorial: " << MYCONFIG::my_global::FLAGS_double_factorial << endl;
  cerr << "freq-word-count: " << MYCONFIG::my_global::FLAGS_freq_word_count << endl;
  cerr << "number-of-empty-words: " << MYCONFIG::my_global::FLAGS_number_of_empty_words << endl;
  cerr << "scale: " << MYCONFIG::my_global::FLAGS_scale << endl;
  cerr << "scale-value: " << MYCONFIG::my_global::FLAGS_scale_value << endl;
  cerr << "distortion: " << MYCONFIG::my_global::FLAGS_distortion << endl;
  cerr << endl;
  cerr << "sample: " << MYCONFIG::my_global::FLAGS_sampling_type << endl;
  cerr << "N_training_re_start: " << Ibm1_Fertility::N_training_re_start << endl;
  cerr << "N_training: " << Ibm1_Fertility::N_training << endl;
  cerr << "N_decoding_re_start: " << Ibm1_Fertility::N_decoding_re_start << endl;
  cerr << "N_decoding: " << Ibm1_Fertility::N_decoding << endl;
  cerr << "SMALL_LEX_SMOOTH: " << WORDALIGN_COMMON::SMALL_LEX_SMOOTH << endl;
  cerr << "SMALL_FERT_SMOOTH: " << WORDALIGN_COMMON::SMALL_FERT_SMOOTH << endl;

  /*
    processing
  */

  Ibm1_Fertility ibm1_fertility;
  if (mapper) {
    ibm1_fertility.map(cin, cout);
  }

  return 0;
}
