
// fert1.cc

#include "config/global.h"
#include "ibm1-classic/ibm1-classic.h"
#include "ibm1-fertility/ibm1-fertility.h"
#include "wordalign-common/eval.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>

using namespace std;

namespace IBM1_FERTILITY {

  const int Ibm1_Fertility::TOPK = 20;
  const int Ibm1_Fertility::TOPK2 = 8;

  int Ibm1_Fertility::N_training_re_start = 3;  // 3 is too small, and 12 is good
  int Ibm1_Fertility::N_training = 10;
  int Ibm1_Fertility::N_decoding_re_start = 12;  // 3 is too small, and 12 is good
  int Ibm1_Fertility::N_decoding = 100;

  // important, for debugging
  const bool VITERBI_INIT = true;  // init alignments for Gibbs
  const bool RUN_SRAND = false;

  Ibm1_Fertility::Ibm1_Fertility() {
    if (MYCONFIG::my_global::FLAGS_sampling_type == "for-toy-debugging") {
      N_training_re_start = 10;
      N_training = 1000;
      N_decoding_re_start = 10;
      N_decoding = 1000;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "quick") {
      N_training_re_start = 2;
      N_training = 5;
      N_decoding_re_start = 2;
      N_decoding = 5;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "hmm") {
      // N_training_re_start = 2;
      // N_training = 5;
      N_decoding_re_start = 2;
      N_decoding = 5;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "final") {
      N_training_re_start = 2;
      N_training = 5;
      // N_decoding_re_start = 2;
      // N_decoding = 5;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "standard") {

      // 

    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "improved") {

      N_training_re_start *= 4;

    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "long") {
      N_training_re_start *= 5;
      N_training *= 10;
      N_decoding_re_start *= 20;
      N_decoding *= 20;
    } else if (MYCONFIG::my_global::FLAGS_sampling_type == "longest") {
      N_training_re_start *= 10;
      N_training *= 20;
      N_decoding_re_start *= 50;
      N_decoding *= 100;
    } else {
      assert(false);
    }
  }

  Ibm1_Fertility::~Ibm1_Fertility() {
  }

  void Ibm1_Fertility::map(istream& in, ostream& out) {
    
    vector<pair<vector<string>, vector<string> > > all_input;

    // Note: the first index is trg
    WORDALIGN_COMMON::Prob_Lex prob_lex;
    WORDALIGN_COMMON::Prob_Jump prob_jump;
    unordered_map<string, double> freq_lambda;
    double lambda_non_empty_word = -1;
    double lambda_empty_word = -1;

    IBM1_CLASSIC::Ibm1_Classic::init_data_once(
					       //
					       &all_input,
					       &prob_lex,
					       &prob_jump,
					       &freq_lambda,
					       &lambda_non_empty_word,
					       &lambda_empty_word
					       );

    assert(lambda_non_empty_word != -1);
    assert(lambda_empty_word != -1);

    cerr << endl << endl;
    cerr << "model: ibm1_fertility" << endl;

    Gibbs_training_and_testing(all_input, 
			       0, // to be removed
			       0,  // to be removed
			       4, // to be removed
			       N_training,
			       N_decoding,
			       //
			       &prob_lex, 
			       &freq_lambda,
			       &lambda_non_empty_word,
			       &lambda_empty_word
			       );
    
    return;
  }

  void Ibm1_Fertility::Gibbs_training_and_testing(const vector<pair<vector<string>, vector<string> > >& all_input,
					 int round,
					 int LOOP_FROM,
					 int LOOP_TO,
					 int N_training,
					 int N_decoding,
					 //
					 WORDALIGN_COMMON::Prob_Lex* prob_lex,
					 unordered_map<string, double>* freq_lambda,
					 double* lambda_non_empty_word,
					 double* lambda_empty_word
					 ) {

    for (int loop = LOOP_FROM; loop < LOOP_TO; ++loop) {
      // compute expected counts by Gibbs
      unordered_map<string, unordered_map<string, double> > total_expected_count_lex;
      unordered_map<string, unordered_map<int, double> > total_expected_count_fert;

      int line_count = 0;
      const int LINE_COUNT = all_input.size();
      int STEP = LINE_COUNT / 100;
      if (STEP == 0) {
	STEP = 1;
      }
      
      for (vector<pair<vector<string>, vector<string> > >::const_iterator it = all_input.begin(); it != all_input.end(); ++it) {
	
	if (line_count % STEP == 0) {
	  cerr << ".";
	}
	++line_count;

	// for decoding
	unordered_map<int, int> v_max_align;
	unordered_map<int, unordered_set<int> > v_max_fert;
	double max_log_viterbi = 0;

	Gibbs_one_sentence(*it, 
			   *prob_lex, 
			   *freq_lambda,
			   *lambda_non_empty_word,
			   *lambda_empty_word,
			   N_training,
			   false,
			   //
			   &v_max_align, &v_max_fert, &max_log_viterbi,
			   &total_expected_count_lex, 
			   &total_expected_count_fert
			   );
	
      }  // for
      cerr << endl;

      if (MYCONFIG::my_global::FLAGS_debug) {
	cerr << "total expected counts" << endl;
	for (unordered_map<string, unordered_map<string, double> >::iterator it1 = total_expected_count_lex.begin(); it1 != total_expected_count_lex.end(); ++it1) {
	  const string& trg = it1->first;
	  unordered_map<string, double>& map_tmp = it1->second;
	  for (unordered_map<string, double>::iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	    const string& src = it2->first;
	    double& count = it2->second;
	    cerr << count << "\t" << trg << "\t" << src << endl;
	  }
	}
	cerr << endl;
      }
    
      /*
	update_parameters
      */
      update_parameters(total_expected_count_lex,
			total_expected_count_fert,
			//
			prob_lex,
			freq_lambda,
			lambda_non_empty_word,
			lambda_empty_word
			);
      // eval
      if (!MYCONFIG::my_global::FLAGS_training_only) {
	// decode and eval

	const string str1 = "ibm1_fertility-ibm1_fertility-1";
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result1;
	Gibbs_decode_new(*prob_lex, *freq_lambda, *lambda_non_empty_word, *lambda_empty_word, N_decoding, round, loop, str1, &result1);
	WORDALIGN_COMMON::Eval::eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result1, "fert", loop);
	cerr << "INFO: Gibbs decoding done at ";
	system("date");

	/*
	const string str2 = "ibm1_fertility-ibm1_fertility-2";
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result2;
	Gibbs_decode_new(prob_lex, prob_fert, prob_p0p1, 2 * N_decoding, round, loop, str2, &result2);
	eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result2, loop);
	cerr << "INFO: Gibbs decoding done at ";
	system("date");

	const string str3 = "ibm1_fertility-ibm1_fertility-3";
	unordered_map<int, unordered_map<int, unordered_map<int, bool> > > result3;
	Gibbs_decode_new(prob_lex, prob_fert, prob_p0p1, 4 * N_decoding, round, loop, str3, &result3);
	eval_new(MYCONFIG::my_global::FLAGS_a_file.c_str(), result3, loop);
	cerr << "INFO: Gibbs decoding done at ";
	system("date");
	*/

	cerr << endl;
      }
    }  // loop

    return;
  }

  // Gibbs sampling
  void Ibm1_Fertility::Gibbs_one_sentence(const pair<vector<string>, vector<string> >& v_trg_src,
				 const WORDALIGN_COMMON::Prob_Lex& prob_lex,
				 const unordered_map<string, double>& freq_lambda,
				 double lambda_non_empty_word,
				 double lambda_empty_word,
				 int N,
				 bool decoding,
				 //
				 unordered_map<int, int>* v_max_align,
				 unordered_map<int, unordered_set<int> >* v_max_fert,
				 double* max_log_viterbi,
				 unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
				 unordered_map<string, unordered_map<int, double> >* total_expected_count_fert
) {

    // the un-normalized probability of the bitext sentence
    double data_uprob = 0;
    // for expected counts
    unordered_map<string, unordered_map<string, double> > expected_count_lex;
    unordered_map<string, unordered_map<int, double> > expected_count_fert;
    unordered_map<string, double> expected_count_e;

    const vector<string>& v_trg = v_trg_src.first;
    const vector<string>& v_src = v_trg_src.second;
    assert(v_trg.size() > 1);
    assert(v_src.size() >= 1);

    // init v_align and v_fert
    unordered_map<int, unordered_set<int> > v_fert;
    unordered_map<int, int> v_align;

    for (int j = 0; j < (int)v_src.size(); ++j) {

      // empty word locates at pos
      int max_align = v_trg.size() - 1;
      assert(v_trg[max_align] == MYCONFIG::my_global::FLAGS_empty_word);

      if (!VITERBI_INIT) {
	// random
	int rand_total = v_trg.size();
	assert(rand_total >= 2);
	max_align = rand() % rand_total;
      } else {
	// viterbi
	
	// update to Viterbi alignments instead of empty word alignments
	const string& src = v_src[j];
	double max_prob = 0;
	for (int i = 0; i < (int)v_trg.size(); ++i) {
	  const string& trg = v_trg[i];

	  double prob = prob_lex.get(trg, src);
	  if (prob > max_prob) {
	    max_prob = prob;
	    max_align = i;
	  }
	}
      }

      v_align[j] = max_align;
      v_fert[max_align].insert(j);
    }

    // make sure the fertility set is not "empty"
    for (int i = 0; i < (int)v_trg.size(); ++i) {
      if (v_fert[i].size() == 0) {
	unordered_set<int> tmp;
	v_fert[i] = tmp;
      }
    }

    // init viterbi alignment
    *v_max_align = v_align;
    *v_max_fert = v_fert;

    double log_viterbi = get_log_viterbi(v_trg, 
					 v_src, 
					 v_align,
					 v_fert,
					 prob_lex,
					 freq_lambda,
					 lambda_non_empty_word,
					 lambda_empty_word
					 );
    
    // init max_log_viterbi
    if (*max_log_viterbi == 0) {
      *max_log_viterbi = log_viterbi;
    }

    int RE_START = N_training_re_start;
    if (decoding) {
      RE_START = N_decoding_re_start;
    }

    if (RUN_SRAND) {
      srand(0);
    }

    for (int i = 0; i < RE_START; ++i) {
      for (int n = 0; n < N; ++n) {
	Gibbs_one_move_new(v_trg, v_src, 
			   prob_lex, freq_lambda, lambda_non_empty_word, lambda_empty_word,
			   i,
			   n,
			   // (n > N/10),
			   (n >= 0),
			   decoding,
			   //
			   v_max_align, v_max_fert, 
			   max_log_viterbi, &log_viterbi,
			   &v_align, &v_fert, 
			   &expected_count_lex, 
			   &expected_count_fert, 
			   // expected_count_e,
			   &data_uprob);
      }
    }

    if (MYCONFIG::my_global::FLAGS_debug) {
      // lex
      cerr << endl;
      for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = expected_count_lex.begin(); it1 != expected_count_lex.end(); ++it1) {
	const string& trg = it1->first;
	const unordered_map<string, double>& map_tmp = it1->second;
	
	for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	  const string& src = it2->first;
	  double count = it2->second;
	  cerr << count << "\t" << trg << "\t" << src << endl;
	}      
      }
      cerr << endl;
    }

    // update total_expected_count_lex
    for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = expected_count_lex.begin(); it1 != expected_count_lex.end(); ++it1) {
      const string& trg = it1->first;
      const unordered_map<string, double>& map_tmp = it1->second;
      unordered_map<string, double>& map_tmp_total = (*total_expected_count_lex)[trg];
      
      for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	const string& src = it2->first;
	double count = it2->second;
	
	map_tmp_total[src] += count / data_uprob;
      }
    }

    // update total_expected_count_fert
    for (unordered_map<string, unordered_map<int, double> >::const_iterator it1 = expected_count_fert.begin(); it1 != expected_count_fert.end(); ++it1) {
      const string& trg = it1->first;
      // assert(trg != MYCONFIG::my_global::FLAGS_empty_word);

      const unordered_map<int, double>& map_tmp = it1->second;
      unordered_map<int, double>& map_tmp_total = (*total_expected_count_fert)[trg];
      
      for (unordered_map<int, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	int fert = it2->first;
	double count = it2->second;

	// assert(fert > 0);  // why did I have this?
	map_tmp_total[fert] += count / data_uprob;
      }
    }

    return;
  }
  
  void Ibm1_Fertility::Gibbs_decode_new(const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			       const unordered_map<string, double>& freq_lambda, 
			       double lambda_non_empty_word, 
			       double lambda_empty_word, 
			       int N_decoding,
			       int round,
			       int loop,
			       const string& str,
			       //
			       unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result) {

    ifstream in(MYCONFIG::my_global::FLAGS_test_file.c_str());

    char afile[1024];
    sprintf(afile, "%s-%d-%d-%s.afile", MYCONFIG::my_global::FLAGS_project_id.c_str(), round, loop, str.c_str());
    ofstream out(afile);

    string line;
    int line_count = 0;
    while (UTIL::getline(in, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, "\t", &v);
      assert(v.size() == 2);
      
      ++line_count;
      
      if (line_count % 2 == 0) {
	cerr << ".";
      }

      // change trg/v[0] to lower case
      UTIL::to_lower(&v[0]);
	
      vector<string> v_trg00, v_src00;
      if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	// reverse
	UTIL::split_string_using(v[1], " ", &v_trg00);
	UTIL::split_string_using(v[0], " ", &v_src00);
      } else {  
	// normal
	UTIL::split_string_using(v[0], " ", &v_trg00);
	UTIL::split_string_using(v[1], " ", &v_src00);
      }
      
      vector<string> v_trg0;
      vector<string> v_src0;
      
      // add <s> and </s> to the source sentence
      v_src0.push_back(MYCONFIG::my_global::FLAGS_BOS);
      for (int i = 0; i < (int)v_src00.size(); ++i) {
	v_src0.push_back(v_src00[i]);
      }
      v_src0.push_back(MYCONFIG::my_global::FLAGS_EOS);

      // add empty words
      int TMP = v_trg0.size();
      if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
	TMP = 0;
      } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 1) {
	TMP = 1;
      } else {
	assert(MYCONFIG::my_global::FLAGS_number_of_empty_words == 2);
	assert(TMP == (int)v_trg0.size());
      }
      for (int tmp = 0; tmp < TMP; ++tmp) {
	v_trg0.push_back(MYCONFIG::my_global::FLAGS_empty_word);
      }
      // assert(v_trg0.size() >= 1);
      // assert(v_src0.size() >= 1);

      // decode one sentence
      unordered_map<string, unordered_map<string, double> > total_expected_count_lex;
      unordered_map<string, unordered_map<int, double> > total_expected_count_fert;
      
      // for decoding
      unordered_map<int, int> v_max_align;
      unordered_map<int, unordered_set<int> > v_max_fert;
      double max_viterbi = 0;
      
      Gibbs_one_sentence(make_pair(v_trg0, v_src0), 
			 prob_lex, 
			 freq_lambda,
			 lambda_non_empty_word,
			 lambda_empty_word,
			 N_decoding,
			 true,
			 //
			 &v_max_align, &v_max_fert, &max_viterbi,
			 &total_expected_count_lex, 
			 &total_expected_count_fert);
      
      // loop
      for (int j = 0; j < (int)v_max_align.size(); ++j) {
	int i = v_max_align[j];
	assert(v_trg0.size() == v_max_fert.size());
	if (i + 1 == (int)v_max_fert.size()) { // ?
	  // out << 0 << "\t" << j + 1 << "\t" << 1 << endl;
	  assert(v_trg0[i] == MYCONFIG::my_global::FLAGS_empty_word);
	} else {
	  if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	    // reverse
	    out << line_count << " " << j + 1 << " " << i + 1 << " S" << endl;
	  } else {
	    // normal
	    out << line_count << " " << i + 1 << " " << j + 1 << " S" << endl;
	  }
	  (*result)[line_count][i+1][j+1] = true;
	}
      }
      // out << "end" << endl;
    }
    cerr << endl;
    in.close();
    out.close();

    return;
  }

}  // namespace
