
// ibm1-fertility.h
#include <vector>
#include <cstdlib>
#include <iostream>
#include "util/unordered-util.h"
#include "wordalign-common/prob-table-util.h"

using namespace std;

namespace IBM1_FERTILITY {

  /*
    A simple fertility model that adds a prior for the fertility distribution.
    It assums *uniform* fertility probability.
    It assumes *uniform* distortion probability.
  */
  class Ibm1_Fertility {

  public:
    static int N_training_re_start;
    static int N_training;
    static int N_decoding_re_start;
    static int N_decoding;
    
    static const int TOPK;
    static const int TOPK2;

  public:
    Ibm1_Fertility();
    ~Ibm1_Fertility();
    
    void map(istream& in, ostream& out);

    void Gibbs_training_and_testing(const vector<pair<vector<string>, vector<string> > >& all_input,
				    int round,
				    int LOOP_FROM,
				    int LOOP_TO,
				    int N_training,
				    int N_decoding,
				    //
				    WORDALIGN_COMMON::Prob_Lex* prob_lex,
				    unordered_map<string, double>* freq_lambda,
				    double* lambda_non_empty_word,
				    double* lambda_empty_word
				    );
    void Gibbs_one_sentence(const pair<vector<string>, vector<string> >& v_trg_src,
			    const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			    const unordered_map<string, double>& freq_lambda,
			    double lambda_non_empty_word,
			    double lambda_empty_word,
			    int N,
			    bool decoding,
			    //
			    unordered_map<int, int>* v_max_align,
			    unordered_map<int, unordered_set<int> >* v_max_fert,
			    double* max_log_viterbi,
			    unordered_map<string, unordered_map<string, double> >* total_expected_count_lex,
			    unordered_map<string, unordered_map<int, double> >* total_expected_count_fert);
    void Gibbs_one_move_new(
			    const vector<string>& v_trg, const vector<string>& v_src,
			    const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			    const unordered_map<string, double>& freq_lambda,
			    double lambda_non_empty_word,
			    double lambda_empty_word,
			    const int loop_1,
			    const int loop_2,
			    bool save,
			    bool decoding,
			    //
			    unordered_map<int, int>* v_max_align,
			    unordered_map<int, unordered_set<int> >* v_max_fert,
			    double* max_log_viterbi,
			    double* log_viterbi,
			    unordered_map<int, int>* v_align,
			    unordered_map<int, unordered_set<int> >* v_fert,
			    unordered_map<string, unordered_map<string, double> >* expected_count_lex,
			    unordered_map<string, unordered_map<int, double> >* expected_count_fert,
			    // unordered_map<string, double>* expected_count_e,
			    double* data_uprob);
    
    void update_parameters(const unordered_map<string, unordered_map<string, double> >& total_expected_count_lex,
			   const unordered_map<string, unordered_map<int, double> >& total_expected_count_fert,
			   //
			   WORDALIGN_COMMON::Prob_Lex* prob_lex,
			   unordered_map<string, double>* freq_lambda,
			   double* lambda_non_empty_word,
			   double* lambda_empty_word
			   );

    void update_for_lex(const vector<string>& v_trg, const vector<string>& v_src,
			const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			int jj, // current j
			int old_align,
			int new_align,
			// 
			double* log_viterbi
			);

    void update_for_distortion(
			       const vector<string>& v_trg, const vector<string>& v_src,
			       int old_align,
			       int new_align,
			       int old_align_old_fert,
			       int old_align_new_fert,
			       int new_align_old_fert,
			       int new_align_new_fert,
			       //
			       double* log_viterbi
			       );

    void update_for_fertility(
			      const vector<string>& v_trg, const vector<string>& v_src,
			      const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			      const unordered_map<string, double>& freq_lambda,
			      double lambda_non_empty_word,
			      double lambda_empty_word,
			      int old_align,
			      int new_align,
			      int old_align_old_fert,
			      int old_align_new_fert,
			      int new_align_old_fert,
			      int new_align_new_fert,
			      //
			      double* log_viterbi
			      );

    void Gibbs_decode_new(const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			  const unordered_map<string, double>& freq_lambda,
			  double lambda_non_empty_word, 
			  double lambda_empty_word, 
			  int N_decoding,
			  int round,
			  int loop,
			  const string& str,
			  //
			  unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* result);

    double get_log_viterbi(
			   const vector<string>& v_trg, 
			   const vector<string>& v_src,
			   const unordered_map<int, int>& v_align,
			   const unordered_map<int, unordered_set<int> >& v_fert,
			   const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			   const unordered_map<string, double>& freq_lambda,
			   double lambda_non_empty_word,
			   double lambda_empty_word
			   );

  };

}
