
// fert1.cc

#include "config/global.h"
#include "ibm1-classic/ibm1-classic.h"
#include "ibm1-fertility/ibm1-fertility.h"
#include "util/common.h"
#include "util/file-util.h"

#include <queue>
#include <stack>

/*
  model:
  pror: adding multinormial to IBM model 1
  poisson: using Poisson distribution for each word. Reject inconsistent cases.
  upoisson: similar to poisson, but all consistent cases are equally likely, which cancels (ni)!
 */

using namespace std;

namespace IBM1_FERTILITY {

  void Ibm1_Fertility::update_for_fertility(
				   const vector<string>& v_trg, const vector<string>& v_src,
				   const WORDALIGN_COMMON::Prob_Lex& prob_lex,
				   const unordered_map<string, double>& freq_lambda,
				   double lambda_non_empty_word,
				   double lambda_empty_word,
				   int old_align,
				   int new_align,
				   int old_align_old_fert,
				   int old_align_new_fert,
				   int new_align_old_fert,
				   int new_align_new_fert,
				   //
				   double* log_viterbi
				   ) {
    
    if (MYCONFIG::my_global::FLAGS_wa_model == "ibm1") {
    } else if ((MYCONFIG::my_global::FLAGS_wa_model != "upoisson") &&
	       (MYCONFIG::my_global::FLAGS_wa_model != "prior") &&
	       (MYCONFIG::my_global::FLAGS_wa_model != "poisson")) {
      assert(false);
    } else {
      assert ((MYCONFIG::my_global::FLAGS_wa_model == "upoisson") ||
	      (MYCONFIG::my_global::FLAGS_wa_model == "prior") ||
	      (MYCONFIG::my_global::FLAGS_wa_model == "poisson"));
      
      if (old_align != new_align) {
	/*
	  remove old
	*/
	if ((MYCONFIG::my_global::FLAGS_wa_model == "upoisson") ||
	    (MYCONFIG::my_global::FLAGS_wa_model == "poisson")) {
	  double old_lambda = lambda_non_empty_word;
	  if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	    old_lambda = lambda_empty_word;
	  } else if (MYCONFIG::my_global::t_freq_[v_trg[old_align]] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	    unordered_map<string, double>::const_iterator it1 = freq_lambda.find(v_trg[old_align]);
	    assert(it1 != freq_lambda.end());
	    old_lambda = it1->second;
	  } else {
	    assert(freq_lambda.find(v_trg[old_align]) == freq_lambda.end());
	  }
	  
	  if (MYCONFIG::my_global::FLAGS_scale == true) {
	    if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	      const int size = v_trg.size() - 1;
	      assert(size >= 1);
	      old_lambda *= size;
	    }
	  }
	  
	  // update
	  *log_viterbi -= log(old_lambda);
	} 
	
	// update
	assert(old_align_old_fert == old_align_new_fert + 1);
	assert (MYCONFIG::my_global::FLAGS_double_factorial == true);
	if (MYCONFIG::my_global::FLAGS_wa_model == "upoisson") {
	  if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	    *log_viterbi -= log(v_trg.size() - 1); // -
	    *log_viterbi += log(old_align_old_fert); // +
	  }
	} else if (MYCONFIG::my_global::FLAGS_wa_model == "poisson") {
	  *log_viterbi += log(old_align_old_fert); // +
	} else {
	  assert(false);
	}

	/* 
	   add new
	*/
	if ((MYCONFIG::my_global::FLAGS_wa_model == "upoisson") ||
	    (MYCONFIG::my_global::FLAGS_wa_model == "poisson")) {
	  double new_lambda = lambda_non_empty_word;
	  if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	    new_lambda = lambda_empty_word;
	  } else if (MYCONFIG::my_global::t_freq_[v_trg[new_align]] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	    unordered_map<string, double>::const_iterator it2 = freq_lambda.find(v_trg[new_align]);
	    assert(it2 != freq_lambda.end());
	    new_lambda = it2->second;
	  } else {
	    assert(freq_lambda.find(v_trg[new_align]) == freq_lambda.end());
	  }
	  
	  if (MYCONFIG::my_global::FLAGS_scale == true) {
	    if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	      const int size = v_trg.size() - 1;
	      assert(size >= 1);
	      new_lambda *= size;
	    }
	  }
	  
	  // update
	  *log_viterbi += log(new_lambda);
	}
	
	// update
	assert(new_align_old_fert == new_align_new_fert - 1);
	assert (MYCONFIG::my_global::FLAGS_double_factorial == true);
	if (MYCONFIG::my_global::FLAGS_wa_model == "upoisson") {
	  if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	    *log_viterbi += log(v_trg.size() - 1); // +
	    *log_viterbi -= log(new_align_new_fert);  // -
	  }
	} else if (MYCONFIG::my_global::FLAGS_wa_model == "poisson") {
	  *log_viterbi -= log(new_align_new_fert); // -
	} else {
	  assert(false);
	}
      } else {
	assert(old_align_old_fert == old_align_new_fert);
	assert(new_align_old_fert == new_align_new_fert);
      }
    }

    return;
  }

  /*
   distortion (wa-model and distortion model)
   1. upoisson: uniform
   2. uniform: uniform
   3. poisson: 
     a) empty-word: give more chances to empty words
     b) phi: proportional to fert
     c) uniform: uniform
  */
  void Ibm1_Fertility::update_for_distortion(const vector<string>& v_trg, const vector<string>& v_src,
				    int old_align,
				    int new_align,
				    int old_align_old_fert,
				    int old_align_new_fert,
				    int new_align_old_fert,
				    int new_align_new_fert,
				    //
				    double* log_viterbi
				    ) {

    if (MYCONFIG::my_global::FLAGS_wa_model != "poisson") {
      return;
    }

    assert(MYCONFIG::my_global::FLAGS_wa_model == "poisson");

    if (MYCONFIG::my_global::FLAGS_distortion == "uniform") {
    } else if (MYCONFIG::my_global::FLAGS_distortion == "empty-word") {
      // proportioanl to k(e_{a_j})
      // pros: empty-word > uniform > phi
      // cons: this results in very big lambda for the empty word
      if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	const int size = v_trg.size() - 1;
	assert(size >= 1);
	assert (MYCONFIG::my_global::FLAGS_scale_value > 0);
	assert (MYCONFIG::my_global::FLAGS_scale_value <= 1);
	*log_viterbi -= log(size * MYCONFIG::my_global::FLAGS_scale_value);
      }
      if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	const int size = v_trg.size() - 1;
	assert(size >= 1);
	assert (MYCONFIG::my_global::FLAGS_scale_value > 0);
	assert (MYCONFIG::my_global::FLAGS_scale_value <= 1);
	*log_viterbi += log(size * MYCONFIG::my_global::FLAGS_scale_value);
	// *log_viterbi += log(v_trg.size() - 1);
      }
    } else if (MYCONFIG::my_global::FLAGS_distortion == "phi") {
      // proportional to \phi
      // this results in very small \lambda for freq words
      assert(v_trg.size() >= 2);

      if (v_trg[old_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	old_align_old_fert *= v_trg.size() - 1;
	old_align_new_fert *= v_trg.size() - 1;
      }

      if (v_trg[new_align] == MYCONFIG::my_global::FLAGS_empty_word) {
	new_align_old_fert *= v_trg.size() - 1;
	new_align_new_fert *= v_trg.size() - 1;
      }

      assert(old_align_old_fert > 0);
      *log_viterbi -= old_align_old_fert * log(old_align_old_fert);


      if (old_align_new_fert > 0) {
	*log_viterbi += old_align_new_fert * log(old_align_new_fert);
      }
      if (new_align_old_fert > 0) {
	*log_viterbi -= new_align_old_fert * log(new_align_old_fert);
      }
      assert(new_align_new_fert > 0);
      *log_viterbi += new_align_new_fert * log(new_align_new_fert);
    } else {
      cerr << "wrong option for distion: " << MYCONFIG::my_global::FLAGS_distortion << endl;
      assert(false);
    }

    return;
  }

  void Ibm1_Fertility::update_for_lex(const vector<string>& v_trg, const vector<string>& v_src,
			     const WORDALIGN_COMMON::Prob_Lex& prob_lex,
			     int jj, // current j
			     int old_align,
			     int new_align,
			     // 
			     double* log_viterbi
			     ) {

    double old_align_prob = prob_lex.get(v_trg[old_align], v_src[jj]);
    if (old_align_prob == 0) {
      old_align_prob = WORDALIGN_COMMON::SMALL_LEX_SMOOTH;
    }
    double new_align_prob = prob_lex.get(v_trg[new_align], v_src[jj]);
    if (new_align_prob == 0) {
      new_align_prob = WORDALIGN_COMMON::SMALL_LEX_SMOOTH;
    }
    
    *log_viterbi += log(new_align_prob);
    *log_viterbi -= log(old_align_prob);
    
    return;
  }

  void Ibm1_Fertility::Gibbs_one_move_new(const vector<string>& v_trg, const vector<string>& v_src, 
				 const WORDALIGN_COMMON::Prob_Lex& prob_lex,
				 const unordered_map<string, double>& freq_lambda,
				 double lambda_non_empty_word,
				 double lambda_empty_word,
				 const int loop_1,
				 const int loop_2,
				 bool after_burn_in,
				 bool decoding,
				 //
				 unordered_map<int, int>* v_max_align,
				 unordered_map<int, unordered_set<int> >* v_max_fert,
				 double* max_log_viterbi,
				 double* log_viterbi,
				 unordered_map<int, int>* v_align,
				 unordered_map<int, unordered_set<int> >* v_fert,
				 unordered_map<string, unordered_map<string, double> >* expected_count_lex,
				 unordered_map<string, unordered_map<int, double> >* expected_count_fert,
				 // unordered_map<string, double>* expected_count_e,
				 double* data_uprob) {
    

    // Gibbs sampling
    for (int jj = 0; jj < (int)v_src.size(); ++jj) {
      // const string& src = v_src[jj];
      const int old_align = (*v_align)[jj];
      const int old_align_old_fert = (*v_fert)[old_align].size();

      vector<double> pp;  // pp = p(a_j | not_a_j, E, F) for each a_j
      double sum = 0;
      
      for (int ii = 0; ii < (int)v_trg.size(); ++ii) {
	// const string& trg = v_trg[ii];
	const int new_align = ii;

	int new_align_old_fert = (*v_fert)[new_align].size();
	int old_align_new_fert = old_align_old_fert;
	int new_align_new_fert = new_align_old_fert;
	if (new_align != old_align) {
	  --old_align_new_fert;
	  ++new_align_new_fert;
	}

	if (MYCONFIG::my_global::FLAGS_debug) {
	  cerr << "source: " << jj << " " << v_src[jj] << endl;
	  cerr << "target: " << ii << " " << v_trg[ii] << endl;
	  cerr << "old_align: " << old_align << " " << v_trg[old_align] << endl;
	  cerr << "old_align_old_fert: " << old_align_old_fert << endl;
	  cerr << "old_align_new_fert: " << old_align_new_fert << endl;
	  cerr << "new_align: " << new_align << " " << v_trg[new_align] << endl;
	  cerr << "new_align_old_fert: " << new_align_old_fert << endl;
	  cerr << "new_align_new_fert: " << new_align_new_fert << endl;
	}
	
	// we will update this variable.
	double my_log_viterbi = *log_viterbi;

	// update lex diff
	update_for_lex(v_trg, v_src,
		       prob_lex,
		       jj,
		       old_align,
		       new_align,
		       //
		       &my_log_viterbi);

	// update distortion
	update_for_distortion(
			      v_trg, v_src,
			      old_align,
			      new_align,
			      old_align_old_fert,
			      old_align_new_fert,
			      new_align_old_fert,
			      new_align_new_fert,
			      //
			      &my_log_viterbi
			      );

	// update fert/lambda diff
	update_for_fertility(
			     v_trg, v_src,
			     prob_lex,
			     freq_lambda,
			     lambda_non_empty_word,
			     lambda_empty_word,
			     old_align,
			     new_align,
			     old_align_old_fert,
			     old_align_new_fert,
			     new_align_old_fert,
			     new_align_new_fert,
			     //
			     &my_log_viterbi
			     );

	// done with my_log_viterbi;

	if (decoding) {
	  if (my_log_viterbi > *max_log_viterbi) {
	    *max_log_viterbi = my_log_viterbi;
	    *v_max_align = *v_align;
	    *v_max_fert = *v_fert;
	    // we do not update v_align and v_fert because we may move to another state
	    // so we have to update v_max_* below
	    // update
	    (*v_max_align)[jj] = new_align;
	    // update
	    (*v_max_fert)[old_align].erase(jj);
	    (*v_max_fert)[new_align].insert(jj);
	  }
	}

	// we will move to this state proportioan to this weight
	double weight = exp(my_log_viterbi);
	pp.push_back(weight);
	sum += weight;
      }  // for ii

      if (MYCONFIG::my_global::FLAGS_debug) {
	for (int i = 0; i < (int)pp.size(); ++i) {
	  cerr << pp[i] / sum << " " << v_src[jj] << " " << v_trg[i] << " " << pp[i] << endl;
	}
      }

      // normalize, so weight become probability
      assert(pp.size() == v_trg.size());
      for (int i = 0; i < (int)pp.size(); ++i) {
	pp[i] /= sum;
      }

      // sample alignment from the conditiona distribution "pp"
      // pp = p(a_j | not_a_j, E, F) for each a_j
      double r = UTIL::rand_r();
      double sum_p = 0;
      int new_align = 0;
      for (int j = 0; j < (int)pp.size(); ++j) {
	sum_p += pp[j];
	if (sum_p >= r) {
	  new_align = j;
	  break;
	}
      }

      // remember old and new align and fert variables
      int new_align_old_fert = (*v_fert)[new_align].size();
      int old_align_new_fert = old_align_old_fert;
      int new_align_new_fert = new_align_old_fert;
      if (new_align != old_align) {
	--old_align_new_fert;
	++new_align_new_fert;
      }

      // update current alignments and fert variables
      (*v_align)[jj] = new_align;
      (*v_fert)[old_align].erase(jj);
      (*v_fert)[new_align].insert(jj);

      // verify
      assert((int)(*v_fert)[old_align].size() == old_align_new_fert);
      assert((int)(*v_fert)[new_align].size() == new_align_new_fert);

      // update lex diff
      update_for_lex(v_trg, v_src,
		     prob_lex,
		     jj,
		     old_align,
		     new_align,
		     //
		     log_viterbi);
      
      // update distortion
      update_for_distortion(
			    v_trg, v_src,
			    old_align,
			    new_align,
			    old_align_old_fert,
			    old_align_new_fert,
			    new_align_old_fert,
			    new_align_new_fert,
			    //
			    log_viterbi
			    );

      // update fert/lambda diff
      update_for_fertility(
			   v_trg, v_src,
			   prob_lex,
			   freq_lambda,
			   lambda_non_empty_word,
			   lambda_empty_word,
			   old_align,
			   new_align,
			   old_align_old_fert,
			   old_align_new_fert,
			   new_align_old_fert,
			   new_align_new_fert,
			   //
			   log_viterbi
			   );

      // done with log_viterbi

      // save this sample
      if (after_burn_in) {
	// update expected_count_lex
	for (int j = 0; j < (int)v_src.size(); ++j) {
	  const string& src = v_src[j];
	  int align = (*v_align)[j];
	  const string& trg = v_trg[align];
	  unordered_map<string, double>& expected_count_lex_tmp = (*expected_count_lex)[trg];
	  ++expected_count_lex_tmp[src];
	}
	
	// update expected_count_fert
	{
	  assert ((MYCONFIG::my_global::FLAGS_wa_model == "ibm1") ||
		  (MYCONFIG::my_global::FLAGS_wa_model == "upoisson") ||
		  (MYCONFIG::my_global::FLAGS_wa_model == "prior") ||
		  (MYCONFIG::my_global::FLAGS_wa_model == "poisson"));

	  for (int i = 0; i < (int)v_trg.size(); ++i) {
	    const string& trg = v_trg[i];
	    unordered_map<int, double>& expected_count_fert_tmp = (*expected_count_fert)[trg];
	    const unordered_set<int> set_fert = get_1(*v_fert, i);
	    int fert = set_fert.size();
	    ++expected_count_fert_tmp[fert];
	  }
	}
	
	++(*data_uprob);
      }

    }  // for jj
      
    return;
  }

  double Ibm1_Fertility::get_log_viterbi(
				const vector<string>& v_trg, 
				const vector<string>& v_src,
				const unordered_map<int, int>& v_align,
				const unordered_map<int, unordered_set<int> >& v_fert,
				const WORDALIGN_COMMON::Prob_Lex& prob_lex,
				const unordered_map<string, double>& freq_lambda,
				double lambda_non_empty_word,
				double lambda_empty_word
				) {
    // init log_viterbi
    double log_viterbi = 0;  // log (N!/n1! n2! ... nm!) \prob p(fj|e_{a_j})

    // lex
    for (int j = 0; j < (int)v_src.size(); ++j) {
      unordered_map<int, int>::const_iterator it = v_align.find(j);
      assert(it != v_align.end());
      int align = it->second;
      double prob = prob_lex.get(v_trg[align], v_src[j]);
      log_viterbi += log(prob);
    }
    
    // fert
    if (MYCONFIG::my_global::FLAGS_wa_model == "ibm1") {
      //
    } else if (MYCONFIG::my_global::FLAGS_wa_model == "prior") {
      for (int phi = 1; phi <= (int)v_src.size(); ++phi) {  // (J)!
	log_viterbi += log(phi);
      }
      for (int i = 0; i < (int)v_trg.size(); ++i) {  // (phi_i)!
	unordered_map<int, unordered_set<int> >::const_iterator it = v_fert.find(i);
	assert(it != v_fert.end());
	const unordered_set<int>& set_tmp = it->second;
	for (int phi = 1; phi <= (int)set_tmp.size(); ++phi) {
	  log_viterbi -= log(phi);
	}
      }
    } else if ((MYCONFIG::my_global::FLAGS_wa_model == "upoisson") ||
	       (MYCONFIG::my_global::FLAGS_wa_model == "poisson")) {
      
      if (MYCONFIG::my_global::FLAGS_wa_model == "poisson") {
	// distortion
	for (int j = 0; j < (int)v_src.size(); ++j) {
	  // get a_j
	  unordered_map<int, int>::const_iterator it = v_align.find(j);
	  assert(it != v_align.end());
	  int align = it->second;
	  // get \phi_i
	  unordered_map<int, unordered_set<int> >::const_iterator it2 = v_fert.find(align);
	  assert(it2 != v_fert.end());
	  const unordered_set<int>& set_tmp = it2->second;
	  const int phi_i = set_tmp.size();
	  
	  log_viterbi += log(phi_i);
	  log_viterbi -= log(v_src.size());
	}
      } else {
	// nothing for upoisson
	assert(MYCONFIG::my_global::FLAGS_wa_model == "upoisson");
      }

      // lambda
      for (int i = 0; i < (int)v_trg.size(); ++i) {  // phi_i ^ lambda_i  * e ^ (-phi_i) / phi_i !
	// retrieve phi_i
	unordered_map<int, unordered_set<int> >::const_iterator it = v_fert.find(i);
	assert(it != v_fert.end());
	const unordered_set<int>& set_tmp = it->second;
	const int phi_i = set_tmp.size();
	// retrieve lambda_i
	double lambda_i = lambda_non_empty_word;
	if (v_trg[i] == MYCONFIG::my_global::FLAGS_empty_word) {
	  lambda_i = lambda_empty_word;
	} else if (MYCONFIG::my_global::t_freq_[v_trg[i]] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	  unordered_map<string, double>::const_iterator it2 = freq_lambda.find(v_trg[i]);
	  assert(it2 != freq_lambda.end());
	  lambda_i = it2->second;
	} else {
	  assert(freq_lambda.find(v_trg[i]) == freq_lambda.end());
	}

	// scalar for the empty word's Poisson distribution
	if (MYCONFIG::my_global::FLAGS_scale == true) {
	  if (v_trg[i] == MYCONFIG::my_global::FLAGS_empty_word) {
	    const int size = v_trg.size() - 1;
	    assert(size >= 1);
	    lambda_i *= size;
	  }
	}

	// update log_viterbi
	log_viterbi += phi_i * log(lambda_i);
	log_viterbi -= lambda_i;
	for (int phi = 1; phi <= phi_i; ++phi) {
	  log_viterbi -= log(phi);
	}
      }
    } else {
      assert(false);
    }

    return log_viterbi;
  }

  void Ibm1_Fertility::update_parameters(const unordered_map<string, unordered_map<string, double> >& total_expected_count_lex,
				const unordered_map<string, unordered_map<int, double> >& total_expected_count_fert,
				//
				WORDALIGN_COMMON::Prob_Lex* prob_lex,
				unordered_map<string, double>* freq_lambda,
				double* lambda_non_empty_word,
				double* lambda_empty_word
				) {

    // expected counts to probabilities for prob_lex
    ofstream out("prob.txt");
    prob_lex->normalize_using_counts(total_expected_count_lex, out, MYCONFIG::my_global::FLAGS_output_prob);
    out.close();
    
    // expected counts to probabilities for lambda_*
    double count_lambda_non_empty_word_1 = 0;
    double count_lambda_non_empty_word_2 = 0;
    double count_lambda_empty_word_1 = 0;
    // double count_lambda_empty_word_2 = 0;
    for (unordered_map<string, unordered_map<int, double> >::const_iterator it1 = total_expected_count_fert.begin(); it1 != total_expected_count_fert.end(); ++it1) {
      const string& trg = it1->first;
      // assert(trg != MYCONFIG::my_global::FLAGS_empty_word);
      const unordered_map<int, double>& map_tmp = it1->second;
      
      {
	double sum1 = 0;
	for (unordered_map<int, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	  double count = it2->second;
	  sum1 += count;
	}
	double sum2 = WORDALIGN_COMMON::total_count_of_e_[trg];
	assert(fabs(sum1 - sum2) < 1);
      }
      
      double count_lambda_freq_1 = 0;
      double count_lambda_freq_2 = 0;
      for (unordered_map<int, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	int fert = it2->first;
	double count = it2->second;
	
	if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	  count_lambda_empty_word_1 += fert * count;
	  //
	  // this has to be collected in another way: it is the sum of all other words!
	  // count_lambda_empty_word_2 += count;
	} else {
	  count_lambda_non_empty_word_1 += fert * count;
	  count_lambda_non_empty_word_2 += count;
	  
	  if (MYCONFIG::my_global::t_freq_[trg] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	    count_lambda_freq_1 += fert * count;
	    count_lambda_freq_2 += count;
	  }
	}
      }  // for
      
      // verify
      if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	// do nothing
      } else if (MYCONFIG::my_global::t_freq_[trg] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	double sum = WORDALIGN_COMMON::total_count_of_e_[trg];
	assert(sum > 0);
	assert(count_lambda_freq_2 > 0);
	if (fabs(count_lambda_freq_2 - sum) >= 1) {
	  cerr << "trg: " << trg << "\t"
	       << "count_lambda_freq_2: " << count_lambda_freq_2 << "\t"
	       << "sum: " << sum << endl;
	}
	assert(fabs(count_lambda_freq_2 - sum) < 1);
	// expected counts to probability
	if (count_lambda_freq_1 < 1.0e-8) {
	  count_lambda_freq_1 = 1.0e-8;
	  cerr << "trg: " << trg << " has a small lambda" << endl;
	}
	if (count_lambda_freq_2 < 1.0e-8) {
	  count_lambda_freq_2 = 1.0e-8;
	  cerr << "trg: " << trg << " has a small lambda" << endl;
	}
	(*freq_lambda)[trg] = count_lambda_freq_1 / count_lambda_freq_2;
      } else {
	// do nothing
      }
    }  // for each trg
    
    // verify
    double sum_non_empty_word = 0;
    for (unordered_map<string, unordered_map<int, double> >::const_iterator it1 = total_expected_count_fert.begin(); it1 != total_expected_count_fert.end(); ++it1) {
      const string& trg = it1->first;
      if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	// do nothing
      } else {
	sum_non_empty_word += WORDALIGN_COMMON::total_count_of_e_[trg];
      }
    }  // for
    assert(fabs(sum_non_empty_word - count_lambda_non_empty_word_2) < 1);
    
    // expected counts to probability
    if (count_lambda_non_empty_word_1 < 1.0e-8) {
      count_lambda_non_empty_word_1 = 1.0e-8;
      cerr << " has a small non-freq lambda" << endl;
    }
    if (count_lambda_non_empty_word_2 < 1.0e-8) {
      count_lambda_non_empty_word_2 = 1.0e-8;
      cerr << " has a small non-freq lambda" << endl;
    }
    *lambda_non_empty_word = count_lambda_non_empty_word_1 / count_lambda_non_empty_word_2;
    
    // expected counts to probability
    if (count_lambda_empty_word_1 < 1.0e-8) {
      count_lambda_empty_word_1 = 1.0e-8;
      cerr << " has a small empty-word lambda" << endl;
    }
    // the total number of empty-word * sentece-length = total-number-of-non-empty-word
    *lambda_empty_word = count_lambda_empty_word_1 / count_lambda_non_empty_word_2;
    

    /*
     output
    */
    if (true) {
      priority_queue<pair<double, string>, vector<pair<double, pair<string, string> > >, WORDALIGN_COMMON::topk3> my_topk;
      prob_lex->get_topk(TOPK, &my_topk);

      cerr << "topk (lex) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, pair<string, string> > p = my_topk.top();
	cerr << "\"" << p.first << "\"" << "\t" << "p(" << p.second.first << " | " << p.second.second << ")" << endl;
	//
	const string& trg = p.second.second;
	// cerr << trg << "\t" << MYCONFIG::my_global::t_freq_[trg] << endl;
	if (MYCONFIG::my_global::t_freq_[trg] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	  // if (MYCONFIG::my_global::t_freq_[trg] >= 1) {
	  unordered_map<string, double>::iterator it1 = freq_lambda->find(trg);
	  assert(it1 != freq_lambda->end());
	  cerr << it1->second << "\tp(lambda | " << trg << ")" << " "
	       << MYCONFIG::my_global::t_freq_[trg] << " ";
	  IBM1_CLASSIC::Ibm1_Classic::print_top_translations(trg, *prob_lex);
	}
	my_topk.pop();
      }  // not empty
    } // true
    cerr << endl;

    bool too_small = false;
    // if (MYCONFIG::my_global::FLAGS_debug) {

    if (true) {
      // output lambda for empty word
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      // priority_queue<pair<double, string> > my_topk(topk);
      for (unordered_map<string, double>::iterator it1 = freq_lambda->begin(); it1 != freq_lambda->end(); ++it1) {
	const string& trg = it1->first;
	if (MYCONFIG::my_global::t_freq_[trg] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	  double prob = -1.0 * it1->second;		     
	  						     
	  my_topk.push(make_pair(prob, trg));		     
	  if ((int)my_topk.size() > TOPK) {		     
	    my_topk.pop();				     
	    assert((int)my_topk.size() <= TOPK);
	  }
	}
      }
      
      cerr << "topk (lambda) " << my_topk.size() << endl;
      if ((int)my_topk.size() < TOPK) {
	too_small = true;
      } else {
	assert((int)my_topk.size() == TOPK);
      }

      stack<pair<double, string> > s1;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	s1.push(p);
	my_topk.pop();
      }

      while (!s1.empty()) {
	pair<double, string> p = s1.top();
	cerr << -1 * p.first << "\tp(lambda | " << p.second << ")" << " "
	     << MYCONFIG::my_global::t_freq_[p.second] << " ";
	IBM1_CLASSIC::Ibm1_Classic::print_top_translations(p.second, *prob_lex);
	s1.pop();
      }

      cerr << endl;
      cerr << "--" << endl;
      // cerr << "p(lambda|unk) " << freq_lambda[MYCONFIG::my_global::FLAGS_unknown_word] << endl;
      cerr << "p(lambda|*) " << *lambda_non_empty_word << "\t";
      cerr << "p(lambda|eps) " << *lambda_empty_word << "\t";
      system("date");
      cerr << "--" << endl;
      cerr << endl;
    }
    
    // if (MYCONFIG::my_global::FLAGS_debug) {
    if (!too_small) {
      priority_queue<pair<double, string>, vector<pair<double, string> >, WORDALIGN_COMMON::topk> my_topk;
      // priority_queue<pair<double, string> > my_topk(topk);
      for (unordered_map<string, double>::iterator it1 = freq_lambda->begin(); it1 != freq_lambda->end(); ++it1) {
	const string& trg = it1->first;
	if (MYCONFIG::my_global::t_freq_[trg] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	  double prob = it1->second;
	  
	  my_topk.push(make_pair(prob, trg));
	  if ((int)my_topk.size() > TOPK) {
	    my_topk.pop();
	    assert((int)my_topk.size() <= TOPK);
	  }
	}
      }
      
      cerr << "topk (lambda) " << my_topk.size() << endl;
      while (!my_topk.empty()) {
	pair<double, string> p = my_topk.top();
	cerr << p.first << "\tp(lambda | " << p.second << ")" << " "
	     << MYCONFIG::my_global::t_freq_[p.second] << " ";
	IBM1_CLASSIC::Ibm1_Classic::print_top_translations(p.second, *prob_lex);
	my_topk.pop();
      }

    }

    cerr << endl;
    
    return;
  }

}  // namespace
