
#include "string-util.h"

namespace UTIL {

  void to_lower(string* str) {
    for (int i = 0; i < (int)str->size(); ++i) {
      (*str)[i] = tolower((*str)[i]);
    }
  }

  void split_string_using(const string& full,
			  const char* delim,
			  vector<string>* result) {
    back_insert_iterator< vector<string> > it(*result);
    SplitStringToIteratorUsing(full, delim, it);
  }


  void remove_redundancy(vector<string>* result) {
    vector<string> tmp;
    for (int i = 0; i < (int)result->size(); ++i) {
      // insert str to tmp only if str is not in tmp
      const string& str = (*result)[i];
      bool found = false;
      for (int j = 0; j < (int)tmp.size(); ++j) {
	if (tmp[j] == str) {
	  found = true;
	  break;
	}
      }
      if (found) {
	continue;
      } else {
	tmp.push_back(str);
      }
    }

    // swap tmp and result
    result->clear();
    assert(result->size() == 0);
    for (int j = 0; j < (int)tmp.size(); ++j) {
      result->push_back(tmp[j]);
    }
  }

  // ----------------------------------------------------------------------
  // A wrapper function for string::getline
  // ----------------------------------------------------------------------
  istream& getline( istream& is, string* s, char delimiter) {
    assert(s != NULL);
    return std::getline(is, *s, delimiter);
  }

  // ----------------------------------------------------------------------
  // SplitStringUsing()
  //    Split a string using a character delimiter. Append the components
  //    to 'result'.
  //
  // Note: For multi-character delimiters, this routine will split on *ANY* of
  // the characters in the string, not the entire string as a single delimiter.
  // ----------------------------------------------------------------------
  template <typename ITR> static inline void SplitStringToIteratorUsing(const string& full,
									const char* delim,
									ITR& result) {
    // Optimize the common case where delim is a single character.
    if (delim[0] != '\0' && delim[1] == '\0') {
      char c = delim[0];
      const char* p = full.data();
      const char* end = p + full.size();
      while (p != end) {
	if (*p == c) {
	  ++p;
	} else {
	  const char* start = p;
	  while (++p != end && *p != c);
	  *result++ = string(start, p - start);
	}
      }
      return;
    }
    
    string::size_type begin_index, end_index;
    begin_index = full.find_first_not_of(delim);
    while (begin_index != string::npos) {
      end_index = full.find_first_of(delim, begin_index);
      if (end_index == string::npos) {
	*result++ = full.substr(begin_index);
	return;
      }
      *result++ = full.substr(begin_index, (end_index - begin_index));
      begin_index = full.find_first_not_of(delim, end_index);
    }
  }
  
}  // namespace UTIL

