
#include "config/global.h"
#include "util/common.h"
#include "util/file-util.h"
#include "wordalign-common/eval.h"

namespace WORDALIGN_COMMON {

  void Eval::read_afile(
			const string& file, 
			//
			unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* key_s,
			unordered_map<int, unordered_map<int, unordered_map<int, bool> > >* key_p) {

    ifstream in(file.c_str());
    // ifstream in("/p/mt2/users/shaojun/acl2010/chi-eng/zhao/acl2010.afile");
    string line;
    while (UTIL::getline(in, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, " \t", &v);
      assert(v.size() == 4);

      if (v[1] == "0") {
	continue;
      }
      if (v[2] == "0") {
	continue;
      }

      int line_count = atoi(v[0].c_str());
      int i, j;
      if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
	// reverse
	j = atoi(v[1].c_str());
	i = atoi(v[2].c_str());
      } else {
	// normal
	i = atoi(v[1].c_str());
	j = atoi(v[2].c_str());
      }

      if (v[3] == "S") {
	(*key_s)[line_count][i][j] = true;
      } else if (v[3] == "P") {
	(*key_p)[line_count][i][j] = true;
      } else {
	assert(false);
      }
    }
    in.close();

    return;
  }

  double Eval::eval_new(const string& key_afile, const unordered_map<int, unordered_map<int, unordered_map<int, bool> > >& result, const string& prefix, int loop) {

    unordered_map<int, unordered_map<int, unordered_map<int, bool> > > key_s;  // sure
    unordered_map<int, unordered_map<int, unordered_map<int, bool> > > key_p;  // posible

    // read_afile(MYCONFIG::my_global::FLAGS_a_file.c_str(), false, &key_s, &key_p);
    read_afile(key_afile.c_str(), &key_s, &key_p);
    
    // compute aer
    int A = 0; // aligment from result
    int S = 0; // sure
    int A_S = 0; // A and S
    int A_P = 0; // A and P

    // update S
    for (unordered_map<int, unordered_map<int, unordered_map<int, bool> > >::const_iterator it_index = key_s.begin(); it_index != key_s.end(); ++it_index) {
      // int index = it_index->first;
      for (unordered_map<int, unordered_map<int, bool> >::const_iterator it_i = it_index->second.begin(); it_i != it_index->second.end(); ++it_i) {
	// int i = it_i->first;
	for (unordered_map<int, bool>::const_iterator it_j = it_i->second.begin(); it_j != it_i->second.end(); ++it_j) {
	  // int j = it_j->first;
	  assert(it_j->second == true);
	  
	  ++S;
	}
      }
    }

    // update A, A_S, A_P
    for (unordered_map<int, unordered_map<int, unordered_map<int, bool> > >::const_iterator it_index = result.begin(); it_index != result.end(); ++it_index) {
      int index = it_index->first;
      for (unordered_map<int, unordered_map<int, bool> >::const_iterator it_i = it_index->second.begin(); it_i != it_index->second.end(); ++it_i) {
	int i = it_i->first;
	for (unordered_map<int, bool>::const_iterator it_j = it_i->second.begin(); it_j != it_i->second.end(); ++it_j) {
	  int j = it_j->first;
	  assert(it_j->second == true);
	  
	  if (key_s[index][i][j] == true) {
	    ++A_S;
	    ++A_P;
	  }
	  if (key_p[index][i][j] == true) {
	    ++A_P;
	  }
	  ++A;
	}
      }
    }

    string myprefix = prefix;
    assert(myprefix != "all");
    if (myprefix == "") {
      myprefix = "all";
    }
    double aer = 1 - 1.0 * (A_S + A_P) / (A + S);
    cerr << "AER pre: " << loop << "\t" << 1.0 * A_P / A << endl;
    cerr << "AER rec: " << loop << "\t" << 1.0 * A_S / S << endl;
    cerr << "AER: " << loop << "\t" << "(" << myprefix << ")" << "\t" << aer << endl;

    ofstream out(MYCONFIG::my_global::FLAGS_log_file.c_str(), ios::app);
    if (loop == 0) {
      out << "AER:" << endl;
    }
    out << "AER pre: " << loop << "\t" << 1.0 * A_P / A << endl;
    out << "AER rec: " << loop << "\t" << 1.0 * A_S / S << endl;
    out << "AER: " << loop << "\t" << "(" << myprefix << ")" << "\t" << aer << endl;
    out << endl;
    out.close();

    return aer;
  }

  /*
  void Eval::eval(const string& str, const string& msg, const string& prefix, int loop) {

    int cmd_return;

    // eval
    char cmd_eval[256];
    sprintf(cmd_eval,
	    "%s %s %s ~/project_/zhao_/data_/test-124.afile '%s %d'",
	    "./evaluator-main",
	    "lbfgs.aer",
	    str.c_str(),
	    msg.c_str(),
	    loop);
    cmd_return = system(cmd_eval);
    if (cmd_return) {
      cerr << "FAILED: " << cmd_eval << endl;
      exit(-1);
    }
    
    return;
  }
  */

}  // namespace
