
// fet1-main.cc

#include "init.h"

#include "config/config.h"
#include "config/global.h"
#include "util/string-util.h"
#include "process-one-line.h"

namespace WORDALIGN_COMMON {
  void process(
	       unordered_map<int, int>* length_trg,
	       unordered_map<int, int>* length_src,
	       Parameter_Freq* trg_freq,
	       Parameter_Freq* src_freq,
	       Parameter_Lex* lex,
	       Parameter_Jump* jump,
	       Parameter_Lambda* lambda
	       ) {

    string line;
    int line_count = 0;
    while (UTIL::getline(cin, &line)) {
      vector<string> v;
      UTIL::split_string_using(line, "\t", &v);
      
      vector<string> v_trg;
      vector<string> v_src;
      bool b = Wordalign::process_one_line(line, WORDALIGN_COMMON::TRG_LEN_LIMIT, WORDALIGN_COMMON::SRC_LEN_LIMIT, &v_trg, &v_src);
      
      if (!b) {
	continue;
      }
      
      ++line_count;
      if (line_count % 1000 == 0) {
	cerr << ".";
      }
      
      // len
      int len_trg = v_trg.size();
      assert(len_trg >= 3);
      ++(*length_trg)[len_trg - 2];
      
      int len_src = v_src.size();
      assert(len_src >= 3);
      ++(*length_src)[len_src - 2];
      
      // freq
      for (int i = 0; i < (int)v_trg.size(); ++i) {
	const string& trg = v_trg[i];
	trg_freq->add(trg, 1);
      }

      for (int i = 0; i < (int)v_src.size(); ++i) {
	const string& src = v_src[i];
	src_freq->add(src, 1);
      }
      
      // lex
      for (int i = 0; i < (int)v_trg.size(); ++i) {
	const string& trg = v_trg[i];
	for (int j = 0; j < (int)v_src.size(); ++j) {
	  const string& src = v_src[j];
	  lex->set(trg, src, 1);
	}
      }
      
      // jump
      int full_len_trg = v_trg.size();
      int non_empty_len_trg = full_len_trg / 2;
      if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
	non_empty_len_trg = full_len_trg;
      } else {
	assert(full_len_trg % 2 == 0);
      }
      
      for (int i_pre = 0; i_pre < non_empty_len_trg; ++i_pre) {
	if (i_pre == 0) {
	  assert(jump->get_word_type(v_trg[i_pre]) == BOS);
	} else if ((i_pre >= 1) && (i_pre < non_empty_len_trg - 1)) {
	  assert(jump->get_word_type(v_trg[i_pre]) == NON_EMPTY);
	} else if (i_pre == non_empty_len_trg - 1) {
	  assert(jump->get_word_type(v_trg[i_pre]) == EOS);
	} else {
	  assert(jump->get_word_type(v_trg[i_pre]) == EMPTY);
	}      
	
	for (int i = 0; i < non_empty_len_trg; ++i) {
	  if (i == 0) {
	    assert(jump->get_word_type(v_trg[i]) == BOS);
	  } else if ((i >= 1) && (i < non_empty_len_trg - 1)) {
	    assert(jump->get_word_type(v_trg[i]) == NON_EMPTY);
	  } else if (i == non_empty_len_trg - 1) {
	    assert(jump->get_word_type(v_trg[i]) == EOS);
	  } else {
	    assert(jump->get_word_type(v_trg[i]) == EMPTY);
	  }

	  WORD_TYPE wt_pre = jump->get_word_type(v_trg[i_pre]);
	  WORD_TYPE wt = jump->get_word_type(v_trg[i]);
	  if (wt_pre == BOS) {
	    jump->set(FROM_BOS, i_pre, i, 1);
	  } 
	  if ((wt_pre == NON_EMPTY) && (wt == NON_EMPTY)) {
	    jump->set(OTHERS, i_pre, i, 1);
	  }
	  if (wt == EOS) {
	    jump->set(TO_EOS, i_pre, i, 1);
	  }
	}
      }
    // lambda
      for (int i = 0; i < (int)v_trg.size(); ++i) {
	const string& trg = v_trg[i];
	lambda->set(trg, 1);
      }
    }  // while
    cerr << endl;
    
    return;
  }
  
  void doit() {
    // length statistics
    unordered_map<int, int> length_trg;
    unordered_map<int, int> length_src;

    // use this to remember the frequency of words in the target language

    // freq
    Parameter_Freq trg_freq;
    Parameter_Freq src_freq;

    // lex
    Parameter_Lex lex;

    // jump
    Parameter_Jump jump;
    // lambda
    Parameter_Lambda lambda;
    
    process(&length_trg, &length_src, &trg_freq, &src_freq, &lex, &jump, &lambda);
    
    
    // length_trg
    {
      for (unordered_map<int, int>::const_iterator it1 = length_trg.begin(); it1 != length_trg.end(); ++it1) {
	const int& length = it1->first;
	int freq_count = it1->second;
	cout << "length_trg" << "\t" << length << "\t" << freq_count << endl;
      }
      cout << endl;
    }
    
    // length_src
    {
      for (unordered_map<int, int>::const_iterator it1 = length_src.begin(); it1 != length_src.end(); ++it1) {
	const int& length = it1->first;
	int freq_count = it1->second;
	cout << "length_src" << "\t" << length << "\t" << freq_count << endl;
      }
      cout << endl;
    }
    
    // freq
    trg_freq.dump(cout, "trg-freq");
    cout << endl;

    src_freq.dump(cout, "src-freq");
    cout << endl;

    // lex
    lex.normalize();
    lex.dump(cout, SMALL_LEX_SMOOTH);
    cout << endl;
    
    // jump
    jump.normalize();
    jump.dump(cout, SMALL_JUMP_SMOOTH);
    cout << endl;
    
    // lambda
    lambda.normalize(trg_freq);
    lambda.dump(trg_freq, src_freq, SMALL_LAMBDA_SMOOTH, cout);
    cout << endl;
    
    return;
  }

}

