
#include "prob-table-util.h"
#include "config/global.h"
#include "util/string-util.h"

namespace WORDALIGN_COMMON {

  /////////////////////////////////////////////////////////////////////
  // Parameter_Lex

  void Parameter_Lex::normalize() {

    for (unordered_map<string, unordered_map<string, double> >::iterator it1 = parameter_lex_.begin(); it1 != parameter_lex_.end(); ++it1) {
      const string& trg = it1->first;
      UTIL::do_nothing(trg);

      unordered_map<string, double>& tmp = it1->second;
      
      double sum = 0;
      for (unordered_map<string, double>::iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	// int width = it1->first;
	double count = it2->second;
	sum += count;
      }

      assert(sum > 0);

      for (unordered_map<string, double>::iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	// int width = it1->first;
	double& count = it2->second;
	count /= sum;
      }
    }

    return;
  }

  void Parameter_Lex::dump(ostream& out, double eps) const {

    for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = parameter_lex_.begin(); it1 != parameter_lex_.end(); ++it1) {
      const string& trg = it1->first;
      const unordered_map<string, double>& tmp = it1->second;
      
      for (unordered_map<string, double>::const_iterator it2 = tmp.begin(); it2 != tmp.end(); ++it2) {
	const string& src = it2->first;
	double parameter = it2->second;

	if (parameter > eps) {
	  out << "lex" << "\t" << trg << "\t" << src << "\t" << parameter << endl;
	}
      }
    }

    return;
  }

  double Parameter_Lex::get(const string& e, const string& f, double eps) const {
    
    if (eps <= 0) {
      eps = SMALL_LEX_SMOOTH;
    }
    
    const string& trg = e;
    const string& src = f;

    WORDALIGN_COMMON::WORD_TYPE wt_trg = Parameter_Jump::get_word_type(trg);
    WORDALIGN_COMMON::WORD_TYPE wt_src = Parameter_Jump::get_word_type(src);

    if (wt_src == WORDALIGN_COMMON::BOS) {
      if (wt_trg == wt_src) {
	return 1;
      } else {
	return 0;
      }
    } else if (wt_src == WORDALIGN_COMMON::EOS) {
      if (wt_trg == wt_src) {
	return 1;
      } else {
	return 0;
      }
    } else {
      assert(wt_src == WORDALIGN_COMMON::NON_EMPTY);
      if (wt_trg == WORDALIGN_COMMON::BOS) {
	return 0;
      } else if (wt_trg == WORDALIGN_COMMON::EOS) {
	return 0;
      } else {
	// go on
      }
    }

    assert(wt_trg != WORDALIGN_COMMON::BOS);
    assert(wt_trg != WORDALIGN_COMMON::EOS);
    assert(wt_src != WORDALIGN_COMMON::BOS);
    assert(wt_src != WORDALIGN_COMMON::EOS);
    assert(wt_src != WORDALIGN_COMMON::EMPTY);
    assert(wt_src == WORDALIGN_COMMON::NON_EMPTY);

    unordered_map<string, unordered_map<string, double> >::const_iterator it1 = parameter_lex_.find(trg);
    if (it1 == parameter_lex_.end()) {
      return eps;
    }
	   
    const unordered_map<string, double>& tmp = it1->second;
    unordered_map<string, double>::const_iterator it2 = tmp.find(src);
    if (it2 == tmp.end()) {
      return eps;
    }

    double d = it2->second + eps;
    if (d >= 1) {
      d = 1;
    }
    return d;
  }

  void Parameter_Lex::set(const string& e, const string& f, double d) {
    const string& trg = e;
    const string& src = f;
    parameter_lex_[trg][src] = d;

    return;
  }

  void Parameter_Lex::add(const string& e, const string& f, double d) {
    const string& trg = e;
    const string& src = f;
    parameter_lex_[trg][src] += d;

    return;
  }

  void Parameter_Lex::get_keys(vector<string>* keys) const {
    assert(keys != NULL);
    assert(keys->size() == 0);
    
    for (unordered_map<string, unordered_map<string, double> >::const_iterator it = parameter_lex_.begin(); it != parameter_lex_.end(); ++it) {
      const string& str = it->first;
      keys->push_back(str);
    }
    return;
  }

  int Parameter_Lex::key_size() const {
    int size = parameter_lex_.size();
    return size;
  }

  int Parameter_Lex::get_total_size() const {
    int size = 0;
    for (unordered_map<string, unordered_map<string, double> >::const_iterator it = parameter_lex_.begin(); it != parameter_lex_.end(); ++it) {
      const unordered_map<string, double> tmp = it->second;
      size += tmp.size();
    }
    return size;
  }

  /*
  void Parameter_Lex::clear() {
    for (unordered_map<string, unordered_map<string, double> >::iterator it1 = parameter_lex_.begin(); it1 != parameter_lex_.end(); ++it1) {
      unordered_map<string, double>& map_tmp = it1->second;
      for (unordered_map<string, double>::iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	it2->second = 0;
      }
    }
    return;
  }
  */

  /*
  void Parameter_Lex::init(const string& info) {
    if (info == "uniform") {
      for (unordered_map<string, unordered_map<string, double> >::iterator it1 = parameter_lex_.begin(); it1 != parameter_lex_.end(); ++it1) {
	unordered_map<string, double>& map_tmp = it1->second;
	const int size = map_tmp.size();
	for (unordered_map<string, double>::iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	  it2->second = 1.0 / size;
	}
      }
    } else {
      assert(false);
    }
    return;
  }
  */

  /*
  void Parameter_Lex::normalize_using_counts(const unordered_map<string, unordered_map<string, double> > total_expected_count_lex, ostream& out) {

    parameter_lex_.clear();

    for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = total_expected_count_lex.begin(); it1 != total_expected_count_lex.end(); ++it1) {
      const string& trg = it1->first;
      
      double sum = 0;
      const unordered_map<string, double>& map_tmp = it1->second;
      for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	double parameter = it2->second;
	sum += parameter;
      }
      
      unordered_map<string, double>& tmp2 = parameter_lex_[trg];
      for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	const string& src = it2->first;
	double parameter = it2->second;
	double parameter2 = SMALL_LEX_SMOOTH + parameter / sum;
	tmp2[src] = parameter2;
      }
    }
    return;
  }
  */

  void Parameter_Lex::get_topk(int TOPK,
			  //
			  priority_queue<pair<double, string>, vector<pair<double, string> >, topk>* my_topk) const {

    // priority_queue<pair<double, string> > my_topk(topk);
    for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = parameter_lex_.begin(); it1 != parameter_lex_.end(); ++it1) {
      const string& trg = it1->first;
      const unordered_map<string, double>& map_tmp = it1->second;
      
      for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	double parameter = it2->second;
	my_topk->push(make_pair(parameter, it2->first + " | " + trg));
	if ((int)my_topk->size() > TOPK) {
	  my_topk->pop();
	  assert((int)my_topk->size() <= TOPK);
	}
      }
    }
    return;
  }

  void Parameter_Lex::get_topk(int TOPK,
			  //
			  priority_queue<pair<double, string>, vector<pair<double, pair<string, string> > >, topk3>* my_topk) const {

    // priority_queue<pair<double, string> > my_topk(topk);
    for (unordered_map<string, unordered_map<string, double> >::const_iterator it1 = parameter_lex_.begin(); it1 != parameter_lex_.end(); ++it1) {
      const string& trg = it1->first;
      
      if (trg == MYCONFIG::my_global::FLAGS_empty_word) {
	continue;
      }
      if (MYCONFIG::my_global::t_freq_[trg] >= MYCONFIG::my_global::FLAGS_freq_word_count) {
	const unordered_map<string, double>& map_tmp = it1->second;
	
	for (unordered_map<string, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	  double parameter = it2->second;
	  my_topk->push(make_pair(parameter, make_pair(it2->first, trg)));
	  if ((int)my_topk->size() > TOPK) {
	    my_topk->pop();
	    assert((int)my_topk->size() <= TOPK);
	  }
	}
      }
    }
    return;
  }

  /////////////////////////////////////////////////////////////////////
  // Parameter_Jump

  void Parameter_Jump::set(JUMP_TYPE jt, int i_pre, int i, double d) {

    if (MYCONFIG::my_global::FLAGS_combine_jump) {
      jt = OTHERS;
    }

    parameter_jump_[jt][abs(i_pre - i)] = d;
  }

  void Parameter_Jump::set2(JUMP_TYPE jt, int width, double d) {

    assert(width >= 0);

    if (MYCONFIG::my_global::FLAGS_combine_jump) {
      jt = OTHERS;
    }

    parameter_jump_[jt][width] = d;
  }

  void Parameter_Jump::add(JUMP_TYPE jt, int i_pre, int i, double d) {

    if (MYCONFIG::my_global::FLAGS_combine_jump) {
      jt = OTHERS;
    }

    parameter_jump_[jt][abs(i_pre - i)] += d;
  }

  void Parameter_Jump::add2(JUMP_TYPE jt, int width, double d) {

    assert(width >= 0);

    if (MYCONFIG::my_global::FLAGS_combine_jump) {
      jt = OTHERS;
    }

    parameter_jump_[jt][width] += d;
  }

  void Parameter_Jump::normalize() {

    for (int jt = 0; jt < 3; ++jt) {

      if (MYCONFIG::my_global::FLAGS_combine_jump) {
	if (jt != WORDALIGN_COMMON::OTHERS) {
	  continue;
	}
      }

      double sum = 0;
      for (unordered_map<int, double>::const_iterator it1 = parameter_jump_[jt].begin(); it1 != parameter_jump_[jt].end(); ++it1) {
	// int width = it1->first;
	double count = it1->second;
	sum += count;
      }
    
      assert(sum > 0);
      
      for (unordered_map<int, double>::const_iterator it1 = parameter_jump_[jt].begin(); it1 != parameter_jump_[jt].end(); ++it1) {
	const int& width = it1->first;
	double parameter = it1->second;
	parameter_jump_[jt][width] = parameter / sum;
      }

    }

    return;
  }

  void Parameter_Jump::dump(ostream& out, double eps) const {

    for (int jt = 0; jt < 3; ++jt) {
      for (unordered_map<int, double>::const_iterator it1 = parameter_jump_[jt].begin(); it1 != parameter_jump_[jt].end(); ++it1) {
	const int& width = it1->first;
	double parameter = it1->second;
	
	if (parameter > eps) {
	  out << "jump" << "\t" << jt << "\t" << width << "\t" << parameter << endl;
	}
      }
    }

    return;
  }

  double Parameter_Jump::get0(JUMP_TYPE jt, int width, int III) const {
    assert(width >= 0);

    if (MYCONFIG::my_global::FLAGS_combine_jump) {
      assert(jt == OTHERS);
    }
    
    double parameter = SMALL_JUMP_SMOOTH + get_1(parameter_jump_[jt], width, SMALL_JUMP_SMOOTH);

    return 0.8 * parameter + 0.2 * 1.0 / III;
  }

  double Parameter_Jump::get(JUMP_TYPE jt, int i_pre, int i, int III, double eps) const {

    if (MYCONFIG::my_global::FLAGS_combine_jump) {
      jt = OTHERS;
    }

    if (false) {
      cerr << endl;
      cerr << "jt: " << jt << endl;
      cerr << "i_pre: " << i_pre << endl;
      cerr << "i: " << i << endl;
    }

    if (jt == 1) {
      // cerr << "aha" << endl;
    }

    double p0 = MYCONFIG::my_global::FLAGS_p0_;

    int II = III / 2;
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      II = III;
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(III % 2 == 0);
    } else {
      assert(false);
    }

    if (p0 == 0) {
      if (i_pre >= II) {
	return 0;
      }
      if (i >= II) {
	return 0;
      }
    }

    if (jt == FROM_BOS) {
      // i_pre has to be 0
      if (i_pre != 0) {
	return 0;
      } else {
	assert(i_pre == 0);
	// it can not jump beyong II
	if (i > II) {
	  return 0;
	}
      }
    } else if (jt == TO_EOS) {
      // i has to be II
      if (i != II - 1) {
	return 0;
      } else {
	assert(i == II - 1);
	// goes to </s> with prob 1
	return 1;
      }
    } else {
      /*
      if (i_pre == II) {
	if (i != II) {
	  return 0;
	}
      }
      */
    }

    // emtpy words
    if (i >= II) {
      // jump to empty words
      // Note: we can jump to empty words at the very begining
      if (abs(i_pre - i) % II == 0) {
	return p0;
      } else {
	return 0;
      }
    } 

    assert(i < II);
    
    // jump from empty words
    if (i_pre >= II) {
      i_pre -= II;
    }

    assert(i_pre < II);

    static unordered_map<int, unordered_map<int, unordered_map<int, double> > > cache;
    static bool use_cache = true;
    
    double parameter = get0(jt, abs(i_pre - i), III);
    // cerr << "parameter: " << parameter << endl;

    double sum = 0;
    unordered_map<int, double> m;
    if (!use_cache || (cache[jt][i_pre].find(III) == cache[jt][i_pre].end())) {
      if (use_cache) {
	assert(cache[jt][i_pre][III] == 0);
      }
      for (int ii = 0; ii <  III; ++ii) {
	double tmp = get0(jt, abs(i_pre - ii), III);
	sum += tmp;
	m[ii] = tmp;
      }
      if (sum == 0) {
	assert(parameter == 0);
      }
      cache[jt][i_pre][III] = sum;

    } else {
      sum = cache[jt][i_pre][III];
    }
    
    assert(sum >= 0);
    if (sum == 0) {
      assert(parameter == 0);
      return 0;
    } else {
      return (1 - p0) * parameter / sum;
    }
  }

  int Parameter_Jump::key_size() const {
    int size = 0;
    for (int jt = 0; jt < 3; ++jt) {
      size += parameter_jump_[jt].size();
    }
    return size;
  }

  /*
  void Parameter_Jump::clear() {
    for (int jt = 0; jt < 3; ++jt) {
      parameter_jump_[jt].clear();
    }
    return;
  }
  */

  WORD_TYPE Parameter_Jump::get_word_type(const string& str) {

    if (str == MYCONFIG::my_global::FLAGS_BOS) {
      return BOS;
    } else if (str == MYCONFIG::my_global::FLAGS_EOS) {
      return EOS;
    } else if (str == MYCONFIG::my_global::FLAGS_empty_word) {
      return EMPTY;
    } else {
      assert(str != MYCONFIG::my_global::FLAGS_BOS);
      assert(str != MYCONFIG::my_global::FLAGS_EOS);
      assert(str != MYCONFIG::my_global::FLAGS_empty_word);
      return NON_EMPTY;
    }
  }

  /*
  void Parameter_Jump::init(const string& info) {
    if (info == "uniform") {
      for (unordered_map<int, double>::iterator it1 = parameter_jump_[jt].begin(); it1 != parameter_jump_[jt].end(); ++it1) {
	// it2->second = 1.0 / size;
	it1->second = 1.0;
      }
    } else {
      assert(false);
    }
  }
  */

  /*
  void Parameter_Jump::normalize_using_counts(const unordered_map<int, unordered_map<int, double> > total_expected_count_jump, ostream& out, bool output) {

    unordered_map<int, double> sum_width;
    for (unordered_map<int, unordered_map<int, double> >::const_iterator it1 = total_expected_count_jump.begin(); it1 != total_expected_count_jump.end(); ++it1) {
      const int& i_pre = it1->first;
      
      const unordered_map<int, double>& map_tmp = it1->second;
      for (unordered_map<int, double>::const_iterator it2 = map_tmp.begin(); it2 != map_tmp.end(); ++it2) {
	const int& i = it2->first;
	double count = it2->second;
	
	// cerr << "width " << abs(src - trg) << endl;
	
	sum_width[abs(i - i_pre)] += count;
      }
    }
    
    // compute sum for normalization
    double sum = 0;
    for (unordered_map<int, double>::const_iterator it1 = sum_width.begin(); it1 != sum_width.end(); ++it1) {
      double parameter = it1->second;
      sum += parameter;
    }
    
    assert(sum > 0);
    
    clear();

    for (unordered_map<int, double>::const_iterator it1 = sum_width.begin(); it1 != sum_width.end(); ++it1) {
      const int& width = it1->first;
      double parameter = it1->second;
      parameter_jump_[jt][width] = parameter / sum;
      
      if (true || MYCONFIG::my_global::FLAGS_output_parameter) {
	out << "jump" << "\t" << width << "\t" << parameter / sum << endl;
      }
    }

    return;
  }
  */

  // Parameter_Lamdba
  void Parameter_Lambda::normalize(const Parameter_Freq& parameter_trg_freq) {

    for (unordered_map<string, unordered_map<int, unordered_map<int, double> > >::const_iterator it1 = parameter_lambda_raw_.begin(); it1 != parameter_lambda_raw_.end(); ++it1) {
      const string& trg = it1->first;
      const unordered_map<int, unordered_map<int, double> >& tmp1 = it1->second;

      if (trg == "_empty_word_") {
	assert(trg != "");
      }

      assert(parameter_lambda_normalized_[trg] == 0);

      double sum1 = 0;
      double sum2 = 0;
      for (unordered_map<int, unordered_map<int, double> >::const_iterator it2 = tmp1.begin(); it2 != tmp1.end(); ++it2) {

	int fert = it2->first;
	const unordered_map<int, double>& tmp2 = it2->second;

	for (unordered_map<int, double>::const_iterator it3 = tmp2.begin(); it3 != tmp2.end(); ++it3) {
	  int shared_by = it3->first;
	  double cnt = it3->second;
	  
	  assert(cnt > 0);
	  sum1 += cnt;
	  sum2 += fert * cnt / shared_by;
	}
      }	

      double count = parameter_trg_freq.get(trg);
      assert(fabs(count - sum1) < 1);
      
      parameter_lambda_normalized_[trg] = sum2 / sum1;
    }

    return;
  }

  double Parameter_Lambda::get(const string& str, double eps) const {
    unordered_map<string, double>::const_iterator it = parameter_lambda_normalized_.find(str);
    if (it == parameter_lambda_normalized_.end()) {
      return eps;
    } else {
      return it->second;
    }
  }


  double Parameter_Lambda::get_lambda_for_empty_word(
						     const Parameter_Freq& parameter_trg_freq,
						     const Parameter_Freq& parameter_src_freq
						     ) const {

    double d = get(MYCONFIG::my_global::FLAGS_empty_word, SMALL_LAMBDA_SMOOTH);
    if (d < SMALL_LAMBDA_SMOOTH) {
      // d = SMALL_LAMBDA_SMOOTH;
    }
    return d;
  }

  double Parameter_Lambda::get_lambda_for_non_empty_word(
							 const Parameter_Freq& parameter_trg_freq,
							 const Parameter_Freq& parameter_src_freq
							 ) const {
    int trg_sum = parameter_trg_freq.get_non_empty_word_size();
    assert(trg_sum > 0);

    int src_sum = parameter_src_freq.get_non_empty_word_size();
    assert(src_sum > 0);

    double lambda = 1.0 * src_sum / trg_sum;
    return lambda;
  }

  void Parameter_Lambda::set(const string& str, double d) {
    parameter_lambda_normalized_[str] = d;
  }

  void Parameter_Lambda::add(const string& str, int fert, int shared_by, double d) {
    if (Parameter_Jump::get_word_type(str) == EMPTY) {
    } else if (Parameter_Jump::get_word_type(str) == NON_EMPTY) {
      assert(shared_by == 1);
    }
    parameter_lambda_raw_[str][fert][shared_by] += d;
  }

  void Parameter_Lambda::dump(
			      const Parameter_Freq& parameter_trg_freq, 
			      const Parameter_Freq& parameter_src_freq,
			      double eps,
			      ostream& out
			      ) const {

    for (unordered_map<string, double>::const_iterator it1 = parameter_lambda_normalized_.begin(); it1 != parameter_lambda_normalized_.end(); ++it1) {
      const string& trg = it1->first;
      double lambda = it1->second;

      if (trg == "_empty_word_") {
	assert(trg != "");
      }

      if ((lambda > eps) || (trg == "_empty_word_")) {
	const int fert = 1;
	const int shared_by = 1;
	out << "lambda" << "\t" << trg << "\t" << fert << "\t" << shared_by << "\t" << lambda << endl;
      }
    }

    return;
  }

  int Parameter_Lambda::key_size() const {
    int size = parameter_lambda_normalized_.size();
    return size;
  }

  // Parameter_Freq
  Parameter_Freq::Parameter_Freq() {
    non_empty_word_size = 0;
    empty_word_size = 0;
  }

  double Parameter_Freq::get(const string& str) const {

    unordered_map<string, int>::const_iterator it = parameter_freq.find(str);
    if (it == parameter_freq.end()) {
      return 0;
    } else {
      return it->second;
    }
  }

  void Parameter_Freq::add(const string& str, int freq) {

    parameter_freq[str] += freq;
    
    if (Parameter_Jump::get_word_type(str) == NON_EMPTY) {
      non_empty_word_size += freq;
    } else if (Parameter_Jump::get_word_type(str) == EMPTY) {
      empty_word_size += freq;
    }

    return;
  }

  void Parameter_Freq::dump(ostream& out, const string& info) const {

    for (unordered_map<string, int>::const_iterator it1 = parameter_freq.begin(); it1 != parameter_freq.end(); ++it1) {
      const string& trg = it1->first;
      int freq = it1->second;

      out << info << "\t" << trg << "\t" << freq << endl;
    }

    return;
  }

  int Parameter_Freq::key_size() const {
    int size = parameter_freq.size();
    return size;
  }

  int Parameter_Freq::get_non_empty_word_size() const {
    
    return non_empty_word_size;
  }

  int Parameter_Freq::get_empty_word_size() const {

    return empty_word_size;
  }

}  // namespace UTIL
