
//
// string utilities
//

#ifndef _util_parameter_table_h_
#define _util_parameter_table_h_

#include "wordalign-common/topk.h"
#include "util/common.h"
#include "util/unordered-util.h"
#include "config/global.h"
#include <queue>
#include <vector>

namespace WORDALIGN_COMMON {

  const int TRG_LEN_LIMIT = 50;
  const int SRC_LEN_LIMIT = 50;

  const double SMALL_LEX_SMOOTH = 1.0e-8;
  const double SMALL_JUMP_SMOOTH = 1.0e-4;
  const double SMALL_LAMBDA_SMOOTH = 1.0e-4;
  
  // Parameter_Lex
  // used by IBM1 and HMM
  class Parameter_Lex {
  private:
    // The first string is e; the second string is f: p(f|e)
    // After nomalization, sum over f is 1
    unordered_map<string, unordered_map<string, double> > parameter_lex_;
  public:

    void normalize();
    void dump(ostream& out, double eps) const;

    double get(const string& e, const string& f, double eps) const;
    void set(const string& e, const string& f, double d);
    void add(const string& e, const string& f, double d);
    int key_size() const;
    int get_total_size() const;
    // void init(const string& info);
    // void clear();
    // void normalize_using_counts(const unordered_map<string, unordered_map<string, double> > total_expected_count_lex, ostream& out);

    void get_keys(vector<string>* keys) const;
    void get_topk(int TOPK,
		  //
		  priority_queue<pair<double, string>, vector<pair<double, string> >, topk>* my_topk) const;
    void get_topk(int TOPK,
		  //
		  priority_queue<pair<double, string>, vector<pair<double, pair<string, string> > >, topk3>* my_topk) const;

  };


  enum JUMP_TYPE {FROM_BOS = 0, OTHERS = 1, TO_EOS = 2};
  enum WORD_TYPE {BOS = 0, NON_EMPTY = 1, EMPTY = 2, EOS = 3};

  // Parameter_Jump
  // used by IBM1 and HMM
  class Parameter_Jump {
  private:
    unordered_map<int, double> parameter_jump_[3];
    double get0(JUMP_TYPE jt, int width, int III) const;
    
  public:
    void normalize();
    void dump(ostream& out, double eps) const;

    double get(JUMP_TYPE jt, int i_pre, int i, int I, double eps) const;
    void set(JUMP_TYPE jt, int i_pre, int i, double d);
    void set2(JUMP_TYPE jt, int width, double d);
    void add(JUMP_TYPE jt, int i_pre, int i, double d);
    void add2(JUMP_TYPE jt, int width, double d);
    int key_size() const;
    // void init(const string& info);
    // void clear();
    // void normalize_using_counts(const unordered_map<int, unordered_map<int, double> > total_expected_count_jump, ostream& out, bool output);

    static WORD_TYPE get_word_type(const string& str);

  };

  class Parameter_Freq {
  private:
    unordered_map<string, int> parameter_freq;
    int non_empty_word_size;
    int empty_word_size;

  public:
    Parameter_Freq();
    double get(const string& str) const;
    void add(const string& str, int freq);
    int key_size() const;
    int get_non_empty_word_size() const;
    int get_empty_word_size() const;

    void dump(ostream& out, const string& info) const;
  };

  class Parameter_Lambda {
  private:
    unordered_map<string, unordered_map<int, unordered_map<int, double> > > parameter_lambda_raw_;
    unordered_map<string, double> parameter_lambda_normalized_;    

  public:
    void normalize(const Parameter_Freq& parameter_freq);
    double get(const string& str, double eps) const;
    double get_lambda_for_non_empty_word(
					 const Parameter_Freq& parameter_trg_freq,
					 const Parameter_Freq& parameter_src_freq
					 ) const;
    double get_lambda_for_empty_word(
				     const Parameter_Freq& parameter_trg_freq,
				     const Parameter_Freq& parameter_src_freq
				     ) const;
    void set(const string& str, double d);
    void add(const string& str, int fert, int shared_by, double d);
    int key_size() const;
    
    void dump(
	      const Parameter_Freq& parameter_trg_freq, 
	      const Parameter_Freq& parameter_src_freq,
	      double eps,
	      ostream& out
	      ) const;
  };

}  // namespace UTIL

#endif  // _util_parameter_table_h_
