
// fet1-main.cc

#include "config/config.h"
#include "config/global.h"
#include "util/string-util.h"
#include "process-one-line.h"

namespace WORDALIGN_COMMON {

  bool Wordalign::process_one_line(const string& line,
				   int trg_len_limit,
				   int src_len_limit,
				   //
				   vector<string>* v_trg,
				   vector<string>* v_src) {
    
    vector<string> v;
    UTIL::split_string_using(line, "\t", &v);
    if (v.size() != 2) {
      return false;
    }
      
    // change trg/v[0] to lower case
    UTIL::to_lower(&v[0]);
    
    vector<string> v_trg0, v_src0;
    if (MYCONFIG::my_global::FLAGS_data_format_reverse) {
      // reverse
      UTIL::split_string_using(v[1], " ", &v_trg0);
      UTIL::split_string_using(v[0], " ", &v_src0);
    } else {  
      // normal
      UTIL::split_string_using(v[0], " ", &v_trg0);
      UTIL::split_string_using(v[1], " ", &v_src0);
    }
    
    if (trg_len_limit > 0) {
      if ((int)v_trg0.size() > trg_len_limit) {
	// cout << "ignore\t" << line << endl;
	return false;
      }
    }
    
    if (src_len_limit > 0) {
      if ((int)v_src0.size() > src_len_limit) {
	// cout << "ignore\t" << line << endl;
	return false;
      }
    }
    
    assert(v_trg0.size() >= 1);
    assert(v_src0.size() >= 1);

    // add <s> and </s> to the source sentence
    v_src->push_back(MYCONFIG::my_global::FLAGS_BOS);
    for (int i = 0; i < (int)v_src0.size(); ++i) {
      v_src->push_back(v_src0[i]);
    }
    v_src->push_back(MYCONFIG::my_global::FLAGS_EOS);
    
    // add <s> and </s> to the target sentence
    v_trg->push_back(MYCONFIG::my_global::FLAGS_BOS);
    for (int i = 0; i < (int)v_trg0.size(); ++i) {
      v_trg->push_back(v_trg0[i]);
    }
    v_trg->push_back(MYCONFIG::my_global::FLAGS_EOS);

    // add empty words if any
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 1) {
      v_trg->push_back(MYCONFIG::my_global::FLAGS_empty_word);
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      for (int i = 0; i < (int)v_trg0.size(); ++i) {
	v_trg->push_back(MYCONFIG::my_global::FLAGS_empty_word);
      }
      v_trg->push_back(MYCONFIG::my_global::FLAGS_empty_word);
      v_trg->push_back(MYCONFIG::my_global::FLAGS_empty_word);
    }

    //
    assert(v_src0.size() >= 1);
    assert(v_src->size() >= 3);
    assert(v_src0.size() + 2 == v_src->size());
    
    assert(v_trg0.size() >= 1);
    assert(v_trg->size() >= 3);
    if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 0) {
      assert(v_trg0.size() + 2 == v_trg->size());
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 1) {
      assert(v_trg0.size() + 2 + 1 == v_trg->size());
    } else if (MYCONFIG::my_global::FLAGS_number_of_empty_words == 2) {
      assert(v_trg0.size() + 2 + v_trg0.size() + 2 == v_trg->size());
    } else {
      assert(false);
    }
    
    return true;
  }
  
}  // namespace
